;;Copyright (C) 2001 Soren Goodman
;;Algnalogthmic Music Score Generator
;;See readme.txt for more info

;;documentation.txt
;;all accessible from within lisp :)

;;to do...UPDATE DOCUMENTATION...I've been a bit lazy lately, since I moved the documentation to a seperate file,
;;so there are a few outdated prompts...I'll update buffer, multi-ramp and sequencer now, since they all need
;;a a buffer-ref number to work....buffer no longer stores in an array...

(in-package :cl-user)

(defvar *help-verbose* nil)

(setq *INTRODUCTION*
"******************************INTRODUCTION******************************
The Algnalogthmic Music Score Generator has been designed to be a powerful creative tool for
creating music score files, such as midi, csound-scores, traditional (or untraditional) music 
notation files.  It is intended to be an interactive tool, good music will only come if you
spend enough time and energy constructing the structure of a piece, you cannot simply hit 'go'
until you've been working for a while and already hit 'go' a number of times...

I wrote this code because I discovered I love creating/shaping form and instrument interaction
processes, as well as the tools used to write music in the electronic music medium.  I started
writing this with the intention of creating a set of tools, similar to an analog synthesizer, or
many of todays modern synthesis/sequencer tools, but with the sole intention of creating musical
scores, patterns of notes, so to speak.  I wanted to be able to write a piece of music in form/process, 
then be able to go into the software and have it realize the actual little details of notes and such..
I have a tendancy to get too lazy when writing music by hand much of the time, so this code
is intending to fix that problem.")




(setq *BASIC-FORMAT*
"******************************BASIC-FORMAT******************************
The best way to see how the program works is to look at a sample score file.  Basically, there are 
a number of elements in a score-format.  A typical music score contains :RHYTHM :PITCH :DURATION and 
:DYNAMICS data.  There are also places for :INSTRUMENT, :CLOCK-OFFSET (for making a phrase start
later) and :TEMPO.  There are a number of high level functions, such as ALL-SEGMENTS, that deal with
how each element-list is created.  MODULAR-phrase simply has a parameter for each element and expects 
one of three things.  1) straight data, i.e. a list of numbers.  2) a function call that will eventually
end up with said data, or 3) nil.  If the output is nil, there is some sort of default setting, usually contained
within the associated instrument.

The basic output format is the list, and everything is divided beat by beat.  So the output for a :RHYTHM slot
could look something like (1 0 0 1 0 1 0 0 0 0 1 1 0), with 1 being an attack and 0 being no new attack.  Some
output (such as rhythm) is in a per beat format, whereas some output (such as pitch) is in a per-attack output.
It is also possible to create rhythmic output that is initially per-attack (i.e., beats to next attack..so the
above list would translate to '(3 2 6 1 2)).  :tempo is smallest-beat-per-minute, and :clock-offset is simply
number of beats.  For the :INSTRUMENT data, a special set of objects has been created.  Each instrument
contains relevent data, such as max-volume, max-duration, range-hi, range-lo, channel (for midi), etc.  These
are implemented in the performance (saving) of a piece as well as being available to any section that may use
them (there is no built in error checking, however, for range and volume...so if a function ignores 'range-hi, 
then there might be pitches higher than that).")



(setq *FUNCTION-DETAILS*
"******************************FUNCTION-DETAILS******************************
For a list of functions type (available-functions).  For info on a specific function
type (help 'function).

There are a number of different types of functions in the ALGNOLOGTHMIC MUSIC SCORE GENERATOR.  
They are:
  Generators - generate output 
(CHANCE RANDOM-FROM-PC MULTI-RAMP RAMP)
   Filters - create output by toying with some sort of list input
(SPLIT-LIST VARIATION TRANSPOSE FIRST-N LAST-N, other CL functions such as +, -, last, etc...)
   Controllers - organize the way functions are realized
(SEQUENCE MULTI-RAMP RAMP CHANCE)

Obviously there is some overlap..for example, multi-ramp actually generates output, but this is often
used as an envelope for some parameter in another function (such as an increasing N in a call to FIRST-N).
MULTI-RAMP is usually only used as a generator for simple elements, such as dynamics, not for more complex
elements such as pitch. 

All of these functions are designed to be flexible...find fun new ways of making them work together and tell
me about it!") 

(defun help (&optional function &rest more &aux (more (clean-up-list more)))
  (cond ((null function)
         (format t ";;Copyright (C) 2001 Soren Goodman
;;Algnalogthmic Music Score Generator
;;See readme.txt for more info on license

;;See the variables *INTRODUCTION*, *BASIC-FORMAT* and *FUNCTION-DETAILS* for documentation.
;;type (help 'function-name1 'function-name2 ...) for info on a specific function, or, for a list of functions, type
;;(help 'functions) or (help 'ramps, 'utility, 'pitch or 'main).  For just the function prompts, 
;;type (help 'non-verbose), to turn documentation back on type (help 'verbose) 


;;For complete documentation spewed at you type (ALGNALOGTHMIC).~%"))
        ((eq function 'functions)
         (format t "~S~%" (available-functions t)))
        ((eq function 'verbose)(setq *help-verbose* t))
        ((eq function 'non-verbose)(setq *help-verbose* nil))
        ((eq function 'ramps)(format t "~A~%" *ramp-functions*))
        ((eq function 'utility)(format t "~A~%" *utility-functions*))
        ((eq function 'pitch)(format t "~A~%" *pitch-functions*))
        ((eq function 'main)(format t "~A~%" *main-functions*))
        ((member function (available-functions))
         (if *help-verbose*
             (let ((function-documentation (eval (read-from-string 
                                                  (concatenate 'string "*" (symbol-name function) "-documentation*")))))
               (format t "~A~%" function-documentation))
           (let ((function-prompt (eval (read-from-string 
                                         (concatenate 'string "*" (symbol-name function) "-prompt*")))))
             (format t "~A~%" function-prompt))))
        (t (format t "Sorry, no documentation available for function ~S.~%" function)))
  (when more (help (first more) (rest more))))

(defun algnalogthmic ()
  (format t "~%;;***********************~%~S~%~%" *INTRODUCTION*)
  (format t "~%;;***********************~%~S~%~%" *BASIC-FORMAT*)
  (format t "~%;;***********************~%~S~%~%" *FUNCTION-DETAILS*))


;;this is too big, causing problems...
#|                
;List: a standard list.  If the data needs to be a list, do not use a quote '.  (get-data) is not called
;until the end of a call to sequence.  The elements in LIST can be function calls, and only the selected
;call will be eval'd.
;   
;(sequencer (function-call))
;(sequencer '(item1 item2 item3))
;(sequencer '((function-call1)(function-call2)))
;
;Type: 'cycle [default] - go through the list in order
;       'pattern - haven't figured this out yet  
;       'step - go through each element in order.  Cycles becomes number of times to repeat each element,
;i.e., (sequencer '(1 2 3) :type 'step :cycles 2) --> '(1 1 2 2 3 3 1 1 2 2 3 3 ...) 
;       'step-hold - same as step, but when it gets to the last element (subject to 'bounce and :cycles)
;it repeats that atom.  i.e. (sequencer '(1 2 3) :type 'step-hold) --> '(1 2 3 3 3 3 3 3 3 ...);
;
;Note that for :type 'step, the entire sequence is buffer'd and a list of size NUM-ATTACKS is output one at a time
;when called.      
;
;Cycles: number of time to run through the list, can be any number.  If type is set to 'step, than this
;becomes the number of times to repeat each element and is rounded.  
;                  
;Motion: 'forward [default] - (front-->end)
;        'reverse - (end-->front)
;        'bounce - 1 cycle = (front-->end-->front)
;
;
;Start-pos: fixnum for position (nth) or float for percent %.  Starts at 0 [default].
;
;Reset-location: if non-nil, causes sequencer to reset the value of location from 0 --> 1 for each element of
;the list.  Default is T for :type 'cycle, nil for :type 'step.
;
;Phew!  A monster!|#

(defparameter *sequencer-documentation* 
  "This function has a number of different ways it can act.  It can work as a step-sequencer, a patterned 
sequencer (like an arpegiator) as well as a cyclic generator of sorts.  

 (sequencer (list &key (type 'step) (cycles 1) (motion 'forward) (start-pos '0) (reset-location t)
                 (location-curve 'linear)).  See source code for further details.")

(defparameter *sequencer-prompt*
  "(sequencer buffer-ref list :type 'cycle :cycles 1 :motion 'forward :start-pos '0 :reset-location 'default-t)")


;;;(defparameter *riff-documentation*
;;;    "Riff takes a list of a \"riffs\" and outputs the Nth element of the list.  Similar to NTH but
;;;works within the framework of this program by allowing ramp's and further error checking.")

(defparameter *riff-prompt*
  "(riff n list)")

(defparameter *truncate-rhythm-documentation*
  "truncate-rhythm turns makes RHYTHM last exactly BEATS.  It does this by comparing the sum of RHYTHM to BEATS
and modifying the end of RHYTHM.
Ex: (truncate-rhythm '(1 2 2) 4 :mode 'rhythm) --> '(1 2 1).")

(defparameter *truncate-rhythm-prompt* 
  "(truncate-rhythm list beats)")

(defvar *random-from-pc-documentation*
    "Random-from-pc")     

(defparameter *random-from-pc-prompt*
    "(random-from-pc pc-set :range-lo 5 :range-hi 6)")
(defparameter *variation-documentation* 
    "The variation function takes a list of input and has various ways of altering this data to create new but
related data.

(variation list :type 'num :change-percent .2 :min-vary -1 :max-vary 1 :items-to-vary 't)

List - a list
:Type - describes the basic mode, 'num for most applications, 'pch for pitch applications
:chance - a percent (0 - 1.0) indicating chance that each element will change.
:min-vary/max-vary - range of what to add to item being varied.
:elements-to-vary - tells variation to only modify items in this list.  If 't everything is 
modified, if 'nil the output is always the input.")


(defparameter *combine-documentation* 
      "COMBINE takes two phrases and merges them into one monophonic phrase, giving preference
to phrase1 over phrase2.  If cut-durations is set to T, then preference is given to attacks in
phrase2 over the sustain of phrase1 (attacks in phrase1 still dominate).  

Ex. (rhythm is given in a mixed rhythm/duration method, whereby 0 is always a rest 
([]'s show which phrase each attack/duration came from)
Note:  You must call combine w/ two phrase-symbols, this example uses rhythm lists
as examples...
(combine '(2 0 2 3 0 0 3) '(0 1 2 3 1 5 2)) ---> (2 [1] 1 [2] 2 [1] 3 [1] 2 [2] 3 [1])
(combine '(2 0 3) '(0 1 3 5) :options '(cut-durations)) ---> (1 [1] 1 [2] 1 [2] 2 [1] 5 [2])")

(defvar *split-list-documentation*
    "Takes LIST and outputs a single random derivative list of length LENGTH and
including INCLUDE.  Useful for sectionic melodies, semi-repetitive rhythms, etc.  
Anything in include is included, still ordered to the list.  If an element in include is not
in the given list, it is included at the end.

Default LENGTH is 1, so it becomes essentially pick-one, but if include is present that takes over.")


(defvar *chance-documentation*
    "CHANCE has a MAX and MIN value (fixnum), CHANCE (1.0 = 100% = max) and more-erratic
which, when set higher than one, increases the chances of getting a 0, without lowering the overall values.
I.E. more-erratic = 1; (0 1 0 1 1 2 1 2 2 3 2 3 4 2 3 4 ...).

I will be adjusting properties soon...at the moment a higher CHANCE increases the likelyhood of achieving MAX.
This is not always desirable, as max of 1 = quick attack, max of 16 = slow attack...I will fix soon.

If the cur-phrase is rhythm and min is 0, than the function loops until it is not 0 and adds the count.")


(defparameter *segment-documentation*
      "If segment is called with a function than it will handle num-attacks and location.  Output to segment
can be a list, a series of lists or a series of atoms.  This removes the need for most buffers, as segment
adjusts cur-position and location depending on the size of the current output.  Segment does not error
check for output > num-attacks, as this should be done on a higher level (modular-phrase) and is not
always desirable.")

(defvar *composition-documentation*
    "A composition is a collection of sections, each with a specified clock-offset in a specified sequence.
Use the function LAST-BEAT when you want one section to directly follow the next.  
Input format is (composition '((section-symbol1 clock-offset-symbol1)(etc.))")

(defparameter *multi-ramp-documentation*
    "Multi-ramp creates a ramp between elements of :ramp-list.  :type is the ramp type (use sequencer for changing)
and :quantize is the value to make each element divisible by.  Location is reset for each segment.

Note that if you have (multi-ramp '(1 2 3) :quantize 1) the segments are 1->2 and 2->3, so the output will look like
(1 1 2 2 2 2 3 3)...i.e., extra 2's...this is intentional... 

Multi-ramp can be used to create envelopes for other functions, as well as being used as a step sequencer.

Quantize is a number to quantize the output to, and type is the type of the ramp.  
The current type options are:
  linear
  pch-linear
  step [better to use sequencer]") 


(defparameter *multi-ramp-prompt*
  "(multi-ramp (ramp-list &optional (ramp-ref 0) &key (type 'linear)(segment-ramp 'linear)(quantize nil)")

(defvar *ramp-documentation* 
    "Creates a QUANTIZED ramp between ITEM1 and ITEM2 of TYPE.")

(defvar *ramp-prompt*
    "(ramp item1 item2 &key (type 'linear)(quantize nil))")


(defvar *buffer-documentation*
    "BUFFER is intended as a multifunctional utility.  It has the ability to \"convert\" output, from single
atoms to lists of atoms and vise versa.  It also has the ability to set a limit on the length of an incoming list,
truncating and only outputing a phrase of it.  It can also be used to lengthen data by repeatition.

Use buffer to:
1) create a list from a function that only outputs single numbers (use segment for simple applications).
2) create a list of a certain length and repeat/loop it for the full duration
3) make a list output only single atoms at a time, until the list is expired (at which point it re'evals for a list)

(buffer input &key 
        (output ['single], 'list)
        (storage-length ['1.0% or '5] or Fixnum/amount)
        (repeat [0] fixnum=num-repeats, float=% num-attacks to repeat.  This only applies to subsequent calls.))

Input - a 'single or 'list

Output - format of output, 'single or 'list

Output-length - Percent-float or fixnum.  Dictates how big to make the buffer.  If input is a list, then it is
truncated (or called again, if necessary) until it reaches the output-length.  If input is a single item then
it is called again (with adjusting location) until it reaches output-length.  
If output-length is a float, it is a percent of 1) length of input or 2) num-attacks (for single-input)
If output-length is a fixnum, it designates the length of the output list.

Repeat - What to do when buffer is called again.  When the buffer is filled to list-length, it outputs it's data.
If repeat is not set to 0, then buffer will output the same data again.  A fixnum dictates how many times to
repeat the data (repeat of 1 = 2x).  A float makes buffer decide what percentage of num-attacks to
repeat for.")

(defvar *buffer-prompt*
    "(buffer buffer-ref function :output 'single :storage-length 5 :repeat 0)") 



