;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;September 2001
;;;thesis project
;;;instrument-definitions.lisp

(in-package :cl-user)


(def-CLASS (instrument
            (:documentation
"An instrument stores data about it's capabilities and output :channel (midi#, csound-instrument, etc.)."))
    ()
  (range-hi '13.07)
  (range-lo '3.00)
  (max-volume '127)
  (mute nil) ;;if nil, don't perform
  (max-duration 100000000)
  (pitch-template nil)
  (polyphony 1) ;;a value of nil means unlimited
  (channel '1 :type fixnum :allocation :instance))

(defvar default-instrument (make-instance 'instrument))

(def-class csound-instrument
    (instrument)
  (range-hi '16.00)
  (range-lo '1.00)
  (max-volume 127) ;;this is converted later...can be floats
  (mute nil)
  (max-duration 100000000)
  (pitch-template nil)
  (polyphony nil)
  (channel 1) ;;csound instruments are also saved as midi
  )

(defvar default-csound-instrument (make-instance 'csound-instrument))

(def-class (musical-instrument 
            (:documentation
"Musical instruments are for performance by a human.")) 
    (instrument) 
  family
  (max-volume '127))

(def-class (bowed-string
            (:documentation
"The bowed-string family includes violin, viola, cello and bass."))
    (musical-instrument)
  (max-duration '60000)
  (family 'bowed-string))

(def-class violin (bowed-string)
  (range-lo '6.07)
  (range-hi '10.07))

(defvar violin (make-instance 'violin))

(def-class viola (bowed-string)
  (range-lo '6.00)
  (range-hi '9.04))

(defvar viola (make-instance 'viola))

(def-class cello (bowed-string)
  (range-lo '5.00)
  (range-hi '8.07))

(defvar cello (make-instance 'cello))

(def-class contra-bass (bowed-string)
  (range-lo '4.04)
  (range-hi '7.00))

(defvar contra-bass (make-instance 'contra-bass))

(def-class wind (musical-instrument)
  (max-duration '20000)
  (family 'wind))

(def-class flute (wind)
  (range-lo '6.11)
  (range-hi '10.02))

(defvar flute (make-instance 'flute))

(def-class piccolo (wind)
  (range-lo '8.07)
  (range-hi '11.00))

(setq piccolo (make-instance 'piccolo))

(def-class clarinet (wind)
  (range-lo '6.04)
  (range-hi '9.07)) 

(defvar clarinet (make-instance 'clarinet))

(def-class percussion (musical-instrument)
  (family 'percussion)
  (channel '10))

(def-class non-pitched-percussion (percussion)
  )

(defvar non-pitched-percussion (make-instance 'non-pitched-percussion))

(def-class pitched-percussion (percussion)
  )

(defvar pitched-percussion (make-instance 'pitched-percussion))

(def-class marimba (pitched-percussion)
  (range-lo '6.09)
  (range-hi '11.00))

(defvar marimba (make-instance 'marimba))

(def-class vibraphone (pitched-percussion)
  (range-lo '7.05)
  (range-hi '10.05))

(setq vibraphone (make-instance 'vibraphone))

(def-class crotales (non-pitched-percussion)
  (range-lo '9.0)
  (range-hi '10.0))

(def-class keyboard (musical-instrument)
  )

(def-class piano (keyboard)
  (channel 1)
  (range-lo '4.09)
  (range-hi '12.00))

(defvar piano (make-instance 'piano))
  




