;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;instrument-macros.lisp
;;;nov 2001instrument-macros.lisp
;;;textures thesis project

;;TO DO:  Fix modular-riff!  It doesn't give the appropriate number of beats when creating individual
;;sectionsk, i.e. if one section has 7 attacks, and the other has 1, ramps on the second will only get
;;1/7th of the way through...they'll think they're going 8* as many steps as they are...

(in-package :cl-user)

#|
These instrument macros allow one to design an instrument without detail to it's parameters,
requiring a simple call to the macro instead of copying and pasting the file.  In other words,
if you make an instrument and want to a) link parameters together or b) make different versions
of an instrument with only slight parameter changes then write the instrument into a macro.

I intend to start building a library of these macros, so that you can simply call an instrument-macro
most of the time instead of having to design one from scratch...much better user interface for the 
unknowledgeable in lisp...Great stuff!
|#

;;NOTE:  Do not add a Quote (') to the beginning of lists on the outer level, only use quotes if you
;;make a function call that requires them, i.e.

;(chance-transpose-random :pc-set (0 11 1 6 5)) ;;no quote on list
;;--> (chance-transpose-random :pc-set (sequencer '(0 11 1 6 5))  ;;quote needed for function call


;;insert any new macro here..
(defparameter *macro-functions* '(m-ramp m-shrink-list m-transpose m-chancew m-split m-truncate m-variation
                                         m-multiply m-rand-reorder m-sequencer
                                         chancew-split variation-split split-variation reorder-sequence))



;;;note to self!  Time to start making smaller macro's that can be used to construct these larger macros,
;;;or even as input to them!  in other words, the max-r, min-r, chance-r, erratic-r could be a macro...


;;;;;;;;;;;;;;;;;;;;;
;;;MACRO UTILITIES;;;
;;;;;;;;;;;;;;;;;;;;;

;;if the data is a plain list, it is put into a multi-ramp and returned, otherwise it is simply returned
;;everything must be quoted now, which is good for consistency, because these aren't defined as macros...



(defun m-ramp (ramp-list &optional quantize &rest quantize-quantize)
  (when quantize (setq quantize (m-ramp quantize (first quantize-quantize)(rest quantize-quantize))))
  (let* ((ramp-list (get-macro ramp-list))
         (ramp-list (if (symbolp ramp-list)
                        (eval ramp-list)
                      ramp-list))
         (buffer-ref (buffer-ref)))
    (cond ((null ramp-list) nil)
          ((atom ramp-list)
           (if quantize
               `'(quantize-to-n ,quantize ',ramp-list)
             `',ramp-list))
          ((eql (first ramp-list) 'quote)
           ramp-list)
          ((available-function (first ramp-list))
           (quote-check ramp-list))
          (t 
           (let* ((end-item (end ramp-list))
                  (type (cond ((member end-item *ramp-types*)                                
                               (setq ramp-list (butlast ramp-list))
                               end-item)
                              ((and (listp end-item)(member (first end-item) *ramp-types*))
                               (when (= (length end-item) 1) 
                                 (setq ramp-list (butlast ramp-list)))
                               (first end-item))
                              (t 'linear)))
                  (quantize (cond (quantize quantize)
                                  ((listp end-item) 
                                   (setq ramp-list (butlast ramp-list))
                                   (if (and (= (length end-item) 1) (numberp (first end-item)))
                                       (first end-item)
                                     (second end-item))))))
             (cond ((member (first ramp-list) (available-functions))
                    `',ramp-list)
                   ((= (length ramp-list) 1)
                    `',@ramp-list)
                   ((= (length ramp-list) 2)
                    `'(ramp ,@ramp-list :type ',type :quantize ,quantize))
                   (t `'(multi-ramp ,buffer-ref ',ramp-list :type ',type :quantize ,quantize))))))))
          

(defun m-rotate (list &optional rotate)
  (let ((list (quote-check (get-macro list)))
        (rotate (m-ramp rotate)))
    (if rotate 
        `'(rotate-list ,list ,rotate)
      list)))

(defun m-repeat (list amount)
  (let* ((list (quote-check (get-macro list)))
         (amount (m-ramp amount))
         (unamount (unquote amount)))
    (cond ((null unamount)
           list)
          ((or (and (numberp unamount)
                    (> (unquote amount) 1))
               (listp (unquote amount)))
        `'(repeat-items ,list ,amount))
          (t list))))

(defun m-morph (list1 list2 &optional pch)
  (let ((list1 (quote-check (get-macro list1)))
        (list2 (quote-check (get-macro list2)))
        (pch (get-data pch)))
    (cond ((null list1)
           list2)
          ((null list2)
           list1)
          (t `'(morph-list ,list1 ,list2 ,pch)))))
      

;;depending on given inputs, returns nth-n, first-n or nthcdr on a possibly rotated repeated list
;;with all parameters m-ramp'd.  Your useful all-in-one list modifier.

;;;(m-list-alter '(1 2 3 4 5) :n 1 :num-items '3 :rotate 2 :repeat '2)
;;;--->'(REPEAT-ITEMS '(NTH-N '1 '3 '(ROTATE-LIST '(1 2 3 4 5) '2)) '2)
;;;--->(4 4 5 5 1 1)
(defun m-list-alter (list &key n num-items rotate repeat)
  (let* ((n (m-ramp n))
         (num-items (m-ramp num-items))
         (rotate (m-ramp rotate))
         (list (m-rotate list rotate)))
    (cond ((null list) nil)
          ((null n) 
           (if (null num-items)
               (m-repeat list repeat)
             (m-repeat `'(first-n ,num-items ,list) repeat)))
          ((null num-items)
           (m-repeat `'(nth-rest ,n ,list) repeat))
          (t
           (m-repeat `'(nth-n ,n ,num-items ,list) repeat)))))

;;use this if you need to ensure that there is a quote in front of an object without making 2 quotes.
(defun quote-check (list)
  (cond ((null list) nil)
        ((not (listp list))
         `',list)
        ((eql (first list) 'quote)
         (quote-check (second list)))
        (t
         `',list)))

(defun m-quantize (list &optional (quantize 1))
  (when list
    (let ((list (get-macro list))
          (quantize (m-ramp quantize)))
    `'(quantize-to-n ,quantize ,list))))
      

(defun m-transpose (list &optional transpose)
  (when list
    (let* ((transpose (if (numberp transpose) transpose
                        (m-ramp transpose)))
           (list (get-macro list))
           (list (quote-check list)))
      (cond ((and (numberp transpose)(zerop transpose))
             list)
            (transpose
             `'(transpose ,list ,transpose))
            (t list)))))

(defun m-chancew (&key (max 1)(min 0)(chance .5)(window-curve 'linear))
  (let ((max (m-ramp max))
        (min (m-ramp min))
        (chance (m-ramp chance)))
    (if (and (numberp (unquote chance))(>= (unquote chance) 1)) 
        max
    `'(chancew :max ,max :min ,min :chance ,chance :window-curve ',window-curve))))

(defun m-split (list &key (length nil)(include '()))
  (let ((list (get-macro list))
        (new-list (quote-check list))
        (length (m-ramp length))
        (include (quote-check include)))    
    (if (null length)
        new-list
      `'(split-list ,new-list :length ,length :include ,include))))

(defun m-truncate (list beats)
  (let* ((list (get-macro list))
         (list (quote-check list))
         (beats (quote-check (m-ramp beats))))
    (cond ((null list) nil)
          ((null beats) list)
          (t
           `'(truncate-rhythm ,list ,beats)))))
  
(defun m-variation (&optional (list '(1)) &key (type 'num) max-vary min-vary (chance .5) (items-to-vary 't))
  (let* ((list (get-macro list))
         (list (quote-check list))
         (max-vary (if (numberp max-vary) (round max-vary) (m-ramp max-vary)))
         (min-vary (if (numberp min-vary) (round min-vary) (m-ramp min-vary)))
         (min-vary (if (null min-vary) 0 min-vary))
         (chance (m-ramp chance))
         (items-to-vary (if (and (listp items-to-vary)
                                 (fboundp (first items-to-vary)))
                            (quote-check (eval (add-to-end items-to-vary list)))
                         (quote-check items-to-vary))))
    (cond ((and (numberp max-vary)(zerop max-vary) list))
          ((null max-vary) list) 
          ((null list) nil)
          (t
           `'(variation ,list :type ',type :max-vary ,max-vary :min-vary ,min-vary 
                        :chance ,chance :items-to-vary ,items-to-vary)))))

(defun m-list> (compare list)
  (let ((compare (m-ramp compare))
        (list (quote-check list)))
    (if (null compare) list
      `'(list> ,compare ,list))))

(defun m-list>= (compare list)
  (let ((compare (m-ramp compare))
        (list (quote-check list)))
    (if (null compare) list
      `'(list>= ,compare ,list))))

(defun m-list< (compare list)
  (let ((compare (m-ramp compare))
        (list (quote-check list)))
    (if (null compare) list
      `'(list< ,compare ,list))))

(defun m-list<= (compare list)
  (let ((compare (m-ramp compare))
        (list (quote-check list)))
    (if (null compare) list
      `'(list<= ,compare ,list))))
    
(defun m-multiply (list &optional (multiplier 1))
  (let* ((multiplier (if (numberp multiplier) multiplier
                       (m-ramp multiplier)))
         (list (quote-check (get-macro list))))
    (cond ((null list) nil)
          ((or (null multiplier) 
               (eq 1 multiplier))
           list)
          ((or (and (numberp multiplier)(zerop multiplier)) (and (numberp list) (zerop list))) 
           (format t "Warning: attempting to multiply by zero!%"))
          ((available-function multiplier)  ;;if multiplier is a function, leave it alone
           `'(times ,list ,multiplier))   
          (t                                
           (if (or (available-function list)     ;;if multiplier is a number and list is not special
                   (symbolp (unquote list)))     ;;multiply now and speed things up
               `'(times ,list ,multiplier)
             (times list multiplier))))))

(defun m-rand-reorder (list &key include-front include-back)
  (if (null list)
      nil
    (let* ((list (get-macro list))
           (include-front (quote-check (get-macro include-front)))
           (include-back (quote-check (get-macro include-back)))
           (quote-list (quote-check list)))           
      (cond ((or (not (null include-front)) (not (null include-back)))
             `'(random-reorder ,quote-list :include-front ,include-front :include-back ,include-back))
            ((and (listp list) (<= (length list)))
             quote-list)
            (t `'(random-reorder ,quote-list))))))
        

(defun m-sequencer (list &rest args)
  (let* ((args (quote-args (get-macro (clean-up-list args))))
         (list (get-macro list))
         (quote-list (quote-check list))
         (buffer-ref (buffer-ref)))
    (cond ((null list) nil)
          ((not (listp list))
           (if args
               `'(sequencer ,buffer-ref ,quote-list ,@args)
             `'(sequencer ,buffer-ref ,quote-list)))
          ((<= (length list) 1)
           list)
          ((null args)
           `'(sequencer ,buffer-ref ,quote-list))
          (t
           `'(sequencer ,buffer-ref ,quote-list ,@args)))))

(defun m-riff (n list)
  (let ((quote-list (quote-check (get-macro list)))
        (n (quote-check (m-ramp n))))
    (if (null n)
        quote-list
      `'(riff ,n ,quote-list))))

(defun m-buffer (function &rest args)
  (let* ((args (quote-args (get-macro (clean-up-list args))))
         (function (get-macro function))
         (quote-function (quote-check function))
         (buffer-ref (buffer-ref)))
    (cond ((null function) nil)
          ((null args)
           `'(buffer ,buffer-ref ,quote-function))
          (t
           `'(buffer ,buffer-ref ,quote-function ,@args)))))

(defmacro m-list (&rest args)
  (let ((args (quote-check (mapcar #'unquote (mapcar #'get-macro args)))))
    `',args))

;;this is for dealing with the :extra keyword.
(defun m-extra (list)
  (let ((data (mapcar (lambda (x) (unquote (get-macro x))) list)))
    `',data))

(defun m-check-list (list)
  (let ((list (unquote list)))
    (if (listp list)
        `',list
      (m-check-list (list list)))))
    

(defun unquote (list)
  (if (or (null list) (atom list)) list
    (let ((item (first list)))
      (if (eql item 'quote)
          (unquote (second list))
        list))))

(defun quote-args (args)
  (if (or (null args) (atom args)) args
    (let ((item (first args)))
      (if (keywordp item)
          (cons item (quote-args (rest args)))
        (cons (quote-check item) (quote-args (rest args)))))))

#|
certain functions *ref-functions*, need to have a unique key to the buffer hash table.   The current
key is stored in *buffer-ref*, and should be incremented each time it is used.  These functions below
check a list for function calls to *ref-functions* functions, and if present add a key.

(ref-check '(sequencer '((sequencer '(1 2 3 4))(multi-ramp '(1 10) :curve 'expt))))
------->(SEQUENCER 31 '((SEQUENCER 32 '(1 2 3 4)) (MULTI-RAMP 33 '(1 10) :CURVE 'EXPT)))


I don't think I need this function anymore...always use macros for making instruments!!!

(defun ref-check (list &optional quoted)
  (setq *utility-error* 'ref-check)
  (cond ((null list) nil)
        ((not (listp list)) list)
        ((eql (first list) 'quote)
         (quote-check (ref-check (second list) t)))
        ((member (first list) *macro-functions*)
         (second (eval list))) ;; we don't want the quotes created by macros here....
        (t
         (let* ((function (pop list))
                (function (if (listp function) (list (ref-function function quoted))
                            (ref-function (list function) quoted)))
                (args
                 (loop for n from 0 below (length list) collect
                       (let ((arg (nth n list)))
                         (ref-check arg)))))
           (append function args)))))

(defun ref-function (list &optional quoted)
  (setq *utility-error* 'ref-function)
  (if (listp list)
      (cond ((eql (first list) 'quote)
             nil)
            ((and (> (length list) 1)
                  (member (first list) *macro-functions*))
             (second (eval list)))  ;;don't want quotes
            (t (let ((function (pop list)))
                 (if (member function *ref-functions*)
                     (progn (push *buffer-ref* list)
                       (incf *buffer-ref*)
                       (cons function list))
                   (cons function list))))
            list)))
        
|#
  

;;;;;;;;;;;;;;;
;;;UTILITIES;;;
;;;;;;;;;;;;;;;

;;given a sequence-list and a a list of num attacks for each item of sequence, gives a total num attacks.
;;i.e., (number-of-attacks '(1 1 2 2 (3) 3) '(5 2 1))
;;---------->16  ;;5 * 2 (1 1) + 2 * 2 (2 2) + 1 (3) + 1 3 
(defun number-of-attacks (sequence attacks-list)
  (let ((sequence (get-data sequence))
        (attacks-list (get-data attacks-list)))
  (loop for n from 0 below (length sequence) sum
        (let ((item (nth n sequence)))
          (if (listp item)
              (if (keywordp (first item))
                  0
                1) 
            (nth (1- item) attacks-list))))))



;;;;;;;;;;;;;;;;;;;
;;;SIMPLE-PHRASE;;;
;;;;;;;;;;;;;;;;;;;

(defun simple-phrase 
    (&key (melody '(5.0))(n-p nil)(num-items-p nil)(transpose nil)(rotate-p nil)(repeat-p nil)
          (rhythm '(4))(n-r nil)(num-items-r nil)(multiply-r nil)(truncate nil)
          (durations '(4))(n-d nil)(num-items-d nil)(multiply-d nil)
          (dynamics 127)(monophonic T)(delay nil)
          (extra nil)(instrument-nums nil))
  "This instrument is for basic entry of data.  Melody, rhythm and durations all take a list
and can modify this list with nth-n, transpose for pitch, and multiply for rhythm and durations.
Rhythm can also be truncated, and dynamics is unstructured."
  
 (initialize 'macros)
  (let ((delay (get-data delay))
        (rhythm-call 
         (m-truncate 
          (m-multiply 
           (m-list-alter rhythm :n n-r :num-items num-items-r) multiply-r)
          truncate))
        (pitch-call (m-list-alter (m-transpose melody transpose) :n n-p :num-items num-items-p
                                  :rotate rotate-p :repeat repeat-p))                     
        (duration-call 
         (m-multiply 
          (m-list-alter durations :n n-d :num-items num-items-d) multiply-d))
        (dynamics-call (m-check-list (m-ramp dynamics)))
        (extra (when extra (m-extra extra)))
        (instrument-nums (m-check-list (get-macro instrument-nums))))
    `(modular-phrase 
      :DELAY ,delay
      :MONOPHONIC ,monophonic
      :RHYTHM ,rhythm-call
      :PITCH ,pitch-call
      :DURATION ,duration-call
      :DYNAMICS ,dynamics-call
      :EXTRA ,extra
      :INSTRUMENT-NUMS ,instrument-nums)))






;;;;;;;;;;;;;;;;;;;
;;;CHANCEw-SPLIT;;;
;;;;;;;;;;;;;;;;;;;

(defun chancew-split 
    (&key (melody '(5.0))(n-p nil)(num-items-p nil)(rotate-p nil)(repeat-p nil)
          (length nil)(include '())(transpose nil)                      
          (max-r 1)(min-r 0)(window-curve-r 'linear)(chance-r .5)
          (multiply-r nil)(truncate nil)   
          (max-d 4)(min-d 0)(window-curve-d 'linear)(chance-d .5)       
          (dynamics 127)(monophonic T)(delay nil))
  "This instrument uses the chance function for both rhythm and duration and a 
transposed split-list for pitch.  All single parameters are multi-ramp'd."  
  
 (initialize 'macros)
  (let ((delay (get-data delay))
        (rhythm-call (m-truncate 
                      (m-multiply
                       (m-chancew :max max-r :min min-r :chance chance-r 
                                  :window-curve window-curve-r)
                       multiply-r)
                      truncate))
        (pitch-call (m-split 
                     (m-list-alter (m-transpose melody transpose) :n n-p :num-items num-items-p
                                   :rotate rotate-p :repeat repeat-p)
                     :length length :include (m-transpose include transpose)))
        (duration-call (m-chancew :max max-d :min min-d :chance chance-d 
                                  :window-curve window-curve-d))
        (dynamics-call (m-ramp dynamics)))
    `(modular-phrase 
      :DELAY ,delay
      :MONOPHONIC ,monophonic
      :RHYTHM ,rhythm-call
      :PITCH ,pitch-call
      :DURATION ,duration-call
      :DYNAMICS ,dynamics-call)))


;;;;;;;;;;;;;;;;;;;;;
;;;VARIATION-SPLIT;;;
;;;;;;;;;;;;;;;;;;;;;

(defun variation-split 
    (&key (melody '(7.0 7.01 7.03 7.04))(n-p nil)(num-items-p nil)  
          (length nil)(include '())(transpose nil)(rotate-p nil)(repeat-p nil)
          (rhythm '(1))(multiplier nil)(n-r nil)(num-items-r nil)
          (max-r nil)(min-r nil)(chance-r .5)(items-r t)(rotate-r nil)(repeat-r nil)
          (max-d 4)(min-d 0)(window-curve-d 'linear)(chance-d .5)   
          (dynamics 127)(monophonic T)(delay nil)(truncate nil))
  "This instrument uses the variation function for rhythm, chancew function duration and 
split-list for pitch."
  
  (initialize 'macros)
  (let ((delay (get-data delay))
        (rhythm-call
         (m-quantize
          (m-truncate 
           (m-variation 
            (m-multiply 
             (m-list-alter rhythm :n n-r :num-items num-items-r 
                           :rotate rotate-r :repeat repeat-r) multiplier)
            :max-vary max-r :min-vary min-r :chance chance-r 
            :items-to-vary items-r :type 'num)
           truncate)))
        (pitch-call 
         (m-split 
          (m-list-alter (m-transpose melody transpose) :n n-p :num-items num-items-p
                        :rotate rotate-p :repeat repeat-p)
          :length length :include (m-transpose include transpose)))
        (duration-call (m-quantize
                        (m-chancew :max max-d :min min-d :chance chance-d 
                                   :window-curve window-curve-d)))
        (dynamics-call (m-ramp dynamics 1)))
    `(modular-phrase 
      :DELAY ,delay
      :MONOPHONIC ,monophonic
      :RHYTHM ,rhythm-call
      :PITCH ,pitch-call
      :DURATION ,duration-call
      :DYNAMICS ,dynamics-call)))


;;;;;;;;;;;;;;;;;;;;;
;;;SPLIT-VARIATION;;;
;;;;;;;;;;;;;;;;;;;;;

(defun split-variation 
    (&key (melody '(7.0 7.01 7.03 7.04))(n-p nil)(num-items-p nil)(transpose nil)                          
          (repeat-p nil)(rotate-p nil)(max-p nil)(min-p nil)(chance-p .5)(items-p t)
          (rhythm '(1))(multiplier nil)(n-r nil)(num-items-r nil)
          (length nil)(include '())
          (max-d 4)(min-d 0)(window-curve-d 'linear)(chance-d .5)          
          (dynamics 127)(monophonic T)(delay nil)(truncate nil))
  "This instrument uses the split-list function for rhythm, chancew function duration and 
variation for pitch."
  
  (initialize 'macros)
  (let ((rhythm-call 
         (m-truncate 
          (m-split
           (m-multiply 
            (m-list-alter rhythm :n n-r :num-items num-items-r) multiplier)
           :length length :include include)
          truncate))    
        (pitch-call (m-variation (m-list-alter (m-transpose melody transpose) :n n-p :num-items num-items-p
                                               :rotate rotate-p :repeat repeat-p)
                                 :max-vary max-p :min-vary min-p :chance chance-p 
                                 :items-to-vary (m-transpose items-p transpose) :type 'pch))
        (duration-call (m-chancew :max max-d :min min-d :chance chance-d 
                                  :window-curve window-curve-d))
        (dynamics-call (m-ramp dynamics)))
    `(modular-phrase 
      :DELAY ,delay
      :MONOPHONIC ,monophonic
      :RHYTHM ,rhythm-call
      :PITCH ,pitch-call
      :DURATION ,duration-call
      :DYNAMICS ,dynamics-call)))


;;;;;;;;;;;;;;;;;;;;;;
;;;REORDER-SEQUENCE;;;
;;;;;;;;;;;;;;;;;;;;;;

(defun reorder-sequencer 
    (&key (melody '(7.0 7.01 7.03 7.04))(n-p nil)(num-items-p nil)
          (transpose nil)(repeat-p nil)(rotate-p nil)
          (rhythm '(1))(multiplier nil)(n-r nil)(num-items-r nil)(rotate-r 0)
          (output 'list)(storage-length '1)
          (max-d 4)(min-d 0)(window-curve-d 'linear)(chance-d .5)          
          (dynamics 127)(monophonic T)(delay nil)(truncate nil))
  "This instrument uses a randomly reordered buffered list for rhythm, chancew function duration and, 
a shrunk list for pitch (i.e. user input)."
  
  (initialize 'macros)
  (let ((rhythm-call 
         (m-truncate 
          (m-rand-reorder
           (m-buffer
            (m-multiply 
             (m-list-alter  rhythm :n n-r :num-items num-items-r :rotate rotate-r) multiplier)
            :output output :storage-length (m-ramp storage-length)))
          truncate))
        (pitch-call (m-list-alter (m-transpose melody transpose) :n n-p :num-items num-items-p
                                  :rotate rotate-p :repeat repeat-p))
        (duration-call (m-chancew :max max-d :min min-d :chance chance-d 
                                  :window-curve window-curve-d))
        (dynamics-call (m-ramp dynamics)))
    `(modular-phrase 
      :DELAY ,delay
      :MONOPHONIC ,monophonic
      :RHYTHM ,rhythm-call
      :PITCH ,pitch-call
      :DURATION ,duration-call
      :DYNAMICS ,dynamics-call)))


#|Modular-riff takes any number of modular-phrases and turns them into a new modular-phrase that 
can sequence between the different phrases by way of the riff function.  Simply call modular-riff 
with a SEQUENCE where 1 is the first phrase and (1) is a 1 beat rest, sequence-args, a list of args for
the call to m-sequencer, a list of ATTACKS-PER-RIFF (for a 5 note 'riff', you don't have to have that 
sequence space referenced 5 times...you also can't do less...) a global delay (the delays within the 
individual phrases are ignored), :pre-eval and :post-eval (for other stuff) and finally a list of 
MODULAR-PHRASEs, which are the actual phrases to be sequenced.

(modular-riff :sequence-list '(1 2 3 1 1 (10) 1) :attacks-per-riff '(1 1 1) :delay 5  
              (list (chancew-split :foo 'baz)
                    (variation-split :bar 'blat)
                    (split-variation :baz 'blorp)))
|#

(defun modular-riff (&key (sequence '(1)) sequence-args (attacks-per-riff '(1)) (delay 0) (monophonic t) 
                          reset-location-list pre-eval post-eval phrases symbolname)
  (setq *function-error* 'modular-riff)
  (initialize 'macros)
  (let* ((num-phrases (length phrases))
         (attacks-per-riff (check-and-modify-length attacks-per-riff num-phrases))
         (total-attacks (when symbolname (number-of-attacks sequence attacks-per-riff)))
         (attack-var (when symbolname (concat-symbol (princ-to-string symbolname) '-attacks)))
         (reset-location-list (check-and-modify-length reset-location-list num-phrases))
         (phrases (mapcar #'(lambda (x) (if (symbolp x) (eval x) x)) phrases))
         (rhythm-call (m-riff (m-sequencer sequence sequence-args)
                         (mapcar #'(lambda (phrase output-length reset-location)
                                     (unquote (m-buffer 
                                               (find-keyword ':rhythm phrase)
                                               :output 'list :storage-length '1.0 :output-length output-length
                                               :reset-location reset-location)))
                           phrases attacks-per-riff reset-location-list)))
         (pitch-call (m-riff (m-sequencer sequence sequence-args)
                        (mapcar #'(lambda (phrase output-length reset-location)
                                    (unquote (m-buffer 
                                              (find-keyword ':pitch phrase)
                                              :output 'list :storage-length '1.0 :output-length output-length
                                              :reset-location reset-location)))
                          phrases attacks-per-riff reset-location-list)))
         (duration-call (m-riff (m-sequencer sequence sequence-args)
                           (mapcar #'(lambda (phrase output-length reset-location) 
                                       (unquote (m-buffer
                                                 (find-keyword ':duration phrase)
                                                 :output 'list :storage-length '1.0 :output-length output-length
                                                 :reset-location reset-location)))
                             phrases attacks-per-riff reset-location-list)))
         (dynamics-call (m-riff (m-sequencer sequence sequence-args)
                           (mapcar #'(lambda (phrase output-length reset-location) 
                                       (unquote (m-buffer
                                                 (find-keyword ':dynamics phrase)
                                                 :output 'list :storage-length '1.0 :output-length output-length
                                                 :reset-location reset-location)))
                             phrases attacks-per-riff reset-location-list)))
         (pre-eval (quote-check pre-eval))
         (post-eval (quote-check post-eval)))
    (when symbolname (set attack-var total-attacks))
    `(modular-phrase :delay ,delay :monophonic ,monophonic :pre-eval ,pre-eval :post-eval ,post-eval
                     :rhythm ,rhythm-call
                     :pitch ,pitch-call
                     :duration ,duration-call
                     :dynamics ,dynamics-call)))



;;;these are the old world macros....yech!




(defmacro chance-random 
    (&key (pc-set '(0 1 3 4)) (range-lo '(6.04)) (range-hi '(9.07))(transpose '(0)) ;pitch
          (max-r '(1))(min-r '(0))(more-erratic-r '(1))(chance-r '(.5)) ;rhythm
          (max-d '(4))(min-d '(0))(more-erratic-d '(1))(chance-d '(.5)) ;duration
          (dynamics '(127)))
  "This instrument uses the chance function for both rhythm and duration and a 
transposed random-from-pc for pitch.  All single parameters are multi-ramp'd (i.e. not pc-set)."  
    
    `'(modular-phrase 
       :RHYTHM '(segment 
                 '(chance :max (multi-ramp '(,@max-r)) :min (multi-ramp '(,@min-r))
                          :more-erratic (multi-ramp '(,@more-erratic-r))
                          :chance (multi-ramp '(,@chance-r))))
       :PITCH '(segment 
                '(transpose
                  '(random-from-pc '(,@pc-set) :range-lo (multi-ramp '(,@range-lo))
                                   :range-hi (multi-ramp '(,@range-hi)))
                  (multi-ramp '(,@transpose))))
       :DURATION '(segment
                   '(chance :max (multi-ramp '(,@max-d)) :min (multi-ramp '(,@min-d))
                            :more-erratic (multi-ramp '(,@more-erratic-d))
                            :chance (multi-ramp '(,@chance-d))))
       :DYNAMICS ',dynamics))


(defmacro chance-split 
    (&key (melody '(7.0 7.01 7.03 7.04))(n nil)(length '(3))(include '())(transpose '(0)) ;pitch
          (max-r '(1))(min-r '(0))(more-erratic-r '(0))(chance-r '(.5)) ;rhythm
          (max-d '(4))(min-d '(0))(more-erratic-d '(0))(chance-d '(.5)) ;duration
          (dynamics '(127)))
  "This instrument uses the chance function for both rhythm and duration and a 
transposed split-list for pitch.  All single parameters are multi-ramp'd."  

  `'(modular-phrase 
     :RHYTHM '(segment 
               '(chance :max (multi-ramp '(,@max-r)) :min (multi-ramp '(,@min-r))
                        :more-erratic (multi-ramp '(,@more-erratic-r))
                        :chance (multi-ramp '(,@chance-r))))
     :PITCH '(segment 
              '(transpose 
                '(split-list '(first-n (multi-ramp '(,@n)) '(,@melody))
                             :length (multi-ramp '(,@length))
                             :include ,include)
                (multi-ramp '(,@transpose))))
     :DURATION '(segment
                 '(chance :max (multi-ramp '(,@max-d)) :min (multi-ramp '(,@min-d))
                          :more-erratic (multi-ramp '(,@more-erratic-d))
                          :chance (multi-ramp '(,@chance-d))))
     :DYNAMICS ',dynamics))


(defmacro chance-pickone 
    (&key (melody '(7.0 7.01 7.03 7.04))(n '(4))(transpose '(0)) ;pitch
          (max-r '(1))(min-r '(0))(more-erratic-r '(1))(chance-r '(.5)) ;rhythm
          (max-d '(4))(min-d '(0))(more-erratic-d '(1))(chance-d '(.5)) ;duration
          (dynamics '(127)))
  "This instrument uses the chance function for both rhythm and duration and a 
transposed split-list for pitch.  All single parameters are multi-ramp'd."  

  `'(modular-phrase 
     :RHYTHM '(segment 
               '(chance :max (multi-ramp '(,@max-r)) :min (multi-ramp '(,@min-r))
                        :more-erratic (multi-ramp '(,@more-erratic-r))
                        :chance (multi-ramp '(,@chance-r))))
     :PITCH '(segment 
              '(transpose 
                '(pick-one '(first-n (multi-ramp '(,@n)) '(,@melody)))
                (multi-ramp '(,@transpose))))
     :DURATION '(segment
                 '(chance :max (multi-ramp '(,@max-d)) :min (multi-ramp '(,@min-d))
                          :more-erratic (multi-ramp '(,@more-erratic-d))
                          :chance (multi-ramp '(,@chance-d))))
     :DYNAMICS ',dynamics))

(defmacro variation-variation
    (&key (input-r '(1 1 1 1))(chance-r '(.2))(min-r '(-1))(max-r '(1))(vary-r t)
          (first-n nil)(last-n nil)(melody '(6.01 6.03 6.04)) (chance-p '(.2))
          (max-p '(2))(min-p '(-2))(vary-p 't)(transpose '(0))
          (max-d '(4))(min-d '(0))(chance-d '(.3))(erratic-d '(0)))
  "This instrument uses the VARIATION function for both rhythm and pitch, and the chance function
for duration.  Most parameters are multi-ramp'd, the list for pitch goes through LAST-N then FIRST-N (in 
that order) before being inputed to the variation module."
  
  `'(modular-phrase
     :RHYTHM '(segment 
               '(variation '(,@input-r) 
                  :type 'num :chance '(multi-ramp '(,@chance-r))
                  :min-vary '(multi-ramp '(,@min-r))
                  :max-vary '(multi-ramp '(,@max-r))
                  :items-to-vary ,vary-r))
     :PITCH '(segment
              '(variation '(transpose (first-n (multi-ramp '(,@first-n)) (last-n (multi-ramp '(,@last-n)) 
                                                                           '(,@melody)))
                             '(multi-ramp '(,@transpose)))
                 :type 'pch :chance '(multi-ramp '(,@chance-p))
                 :min-vary '(multi-ramp '(,@min-p))
                 :max-vary '(multi-ramp '(,@max-p))
                 :items-to-vary ,vary-p))
     :DURATION '(segment
                 '(chance :max '(multi-ramp '(,@max-d)) :min '(multi-ramp '(,@min-d))
                   :chance '(multi-ramp '(,@chance-d))
                   :more-erratic '(multi-ramp '(,@erratic-d))))))

(defmacro sequencer-variation-variation
    (&key (type 'cycle)(cycles '1)(motion 'forward)(reset-location nil)(start-pos 0)
      (input-r1 '(1 1 1 1))(truncate1 '(8))(chance-r1 '(.2))(min-r1 '(-1))(max-r1 '(1))(vary-r1 t)
      (input-r2 '(1 1 1 1))(truncate2 '(8))(chance-r2 '(.2))(min-r2 '(-1))(max-r2 '(1))(vary-r2 t)
      (melody '(8.0))(last-n nil)(first-n nil)(transpose 0)
      (chance-p '(.2))(max-p '(2))(min-p '(-2))(vary-p 't) 
      (max-d1 '(4))(min-d1 '(0))(chance-d1 '(.3))(erratic-d1 '(0))
      (max-d2 '(4))(min-d2 '(0))(chance-d2 '(.3))(erratic-d2 '(0)))
  
  "This macro uses VARIATION for rhythm, SPLIT-LIST for melody and CHANCE for duration.  The rhyhthm and
duration are step-sequenced between two sets, with each parameter multi-ramp'd within.  Reset-location
defaults to NIL for the sequencer (so the ramps are for the entire phrase, but there will be gaps when
the other part is playing...).
The melody is transposed and run through LAST-N then FIRST-N.  All single parameters multi-ramp'd."
  
  `'(modular-phrase
     :RHYTHM '(segment 
               '(sequencer '((truncate-rhythm
                              '(variation '(,@input-r1) 
                                 :type 'num :chance '(multi-ramp '(,@chance-r1))
                                 :min-vary '(multi-ramp '(,@min-r1) :quantize 1)
                                 :max-vary '(multi-ramp '(,@max-r1) :quantize 1)
                                 :items-to-vary ,vary-r1)
                              '(multi-ramp '(,@truncate1)))
                             (truncate-rhythm
                              '(variation '(,@input-r2) 
                                 :type 'num :chance '(multi-ramp '(,@chance-r2))
                                 :min-vary '(multi-ramp '(,@min-r2) :quantize 1)
                                 :max-vary '(multi-ramp '(,@max-r2) :quantize 1)
                                 :items-to-vary ,vary-r2)
                              '(multi-ramp '(,@truncate2))))
                  :type ',type :cycles ,cycles :motion ',motion :reset-location ,reset-location
                  :start-pos ',start-pos))
     :PITCH '(segment
              '(variation '(transpose (first-n (multi-ramp '(,@first-n)) (last-n (multi-ramp '(,@last-n)) 
                                                                           '(,@melody)))
                                       '(multi-ramp '(,@transpose)))
                 :type 'pch
                 :chance '(multi-ramp '(,@chance-p))
                 :min-vary '(multi-ramp '(,@min-p))
                 :max-vary '(multi-ramp '(,@max-p))
                 :items-to-vary ,vary-p))
     :DURATION '(segment
                 '(sequencer '((chance :max '(multi-ramp '(,@max-d1)) :min '(multi-ramp '(,@min-d1))
                                 :chance '(multi-ramp '(,@chance-d1))
                                 :more-erratic '(multi-ramp '(,@erratic-d1)))
                               (chance :max '(multi-ramp '(,@max-d2)) :min '(multi-ramp '(,@min-d2))
                                 :chance '(multi-ramp '(,@chance-d2))
                                 :more-erratic '(multi-ramp '(,@erratic-d2))))))))



(defmacro chance-variation 
    (&key (max-r '(1)) (min-r '(0)) (chance-r '(.2)) (erratic-r '(0))
      (melody '(8.0))(last-n nil)(first-n nil)(transpose 0)
      (chance-p '(.2))(max-p '(2))(min-p '(-2))(vary-p 't)
      (max-d '(1)) (min-d '(0)) (chance-d '(.2)) (erratic-d '(0)))
  
  "CHANCE-VARIATION uses CHANCE for both rhythm and duration, with VARIATION for pitch.  Multi-ramp'd"
  
  `'(modular-phrase
    :rhythm '(segment '(chance :max '(multi-ramp '(,@max-r)) :min '(multi-ramp '(,@min-r)) 
                         :chance '(multi-ramp '(,@chance-r)) :more-erratic '(multi-ramp '(,@erratic-r))))
    :pitch '(segment
             '(variation '(transpose (first-n (multi-ramp '(,@first-n)) (last-n (multi-ramp '(,@last-n)) 
                                                                          '(,@melody)))
                            '(multi-ramp '(,@transpose)))
                :type 'pch
                :chance '(multi-ramp '(,@chance-p))
                :min-vary '(multi-ramp '(,@min-p))
                :max-vary '(multi-ramp '(,@max-p))
                :items-to-vary ,vary-p))
    :duration '(segment '(chance :max '(multi-ramp '(,@max-d)) :min '(multi-ramp '(,@min-d)) 
                           :chance '(multi-ramp '(,@chance-d)) :more-erratic '(multi-ramp '(,@erratic-d))))))


(defmacro sequencer-chance-variation
    (&key (type 'cycle)(cycles '1)(motion 'forward)(reset-location nil)(start-pos 0)
      (max-r1 '(1))(min-r1 '(0))(chance-r1 '(.2))(erratic-r1 '(0))
      (max-r2 '(1))(min-r2 '(0))(chance-r2 '(.2))(erratic-r2 '(0))
      (melody '(8.0))(last-n nil)(first-n nil)(transpose 0)
      (chance-p '(.2))(max-p '(2))(min-p '(-2))(vary-p 't) 
      (max-d1 '(4))(min-d1 '(0))(chance-d1 '(.3))(erratic-d1 '(0))
      (max-d2 '(4))(min-d2 '(0))(chance-d2 '(.3))(erratic-d2 '(0)))
  
  "SEQUENCER-CHANCE-VARIATION uses CHANCE sequenced between two positions for both rhythm and duration,
and transposed last-n first-n VARIATION for pitch."
  
  `'(modular-phrase
     :rhythm '(segment
               '(sequencer
                 '((chance :max '(multi-ramp '(,@max-r1)) :min '(multi-ramp '(,@min-r1)) 
                     :chance '(multi-ramp '(,@chance-r1)) :more-erratic '(multi-ramp '(,@erratic-r1)))
                   (chance :max '(multi-ramp '(,@max-r2)) :min '(multi-ramp '(,@min-r2)) 
                     :chance '(multi-ramp '(,@chance-r2)) :more-erratic '(multi-ramp '(,@erratic-r2))))
                 :type ',type :cycles ,cycles :motion ',motion :reset-location ,reset-location 
                 :start-pos ,start-pos))
     :pitch '(segment
              '(variation '(transpose (first-n (multi-ramp '(,@first-n)) 
                                        (last-n (multi-ramp '(,@last-n)) '(,@melody)))
                             '(multi-ramp '(,@transpose)))
                 :type 'pch
                 :chance '(multi-ramp '(,@chance-p))
                 :min-vary '(multi-ramp '(,@min-p))
                 :max-vary '(multi-ramp '(,@max-p))
                 :items-to-vary ,vary-p))
     :duration '(segment 
                 '(sequencer
                   '((chance :max '(multi-ramp '(,@max-d1)) :min '(multi-ramp '(,@min-d1)) 
                       :chance '(multi-ramp '(,@chance-d1)) :more-erratic '(multi-ramp '(,@erratic-d1)))
                     (chance :max '(multi-ramp '(,@max-d2)) :min '(multi-ramp '(,@min-d2)) 
                       :chance '(multi-ramp '(,@chance-d2)) :more-erratic '(multi-ramp '(,@erratic-d2))))
                   :type ',type :cycles ,cycles :motion ',motion :reset-location ,reset-location 
                   :start-pos ,start-pos))))

 

(defmacro riff-chance-variation
    (&key (sequence '(1))(type 'step)(cycles '1)(motion 'forward)(reset-location nil)(start-pos 0)
      (max-r1 '(1))(min-r1 '(0))(chance-r1 '(.2))(erratic-r1 '(0))
      (max-r2 '(1))(min-r2 '(0))(chance-r2 '(.2))(erratic-r2 '(0))
      (melody '(8.0))(last-n nil)(first-n nil)(transpose 0)
      (chance-p '(.2))(max-p '(2))(min-p '(-2))(vary-p 't) 
      (max-d1 '(4))(min-d1 '(0))(chance-d1 '(.3))(erratic-d1 '(0))
      (max-d2 '(4))(min-d2 '(0))(chance-d2 '(.3))(erratic-d2 '(0)))
  
  "RIFF-CHANCE-VARIATION uses CHANCE for both rhythm and duration, using RIFF (with N sequencer'd) to move
between the layers.  Only supports 2 rhythm/duration pairs, though it can easily be expanded to fit more.
It also uses a transposed last-n first-n VARIATION for pitch."
  
  `'(modular-phrase
     :rhythm '(segment
               '(riff '(sequencer '(,@sequence) :type ',type :cycles ,cycles :motion ,motion 
                         :reset-location ',reset-location :start-pos ,start-pos)
                  '((chance :max '(multi-ramp '(,@max-r1)) :min '(multi-ramp '(,@min-r1)) 
                      :chance '(multi-ramp '(,@chance-r1)) :more-erratic '(multi-ramp '(,@erratic-r1)))
                    (chance :max '(multi-ramp '(,@max-r2)) :min '(multi-ramp '(,@min-r2)) 
                      :chance '(multi-ramp '(,@chance-r2)) :more-erratic '(multi-ramp '(,@erratic-r2))))))
     :pitch '(segment
              '(variation '(transpose (first-n (multi-ramp '(,@first-n)) 
                                        (last-n (multi-ramp '(,@last-n)) '(,@melody)))
                             '(multi-ramp '(,@transpose)))
                 :type 'pch
                 :chance '(multi-ramp '(,@chance-p))
                 :min-vary '(multi-ramp '(,@min-p))
                 :max-vary '(multi-ramp '(,@max-p))
                 :items-to-vary ,vary-p))
     :duration '(segment 
                 '(riff '(sequencer '(,@sequence) :type ',type :cycles ,cycles :motion ,motion 
                           :reset-location ',reset-location :start-pos ,start-pos) 
                    '((chance :max '(multi-ramp '(,@max-d1)) :min '(multi-ramp '(,@min-d1)) 
                        :chance '(multi-ramp '(,@chance-d1)) :more-erratic '(multi-ramp '(,@erratic-d1)))
                      (chance :max '(multi-ramp '(,@max-d2)) :min '(multi-ramp '(,@min-d2)) 
                        :chance '(multi-ramp '(,@chance-d2)) :more-erratic '(multi-ramp '(,@erratic-d2))))))))


(defmacro riff-chance-split 
    (&key (sequence '(1))(type 'step)(cycles '1)(motion 'forward)(reset-location nil)(start-pos 0) 
      (max-r1 '(1))(min-r1 '(0))(chance-r1 '(.2))(erratic-r1 '(0))
      (melody1 '(8.0))(n1 '(0))(items1 nil)(transpose1 '(0))
      (length1 '(2))(include1 nil)
      (max-d1 '(4))(min-d1 '(0))(chance-d1 '(.3))(erratic-d1 '(0))
      (max-r2 '(1))(min-r2 '(0))(chance-r2 '(.2))(erratic-r2 '(0))
      (melody2 '(8.0))(n2 '(0))(items2 nil)(transpose2 '(0))
      (length2 '(2))(include2 nil)
      (max-d2 '(4))(min-d2 '(0))(chance-d2 '(.3))(erratic-d2 '(0)))
      
  "This instrument uses the chance function for both rhythm and duration and a 
transposed split-list for pitch.  There are 2 sets of each and the order is controlled with 
the SEQUENCE list."  
  
  `'(modular-phrase
     :rhythm '(segment
               '(riff '(sequencer '(,@sequence) :type ',type :cycles ,cycles :motion ',motion 
                         :reset-location ',reset-location :start-pos ,start-pos)
                  '((chance :max '(multi-ramp '(,@max-r1)) :min '(multi-ramp '(,@min-r1)) 
                      :chance '(multi-ramp '(,@chance-r1)) :more-erratic '(multi-ramp '(,@erratic-r1)))
                    (chance :max '(multi-ramp '(,@max-r2)) :min '(multi-ramp '(,@min-r2)) 
                      :chance '(multi-ramp '(,@chance-r2)) :more-erratic '(multi-ramp '(,@erratic-r2))))))
     :pitch '(progn (initialize 'buffers)
               (segment 
                '(riff '(sequencer '(,@sequence) :type ',type :cycles ,cycles :motion ',motion
                          :reset-location ',reset-location :start-pos ,start-pos)
                   '((split-list '(transpose 
                                   (nth-n '(multi-ramp '(,@n1)) '(multi-ramp '(,@items1)) '(,@melody1))
                                   (multi-ramp '(,@transpose1)))
                       :length '(multi-ramp '(,@length1)) :include '(,@include1))
                     (split-list '(transpose 
                                   (nth-n '(multi-ramp '(,@n2)) '(multi-ramp '(,@items2)) '(,@melody2))
                                   (multi-ramp '(,@transpose2)))
                       :length '(multi-ramp '(,@length2)) :include '(,@include2))))))
     :duration '(progn (initialize 'buffers)
                  (segment 
                   '(riff '(sequencer '(,@sequence) :type ',type :cycles ,cycles :motion ',motion 
                             :reset-location ',reset-location :start-pos ,start-pos) 
                      '((chance :max '(multi-ramp '(,@max-d1)) :min '(multi-ramp '(,@min-d1)) 
                          :chance '(multi-ramp '(,@chance-d1)) :more-erratic '(multi-ramp '(,@erratic-d1)))
                        (chance :max '(multi-ramp '(,@max-d2)) :min '(multi-ramp '(,@min-d2)) 
                          :chance '(multi-ramp '(,@chance-d2)) :more-erratic '(multi-ramp '(,@erratic-d2)))))))))



  











;;;(setq example 
;;;      (chance-random 
;;;       :pc-set (0 11 1 6 5) :range-lo (9.00 6.06) :range-hi (9.07 8.06)
;;;       :max-r (16 3) :min-r (6 0) :chance-r (1 .6 .1)
;;;       :max-d (16 1) :min-d (4 0) :chance-d (1 .4 .1)))