;;Copyright (C) 2001 Soren Goodman
;;Algnalogthmic Musical Score Generator
;;See readme.txt for more info
;;Main-functions.lisp

;;This file contains the main function sections used to run the program, from realize.lisp and modular-phrase.lisp.

;;To DO:  
;;  Fix variation, hangs on nil list (should signal error).
;;  rewrite so that PHRASE is a structure instead of a class...
;;  Clean up buffer and many funtions...many loose ends, threads to clip...

(in-package :cl-user)

(def-CLASS (phrase 
            (:documentation 
             "A phrase stucture is purely for storage of program output.
Each phrase contains an associated instrument, tempo, clock-offset, rhythm, pitch, duration, and
dynamics slot. A duration of 0 indicates *staccato*."))
    ()
  (instrument nil :type instrument)
  (instrument-nums nil :type list) ;;this is for csound
  (tempo nil :type number)
  (rhythm nil :type list)
  (pitch nil :type list)
  (duration nil :type list)
  (dynamics nil :type list)
  (extra nil :type list) ;;this is for csound stuff
  (clock-offset nil :type integer))


;;all functions used in the program MUST GO HERE.  These are used to distinguish 
;;a normal list from a function call.


(defparameter *pitch-functions* '(transpose transpose-midi convert-pitch pch-to-midi midi-to-pch)) 

(defparameter *utility-functions* '(random-reorder nth-n quantize-to-n pick-one first-n last-n last-beat create 
                             repeat-n-times rotate-list biggest-num number-of-attacks list> list>= list< list<=))

(defparameter *ramp-functions* '(multi-ramp ramp linear-ramp expt-ramp))

(defparameter *lisp-functions* '(expt let let* lambda + - * / append cons length eval quote first second third 
                              fourth fifth sixth seventh eighth ninth tenth last nthcdr rest reverse
                              print round list progn floor ceiling loop float floor ceiling list random))

(defparameter *main-functions* '(riff truncate-rhythm chance chancew split-list segment random-from-pc random-in-range
                              variation sequencer buffer))

(defparameter *ref-functions* '(buffer multi-ramp sequencer))

(defparameter *functions-list* '(*pitch-functions* 
                         *utility-functions* 
                         *ramp-functions* 
                         *main-functions* 
                         *lisp-functions*
                         *other-functions*
                         *ramp-types*))

(defparameter *other-functions* '(multi-ramp-f times sequencer-segment sequencer-step sequencer-step-hold round-to-n 
                                       nth-rest))

(defparameter *non-lisp-functions-list* '(*pitch-functions* *utility-functions* *ramp-functions* *main-functions*))

(defparameter *control-functions* '(distributed-section section modular-phrase composition))

;;these are used for get-data, so that it doesn't signal an error when it encounters these symbols
(defparameter *function-keys* '(* S pch single whole bounce forward backward cycle step pattern 
                          default-t num overlap range-hi range-lo))

(defparameter *ramp-types* '(expt linear))

;;;buffers
(defvar temp '())
(defvar sequencer-temp '())
(defvar *buffer-hash* (make-hash-table :test 'eq))
(defvar segment-buffer '())
(defvar step-output ())
(defvar *random2* nil)
(setq *buffers* '(temp step-output *random2*))    
(defvar *runtime-errors* '())
(setq *error-vars* '(*function-error* *perform-error* *utility-error* *runtime-errors*))
(setq *debug* nil)
(defvar cur-beat 0) ;;this should be reset between create's
(defvar *buffer-ref* 0)


(defconstant whole-note '16)
(defconstant half-note '8)
(defconstant quarter-note '4)
(defconstant eighth-note '2)
(defconstant sixteenth-note '1)
(defconstant measure '16) ;;beats per measure in 4/4.
(defconstant midi-c4 '60)  ;;is this right?? 
(defconstant pch-c4 '8.00) 
(defconstant cps-c4 '261.626)
(defconstant cps-a4 '440.000)
(defconstant bignum '10000000000000000000000)
(defconstant smallnum '.00000000000000000000001)

;;this code uses the csound PCH method, whereby the first number 
;;represents the octave and the second the pitch-class.  8.00 is C4, 8.01 is C#4, 7.11 is B3, etc.

(defparameter *create-verbose* nil)
(defparameter *segment-exact-length* t)
(defparameter *tempo* '240) ;;this works for 1 beat = 16th in finale.
(defparameter *more-erratic* '0) ;;0 - *more-erratic-compare*, increases the chance that a call to chance will be zero.
(defparameter *more-erratic-compare* '10.0) ;;usually there is a comparison (> more-erratic (random *more-erratic-compare*)).
(defvar *phrases* (make-hash-table :test 'eq)) ;;all function output is stored here
(defparameter tempo *tempo*)
(defparameter *default-num-attacks* '15)
(defparameter *average-resolution* 15) ;;get-average tests a function *average-resolution* times
(defparameter *average-denominator* 15) ;;reduces the amount that is added to location so that crazyness doesnt happen
(defparameter *staccato* '200) ;;in msec
(defparameter *print-dot* t) ;;determines whether to call (print-dot) for each call to segment
(defparameter *num-dots* '50)
(defparameter *interpolate-location* nil) ;;makes segment increase location slightly when CUR-PHRASE=rhythm

;;;;;;;;;;;;;;;;;;;
;;;LOCAL-GLOBALS;;;
;;;;;;;;;;;;;;;;;;;

;;These variables are not implicitly passed from function to function, since they may be needed many functions
;;down the chain.  They should be set using (let) statements, and are freed up when the function exits.

(defvar location '0) ;;used for single-functions so that they know where they are, i.e. multi-ramp, etc.
(defvar cur-phrase nil) ;;used by functions that need to know if they are working on duration, rhythm, etc.
(defvar cur-position '0)
(defvar default-phrase (make-instance 'phrase))
(defvar rhythm nil)
(defvar pitch nil)
(defvar duration nil)
(defvar dynamics nil)
(defvar extra nil)
(defvar instrument nil)
(defvar instrument-nums nil) ;;this is for csound
(defvar name 'default-phrase)
(defvar num-attacks 0)
(defvar pitch-type 'pch)
(defvar clock-offset '0)
(defvar function nil)
(defvar reset-location nil)
(defvar total-clock '0)
(defvar phrase-list nil)
(defvar list nil)
(defvar type nil)
(defvar cycles nil)
(defvar motion nil)
(defvar n 0)
(defvar start-pos nil)

(defparameter max-duration 32) ;;this value should get superceded by an instruments value (when I get around to it...)



                           
                           
                           
                           
;;;;;;;;;;;;
;;;CREATE;;;
;;;;;;;;;;;;

(defvar *first-create* t)

(defvar *create-documentation*
    "(create 'test-phrase 100) ---> ((test-phrase . #<phrase @ #x20fd6ad2>)) stored in *PHRASES*")

(defparameter *create-prompt*
    (format nil "(create 'phrase-symbol &optional (num-attacks ~S))~%" *default-num-attacks*))

(defparameter creating nil)

(defun create (function &optional (num-attacks (get-data num-attacks)) (clock-offset (get-data clock-offset))
                        (tempo tempo)
                        &aux (type (first (eval function))))
  (if *first-create* (initialize)
    (initialize 'buffers 'dot))
  (let ((*first-create* nil)
        (creating t)
        (name function))
    (when (eql type 'composition) 
      (format t "~&********************************************"))
    (when (or (eql type 'section)(eql type 'distributed-section))
      (format t "~&--------------------------------------------"))
    (format t "~&Creating ~S of type ~S" function type)
    (cond ((eql type 'section) (eval (eval function)))
          ((eql type 'composition)(eval (eval function)))
          ((eql type 'modular-phrase)
           (let ((num-attacks (if (zerop num-attacks) 
                                  (progn 
                                    (format t "~%Warning, num-attacks is zero.  Setting to ~S.~%" *default-num-attacks*) 
                                    *default-num-attacks*)
                                num-attacks)))
             (format t " for ~A attacks." num-attacks)
             (eval (eval function))
             (format t "Done!")))
          ((eql type 'combine)
           (let ((name function))
             (eval (eval function))))
          ((eql type 'distributed-section)
           (let ((name function))
             (eval (eval function))))           
           (t (format t "~&Sorry, no way to CREATE ~S of type ~S.~%" function type)))
    (when (eql type 'composition) 
      (format t "~&********************************************"))
    (when (or (eql type 'section)(eql type 'distributed-section))
      (format t "~&--------------------------------------------")))
  (when (and *first-create* *runtime-errors*)
    (progn (setq *runtime-errors* (reverse (remove-duplicates *runtime-errors* :test #'equal)) )
      (format t "~&Warning, errors during runtime.  See *runtime-errors* for details.~%"))))


;;;;;;;;;;;;;;;;;
;;;COMPOSITION;;;
;;;;;;;;;;;;;;;;;

(defvar *composition-prompt* "(composition '(section-clock-list=((section1 clock1)(section2 clock2)))")


;;note: no longer uses keys
(defun composition (section-clock-list &optional old)
  (if old (error "Composition no longer uses keys.  Please delete the :section-clock-list key from your call.")
    (let ((num-sections (length section-clock-list)))
      (loop for x from 0 to (1- num-sections) append
            (let* ((section-clock (nth x section-clock-list))
                   (section-clock (if (listp section-clock) section-clock
                                    (list section-clock 0)))
                   (section (first section-clock))
                   (total-clock (round (get-data (second section-clock))))
                   (output-phrases (create section))) ;;wait to add total-clock
              (add-clock-to-section section total-clock)
              output-phrases)))))

(defun add-clock-to-section (function &optional (clock 0))
  (setq *function-error* 'add-clock)
  (let ((clock (get-data clock))
        (type (cond ((null function) ())
                    ((listp function) (first function))
                    ((symbolp function) (add-clock-to-section (eval function) clock)))))
    (when type (cond ((eql type 'section)
                      (let ((phrase-list (find-keyword ':phrase-list function)))
                        (loop for n from 0 to (1- (length phrase-list)) do
                              (add-clock (nth n phrase-list) clock))))
                     (t (format t "Sorry, can't add-clock-to-section; ~S is not an section.~%" function))))))

(defun add-clock (phrase clock &aux (clock (round clock)))
  (let ((phrase (when (symbolp phrase)
                   (fap phrase))))
    (setf (:clock-offset phrase)
      (+ (:clock-offset phrase) clock))))
  

;;;;;;;;;;;;;
;;;SECTION;;;
;;;;;;;;;;;;;

(defvar *section-documentation* '()) 

(defparameter *section-prompt* 
    "(section :phrase-list :instrument-list :tempo-list :num-attacks-list :clock-offset-list)")

(defun section (&key phrase-list instrument-list tempo-list num-attacks-list
                       clock-offset-list)
  (setq *function-error* 'section)
  (let* ((num-phrases (length phrase-list))
         (instrument-list (CHECK-AND-MODIFY-LENGTH instrument-list num-phrases))
         (tempo-list (CHECK-AND-MODIFY-LENGTH tempo-list num-phrases))
         (num-attacks-list (CHECK-AND-MODIFY-LENGTH num-attacks-list NUM-phraseS))
         (clock-list (check-and-modify-length clock-offset-list NUM-phraseS)))
    (loop for x from 0 to (1- num-phrases) collect
          (let* ((phrase (nth x phrase-list))
                 (instrument-name (nth x instrument-list))
                 (instrument (get-data instrument-name))
                 (tempo (get-data (nth x tempo-list)))
                 (num-attacks (round (get-data (nth x num-attacks-list))))
                 (clock-offset (round (get-data (nth x clock-list))))
                 (name phrase))
            (when *create-verbose*
              (format t "phrase ~S, with instrument ~S at a tempo of ~S for ~S attacks at clock ~S.~%" 
                (1+ x) instrument-name tempo num-attacks (+ clock-offset total-clock)))
            (create phrase)))
    t))

;;;;;;;;;;;;;;;;;;;;
;;;MODULAR-PHRASE;;;
;;;;;;;;;;;;;;;;;;;;

(defvar *modular-phrase-documentation* 
    "A modular-phrase consists of :pitch, :rhythm, :duration and :dynamics segments.  To create a modular
phrase, simply enter in a quoted (') function call for each of the above parameters.  They are automatically 
turned into a segment, checked for errors, checked to make sure they are monophonic (with MONOPHONIC set to
T, default) and saved.  To retrieve a saved phrase, type (find-associated-phrase 'phrase-name).

Added to this function are slots for :extra and :instrument-nums.  These are used for Csound scores.")

(defparameter *modular-phrase-prompt* 
  "(modular-phrase :num-attacks :rhythm :pitch :duration :dynamics :extra 
:instrument :instrument-nums :delay  :monophonic :tempo :pre-eval :post-eval)")

(defun modular-phrase (&key pre-eval post-eval (num-attacks num-attacks) rhythm pitch duration dynamics 
                                (instrument instrument) instrument-nums (tempo tempo)(monophonic t)(extra nil) delay)
  (setq *function-error* 'modular-phrase)
  (eval pre-eval)
  (let* ((instrument-nums instrument-nums)
         (instrument (or (get-data instrument) default-instrument))
         (tempo (get-data tempo))
         (max-duration (duration-to-beats (:max-duration instrument)))
         (cur-phrase 'rhythm)
         (rhythm (rhythm-to-durations (segment rhythm))) ;;removes all zeros, adding them to the prev integer
         (num-attacks (length rhythm))
         (cur-phrase 'pitch)
         (pitch (segment pitch))
         (cur-phrase 'duration)
         (duration (if monophonic (fix-durations rhythm (segment duration))
                     (segment duration)))
         (cur-phrase 'dynamics)
         (dynamics (segment dynamics))
         (dynamics (if (listp dynamics) dynamics (list dynamics)))
         (extra (when extra (mapcar (lambda (x) (mapcar #'float (segment x))) extra)))
         (delay (get-data delay))
         (cur-phrase 'instrument-nums)
         (instrument-nums (when instrument-nums (segment instrument-nums)))
         )
    (eval post-eval)
    (save-phrase delay)))


;;all phrases created are stored in *phrases* until (initialize 'phrases)
;;there can only be 1 rendering of each phrase, so if you want a phrase repeated in process but not 
;;exactly you have to copy it to a new symbol and use that as the repeat.  I might add functionality
;;to allow multiple copies in the same section, but not too yet.

(defun save-phrase (&optional delay)
  (setq *function-error* 'save-phrase)
  (let* ((rhythm (remove nil rhythm))
         (rhythm (when rhythm (mapcar #'round rhythm)))
         (pitch (remove nil pitch))
         (pitch (when pitch (mapcar #'make-pch pitch)))
         (duration (remove nil duration))
         (duration (if (null duration) '(1) (mapcar #'round duration)))
         (dynamics (remove nil dynamics))
         (dynamics (if (null dynamics) '(127) (mapcar #'round dynamics)))
         (clock-offset (if clock-offset clock-offset 0))
         (clock-offset (if delay (+ clock-offset delay) clock-offset))
         (instrument-nums (mapcar #'round instrument-nums)))
                       
    (if (or (null rhythm)(null pitch)) (setf phrase nil)
      (setf phrase (make-instance 'phrase 
                     :rhythm rhythm
                     :pitch pitch
                     :duration duration
                     :dynamics dynamics
                     :extra extra
                     :instrument instrument
                     :instrument-nums instrument-nums
                     :tempo tempo
                     :clock-offset clock-offset)))
    (setf (gethash name *phrases*) phrase)))
       

;;;;;;;;;;;;;
;;;SEGMENT;;;
;;;;;;;;;;;;;


(defparameter *segment-prompt*
    "(segment 'function &optional (length num-attacks) (start-position 0)(force-exact-length t))")

;;;NOTE:  DO NOT USE SEGMENT WITHIN OTHER FUNCTIONS.  It must be used at the top level only.
(defun segment (function &optional (length num-attacks)(start-position 0)(force-exact-length t)
                         &aux (num-attacks (if (zerop num-attacks) (+ start-position length) num-attacks)))
  (setq *function-error* 'segment)
  (let* ((num-attacks (get-data num-attacks))
         (segment-buffer '())
         (start-position (get-data start-position))
         (length (get-data length)))
    (if (<= num-attacks 0) (error "SEGMENT:  Sorry, num-attacks must be greater than 0.") 
      (loop as cur-position = (+ start-position (length segment-buffer))
          while (< (- cur-position start-position) length) do
            (let* ((location (/ cur-position (1- num-attacks)))
                   (output (if (eql cur-phrase 'rhythm)
                               (get-data-rhythm function)
                             (get-data function))))
              (when *debug* (format t "location ~S output ~S~%" location output))
              (when *print-dot* (print-dot))
              (if (atom output) (setq segment-buffer (add-to-end segment-buffer output))
                (setq segment-buffer (append segment-buffer output))))))
    (if (or *segment-exact-length* force-exact-length)
        (check-and-modify-length segment-buffer length)
      segment-buffer)))

;;this function ensures that all output to a rhythm segment is non-zero, as well as a little
;;bit of interpolation on location
(defun get-data-rhythm (function)
  (let ((data
         (loop for count from 0 to max-duration doing
               (let ((temp (get-data function)))
                 (when (and (listp temp) (> (apply #'+ temp) 0))
                   (return (mapcar #'round temp)))
                 (when (and (numberp temp)(not (zerop temp)))
                   (return (+ count (round temp))))))))
    (if data data max-duration)))

(defparameter print-next-dot 0)
(defparameter prev 0)

(defun print-dot (&optional (compare1 cur-position) (compare2 num-attacks))
  "Prints *num-dots* dots . based on COMPARE."
  (when (not (= prev compare1))
    (progn (decf print-next-dot)
      (setq prev compare1)))
  (if (<= print-next-dot 0)
      (progn (format t ".")
        (when (> compare2 0)
          (setq print-next-dot (round (/ compare2 *num-dots* .22))))) ;;this is a hack...".22"???
    nil))
  
           
;;;;;;;;;;;;
;;;CHANCE;;;
;;;;;;;;;;;;

(defparameter *chance-prompt*
    "(chance :max 1 :min 0 :chance .5 :more-erratic *more-erratic* :count 0)")

;;this calls chance-f.  If cur-phrase = rhythm, then it loops the call increasing location 
;;according to average-chance * num-beats until we get a non-zero.

(defun chance (&key (max 1) (min 0) (chance .5) (more-erratic *more-erratic*))
  (setq *function-error* 'chance)
  (if (eq cur-phrase 'rhythm)
      (let* ((average-chance (/ (get-average chance) *average-denominator*))
             (temp '0)
             (data
              (loop for count from 0 to max-duration do
                    (let* ((location (+ location  (* count average-chance)))
                           (location (if (> location 1) 1 location)))
                      (setq temp (chance-f max min chance more-erratic))
                      (when (not (zerop temp)) (return (+ count temp)))))))
        (if data data max-duration))
    (chance-f max min chance more-erratic)))

(defun chance-f (max min chance more-erratic &optional (count 0))
  (setq *function-error* 'chance-f)
  (let* ((max (round (get-data max)))
         (min (round (get-data min)))
         (more-erratic (get-data more-erratic))
         (chance (get-data chance))
         (working-max (progn (when (> min max)
                               (setq temp max
                                   max min 
                                   min temp))
                        (- max min))))
    (loop for count from 0 to working-max do
          (let ((make-erratic (< (random *more-erratic-compare*) more-erratic)))
            (cond (make-erratic (return (+ min count)))
                  ((>= chance 1) (return max))
                  ((<= chance 0) (return min))
                  ((= count working-max) (return max))
                  (t
                   (let ((modified-chance (* (- working-max count)
                                             (/ chance (1+ count))))                                           
                         (stoch (random 1.0)))
                     (if (> stoch modified-chance)
                         (return (+ count min))))))))))

;;;;;;;;;;;;;
;;;CHANCEW;;;
;;;;;;;;;;;;;

(defparameter *chancew-documentation*
  "ChanceW, aka chance window, creates output in the range from min-max (can be reversed).  It does this
by creating a set of windows for each possible output, giving each window a chance, and then using
a random function to pick a window.  CHANCE is the probability of generiting MAX, and a curve of type
CURVE is created from CHANCE with each possible output.  Note that ChanceW only accepts integers for
MAX and MIN (as does the old CHANCE function).")

(defparameter *chancew-prompt*
  "(chancew :max 1 :min 0 :chance .5 :window-curve 'linear)")

(defun chancew (&key (max 1) (min 0) (chance .5) (window-curve 'linear))
  (setq *function-error* 'chancew)
  (let* ((max (round (get-data max)))
         (min (round (get-data min)))
         (chance (get-data chance))
         (num-windows (progn (when (> min max)
                               (psetq max min
                                   min max))
                        (- max min -1))))
    (cond ((<= num-windows 0) min)
          ((= num-windows 1) 
           (if (> chance (random 1.0)) max min))
          (t
           (let* ((window-curve (get-data window-curve))
                  (window-chances (window-ramp num-windows chance window-curve))
                  (iching (random 1.0))
                  (window (pick-window iching window-chances)))
             (+ window min))))))

(defun window-ramp (windows end &optional (curve 'linear))
  (setq *function-error* 'ramp-segment)  
  (add-to-end 
   (do* ((count 0 (1+ count))   
         (location 0 (/ count (1- windows)))
         (temp nil))
        ((= count windows)(return (reverse temp)))
     (push (ramp 0 end :type curve) temp))
       1))

(defun pick-window (iching window-list)
  (loop for n from 1 to (length window-list) do
        (let ((item (nth n window-list)))
          (when (< iching item) (return (- (length window-list) n 1))))))
            
;;;;;;;;;;;;;;;;
;;;SPLIT-LIST;;;
;;;;;;;;;;;;;;;;

(defparameter *split-list-prompt*
    "(split-list list :length 2 :include nil)")

(defun split-list (list &key (length 3) include)
  (setq *function-error* 'split-list)
  (let ((list (check-list (get-data list)))
        (length (round (get-data length)))
        (include (check-list (get-data include))))
    (cond ((null list)
           (error "Cannot SPLIT-LIST without a list!"))
          ((> (length include) length)
           (progn 
             (make-error "Warning, split-list error!  INCLUDE is larger than LENGTH, dropping INCLUDE.")
             (split-list list length)))
          ((<= (+ (length list)(length include))
               length)
           (order-to-list list (append list include)))
          (t
           (let* ((de-include-list (remove-items include list))
                  (shrunk-list (pick-n (- length (length include)) de-include-list))
                  (new-list (append include shrunk-list)))
             (if (< (length new-list)
                    length)
                 (split-list list length new-list)
               (order-to-list list new-list)))))))


;;;;;;;;;;;;;
;;;COMBINE;;;
;;;;;;;;;;;;;

(defparameter *combine-prompt*
    "(merge phrase1 phrase2 &rest options ['cut-durations 'favor-pitch-hi 'favor-pitch-lo])")


;;;TO DO:  create support for clock-offset 
(Defun combine (phrase1 phrase2 &key options (attacks (list num-attacks num-attacks)))
  (setq *function-error* 'combine)
  (when (and (> (first attacks) 0)(> (second attacks) 0)) 
    (let ((num-attacks1 (get-data (first attacks)))
          (num-attacks2 (get-data (second attacks))))
      (create phrase1 num-attacks1) 
      (create phrase2 num-attacks2)))
  (format t "~&Combining ~S and ~S." phrase1 phrase2)
  (let* ((phrase1 (find-associated-phrase phrase1))
         (phrase2 (find-associated-phrase phrase2))
         (cut-durations (member 'cut-durations options))
         (favor-pitch (cond ((member 'favor-pitch-hi options) 'hi)
                        ((member 'favor-pitch-lo options) 'lo))))
    (when (or (null phrase1) (null phrase2))
      (error "Combine: input must be valid phrases."))
    (combine-f phrase1 phrase2 cut-durations favor-pitch)))

(defgeneric combine-f (phrase1 phrase2 cut-durations favor-pitch))
    
(defmethod combine-f ((phrase1 phrase)(phrase2 phrase) (cut-durations null) (favor-pitch null))
  (setq *function-error* 'combine-f)
  (let* ((rhythm1 (rhythm-to-durations (:rhythm phrase1)))
         (pitch1 (check-and-modify-length (:pitch phrase1) (length rhythm1)))
         (duration1 (fix-durations rhythm1 (:duration phrase1)))
         (dynamics1 (:dynamics phrase1))
         (rhythm2 (rhythm-to-durations (:rhythm phrase2)))
         (pitch2 (check-and-modify-length (:pitch phrase2) (length rhythm2)))
         (duration2 (fix-durations rhythm2 (:duration phrase2)))
         (dynamics2 (:dynamics phrase2))
         (rhythm-duration1 (combine-rhythm-duration rhythm1 duration1))
         (rhythm-duration2 (combine-rhythm-duration rhythm2 duration2))
         (tempo (:tempo phrase1))
         (clock-offset (:clock-offset phrase1))
         (temp nil)
         (rhythm nil)
         (pitch nil)
         (duration nil)
         (dynamics nil)
         (rhythm-duration2 (substitute 2 1 rhythm-duration2))
         (shortest (max (length rhythm-duration1) (length rhythm-duration2)))
         )
    (loop for n from 0 below shortest do 
          (let ((item1 (nth n rhythm-duration1))
                (item2 (nth n rhythm-duration2)))
            (when *print-dot* (print-dot n shortest))
;;;          (format nil "~S item1 ~S item2 ~S~%" n item1 item2)
            (cond 
             ((null item2)
              (let ((item-number
                     (num-attacks (first-n (1+ n) rhythm-duration1))))
                (setq temp (append (reverse (nthcdr n rhythm-duration1)) temp)
                    pitch (append(reverse (nthcdr item-number pitch1)) pitch)
                    duration (append (reverse (nthcdr item-number duration1)) duration)
                    dynamics (append (reverse (nthcdr item-number dynamics1)) dynamics))
                (return)))
             ((null item1)
              (let ((item-number
                     (num-attacks (first-n (1+ n) rhythm-duration2))))
                (setq temp (append (reverse (nthcdr n rhythm-duration2)) temp)
                    pitch (append(reverse (nthcdr item-number pitch2)) pitch)
                    duration (append (reverse (nthcdr item-number duration2)) duration)
                    dynamics (append (reverse (nthcdr item-number dynamics2)) dynamics))
                (return)))
             ((eq 'S item1)
              (push 'S temp))
             ((zerop item1)
              (cond ((eq item2 'S) 
                     (push 'S temp))
                ((zerop item2) 
                 (push 0 temp))
                (t 
                 (let ((item-number 
                        (num-attacks (first-n (1+ n) rhythm-duration2))))
                   (push item2 temp)
                   (push (nth item-number pitch2) pitch)
                   (push (nth item-number dynamics2) dynamics)))))
             (t
              (let ((item-number (num-attacks (first-n n rhythm-duration1))))
                (push item1 temp)
                (push (nth item-number pitch1) pitch)
                (push (nth item-number dynamics1) dynamics))))))
    (let* ((temp (reverse temp))
           (rhythm (extract-rhythm temp))
           (duration (extract-durations temp))
           (pitch (reverse pitch))
           (dynamics (reverse dynamics)))
     ; (format t "~&temp ~S~%" temp)
      (save-phrase))))


;;;;;;;;;;;;;;;
;;;VARIATION;;;
;;;;;;;;;;;;;;;

(defparameter *variation-prompt*
    "(variation list :type 'num :chance .2 :min-vary -1 :max-vary 1 :items-to-vary 't)")

(defun variation (list &key (type 'num) (chance .2) (min-vary -1)(max-vary 1)(items-to-vary 't))
  (setq *function-error* 'variation)
  (let* ((list (get-data list))
         (type (get-data type))
         (items-to-vary (get-data items-to-vary))
         (items-to-vary (if (null items-to-vary) 
                            t
                          items-to-vary)))
    (variation-f list type chance min-vary max-vary items-to-vary)))

(defgeneric variation-f (list type chance min-vary max-vary items-to-vary))

(defmethod variation-f (list (type (eql 'pch)) chance min-vary max-vary items-to-vary)
  (let ((midi-list (convert-pitch 'pch 'midi list))
        (midi-vary (if (listp items-to-vary) (convert-pitch 'pch 'midi items-to-vary)
                     items-to-vary)))    
    (convert-pitch 'midi 'pch 
      (variation-f midi-list 'num chance min-vary max-vary midi-vary)))) 

(defmethod variation-f (list (type (eql 'num)) chance min-vary max-vary items-to-vary)
  (vary-num list chance min-vary max-vary items-to-vary))

(defmethod variation-f (list type chance min-vary max-vary items-to-vary)
  (format t "Sorry, variation type ~S not yet implemented.~%" type))

(defun vary-num (list chance min-vary max-vary items-to-vary)
  (setq *function-error* 'vary-num)
  (let* ((list (get-data list))
         (item (first list))
         (not-varyp (< (get-data chance) (random 1.0))))
    (cond ((null list) ())
          (not-varyp (cons item (vary-num (rest list) chance min-vary max-vary items-to-vary)))
          (t 
           (let* ((c-items-to-vary (get-data items-to-vary))
                  (c-min-vary (round (get-data min-vary)))
                  (c-max-vary (round (get-data max-vary))))      
            (if (or (eq c-items-to-vary t) (member item c-items-to-vary))
                (cons (vary-item item c-min-vary c-max-vary)
                        (vary-num (rest list) chance min-vary max-vary items-to-vary))
              (cons item (vary-num (rest list) chance min-vary max-vary items-to-vary))))))))

(defun vary-item (item least most)
  (setq *function-error* 'vary-item)
  (let* ((most (- most -1 least))
         (most (if (not (fixnump most))
                   (float most)
                 most)))
    (cond ((<= most 0) 0)
          ((eq cur-phrase 'rhythm)
           (round (+ item least
                     (random most))))
          (t (+ item least
                (random most))))))
  
;;;;;;;;;;;;;;;;;;;;
;;;PITCH-SECTIONS;;;
;;;;;;;;;;;;;;;;;;;;

;;;note that there is no built in error checking for (> range-lo range-hi).  Must be implemented in method.
;;;ex.

;;;  (when (> cur-range-lo cur-range-hi)
;;;    (setq temp cur-range-lo
;;;        cur-range-lo cur-range-hi
;;;        cur-range-hi temp-var))

;;;;;;;;;;;;;;;;;;;;;
;;;RANDOM-IN-RANGE;;;
;;;;;;;;;;;;;;;;;;;;;

(defparameter *last-pitch* '0)

(defun random-in-range (&optional (repeat-pitches nil)(instrument instrument))
  (let* ((instrument (if (null instrument) default-instrument instrument))
         (range-hi (pch-to-midi (get-range (:range-hi instrument))))
         (range-lo (pch-to-midi (get-range (:range-lo instrument))))
         (available (truncate (- range-hi range-lo))))
    (if repeat-pitches (midi-to-pch (+ range-lo (random available)))
      (loop 
        (let ((output (midi-to-pch (+ range-lo (random available)))))
          (when (not (= output *last-pitch*))
            (return (setq *last-pitch* output))))))))
            
;;;;;;;;;;;;;;;;;;;;
;;;RANDOM-FROM-PC;;;
;;;;;;;;;;;;;;;;;;;;

(defun random-from-pc (pc-set &key (range-lo 5) (range-hi 6))
  (setq *function-error* 'single-pc)                  
  (let* ((pc-set (get-data pc-set))
         (range-lo (round-pch (get-range range-lo)))
         (range-hi (round-pch (get-range range-hi)))
         (range-lo (if (> range-lo range-hi)
                       (progn (setq temp range-lo) range-hi)
                     range-lo))
         (range-hi (if (< range-hi range-lo)
                       temp range-hi))
         (available-pitches (find-available-pitches pc-set range-lo range-hi)))
    (if (null available-pitches)
        (format t "No available PC's ~S in range ~S to ~S.~%" pc-set range-lo range-hi)
      (pick-one available-pitches))))

;;;;;;;;;;;;;;;;;;;;
;;;RANDOM-REORDER;;;
;;;;;;;;;;;;;;;;;;;;

(defparameter *random-reorder-documentation* 
  "Reorder takes a list and scrambles the order of the items randomly.")

(defparameter *random-reorder-prompt*
  "(reorder 'list-to-reorder)")

(defun random-reorder (list &key include-front include-back)
  (setq *function-error* 'random-reorder)
  (let ((list (get-data list))
        (include-front (check-list (get-data include-front)))
        (include-back (check-list (get-data include-back))))
    (if (listp list)
        (append include-front
                (loop for x from 0 below (length list) collect
                      (let ((item (pick-one list)))
                        (setq list (remove-first-occurence item list))
                        item))
                include-back)
      (progn (make-error (format nil "RANDOM-REORDER: ~S is not a list!" list))
        list))))

