;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;Sept 2001
;;;textures thesis project
;;;performance.lisp

(in-package :cl-user)

#|
This file contains the performance class and associated methods.  My fist project
is a method PERFORMANCE which translates a full SECTION object into an events-list.

The PERFORMANCE method has 2 inputs, a PERFORMANCE-OUTPUT object (event-list, csound-score, midi, etc.),
and a PERFORMANCE-INPUT object (SECTION object, event-list, INSTRUMENT, INSTRUMENT-LIST, etc.).

TO DO: Fix MAKE-PHRASE-PLAIN-EVENTS, it is recursive and very bulky
       Fix MAKE-PHRASE-MIDI, also way too bulky...    
|#
  


(def-class p-output
    ()
  (data nil :type list))

(def-class event-list 
    (p-output))

(setf event-list (make-instance 'event-list))

(def-class (plain-event-list
            (:documentation 
"a PLAIN-EVENT-LIST ignores the :extra slot."))
    (event-list)
  )

(defvar plain-event-list (make-instance 'plain-event-list))
  
(def-class csound-score 
    (p-output))

(defvar csound-score (make-instance 'csound-score))

(def-class midi 
    (p-output)
  )

(defvar midi (make-instance 'midi))

;;perform looks at 'input and searches phraseS for the first occurance, then performs that into
;;the TYPE object.

(defgeneric perform (input output &optional clock)
  (:documentation 
   "The perform method translates the input format into output and saves in :data of output.~%"))

(defmethod perform (unknown-input unknown-output &optional clock)
  (format t "~%PERFORM method error:  Unknown INPUT ~S;~%Unknown OUTPUT ~S ~S.~%"
    unknown-input unknown-output *perform-error*))

(defmethod perform ((input symbol) output &optional (clock 0)
                    &aux (function (eval input)))
  (setq *perform-error* 'input-symbol)
  (if (listp function)
      (let* ((clock (round (get-data clock)))
             (type (first function)))
        (cond ((or (eq type 'modular-phrase) (eq type 'combine))
               (let ((phrase (find-associated-phrase input)))
                 (format t "~%Performing ~S " input)
                 (perform phrase output clock)))
              ((eq type 'distributed-section)
               (let* ((output-instruments-list (find-keyword ':output-instruments-list function))
                      (output-surnames (find-keyword ':output-surnames function))
                      (output-names (mapcar #'(lambda (instrument-list surname)
                                                (loop for n from 0 below (length instrument-list) collect
                                                      (read-from-string (concatenate 'string 
                                                                          (princ-to-string (nth n instrument-list))
                                                                          (princ-to-string surname)))))
                                      output-instruments-list output-surnames))
                      (phrase-list (apply #'append output-names)))
                 (format t "~%Performing DISTRIBUTEd phrase ~S." input)
                 (prog1 (perform-distribute phrase-list clock)
                   (format t "~&Wrapping up ~S." input))))
              ((eq type 'section) 
               (let ((phrase-list (find-keyword ':phrase-list function)))
                 (format t "~%Performing ~S SECTION." input)
                 (prog1 (perform-section phrase-list output clock)
                   (format t "~&Wrapping up ~A." input))))
              ((eq type 'composition)
               (let ((section-list (unquote (second function))))
                 (format t "~%Performing ~S COMPOSITION." input)
                 (perform-composition section-list output clock)))
              (t (format t "~&Sorry, input type ~S has not yet been implemented.~%" type))))
    (perform function output clock)))
    
(defun perform-composition (section-list output &optional (clock 0))
  (setq *perform-error* 'composition-midi)
  (when section-list 
    (let ((section (first (first section-list))))
      (append (perform section output clock)
              (perform-composition (rest section-list) output clock)))))

;;;we are hitting heap limit here for big files...
(defun perform-section (phrase-list output &optional (clock 0))
  (setq *perform-error* 'perform-section)
  (when phrase-list
    (append (perform (first phrase-list) output clock)
            (perform-section (rest phrase-list) output clock))))

(defun perform-distribute (phrase-list output &optional (clock 0))
  (setq *perform-error* 'perform-section)
  (loop for n from 0 below (length phrase-list) append
        (let ((phrase (nth n phrase-list)))
          (format t "~&Performing ~S " phrase)
          (perform (find-associated-phrase phrase) output clock))))

(defmethod perform ((input phrase) (output midi) &optional (clock 0))
  (declare (ignore clock))
  (setf *perform-error* 'perform-phrase)
  (when (fboundp 'gc)(gc)) ;;GC - garbage collector - is not common lisp
  (setf (:data output)
    (let* ((instrument (get-data (:instrument input)))
           (instrument (if instrument instrument
                         default-instrument))
           (mute (get-data (:mute instrument)))
           (channel (get-data (:channel instrument)))
           (clock-offset (round (get-data (:clock-offset input))))
           (tempo (get-data (:tempo input)))
           (tempo (if (null tempo) *tempo* tempo))
           (rhythm (durations-to-rhythm (check-list (:rhythm input))))
           (pitch (check-list (:pitch input)))
           (pitch-template (:pitch-template instrument)) ;;mostly for percussion
           (pitch (make-midi (if pitch-template (convert-pitches-to-template pitch pitch-template)
                               pitch)))
           (duration (check-list (:duration input)))
           (duration (if duration duration
                       '(1)))
           (dynamics (check-list (:dynamics input)))
           (dynamics (if dynamics dynamics
                       (list (:max-volume instrument))))
           (num-attacks (num-attacks rhythm))
           (pitch (check-and-modify-length pitch num-attacks))
           (duration (check-and-modify-length duration num-attacks)) 
           (dynamics (check-and-modify-length dynamics num-attacks))
           (start-measure (beats-to-measures clock-offset))
           (end-measure (beats-to-measures (last-beat input t))))
      (if (or mute (null pitch) (null rhythm)) 
          (format t "nil")
        (progn (format t "at clock ~S from mm.~S to mm.~S."  clock-offset start-measure end-measure)
          (make-phrase-midi rhythm pitch duration dynamics channel tempo clock-offset))))))


;;;need to rewrite this with a loop, waaaaaaay to bulky and slow.
(defun make-phrase-midi (rhythm pitch duration dynamics channel tempo &optional (clock 0) 
                                 &aux (clock (round clock)))
  (setf *perform-error* 'make-phrase)
  (cond 
   ((null rhythm) ())
   ((zerop (first rhythm))
    (make-phrase-midi (rest rhythm) pitch duration dynamics channel tempo (1+ clock)))
   (t 
    (let* ((time-on (* clock (ms-per-beat tempo)))
           (time-off (+ time-on
                        (if (zerop (first duration))
                            *staccato*
                          (* (ms-per-beat tempo)
                             (first duration)))))
           (event 
            (list (round time-on)
                  (first pitch)
                  (round time-off)
                  (round channel)
                  (first dynamics))))
      (cons event  
            (make-phrase-midi (rest rhythm) (rest pitch) (rest duration) (rest dynamics)
                                       channel tempo (1+ clock)))))))

(defun save (filename data)
  (save-as-midi filename (perform data midi))
  (format t "~%~S" filename))

(defun create-save (filename symbol)
  (if (listp symbol) 
      (let ((function (first symbol))
            (num-attacks (get-data (second symbol))))
        (create function num-attacks)
          (save filename function))
    (progn (create symbol)
          (save filename symbol))))

;;creates and saves n versions, adds n.midi to filename
;;;(create-n-save "test" object 3)-->"test1.midi" "test2.midi" "test3.midi"

(defun create-n-save (filename symbol &optional n &aux (filesystem (machine-type)))
  (let* ((file (cond ((equal filesystem "  Power Macintosh")
                      (mac-namestring (format nil "ccl:textures;midi;~A" filename)))
                     ((equal filesystem "x86")
                      (format nil "c:\\soren\\midi\\~A" filename))
                     (t (format t "Sorry, filesystem ~S is not yet supported.~%" filesystem))))
         (file (if n (concatenate 'string file (write-to-string n) ".midi")
                 (concatenate 'string file ".midi"))))
    (create-save file symbol)
    (if (or (null n)
          (zerop n))
        t
      (create-n-save filename symbol (1- n)))))

(defun create-test (var &optional length instrument)
  (if length
      (create-n-save "test" (list var length))
    (create-n-save "test" var)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;CONVERT-PITCHS-TO-TEMPLATE;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#|template is in the form ((range-lo (list-of-new-pitches) instrument-name [ignored])(range-lo (list-of-new-pitches) ...))
The main goal of this conversion is to allow a set of percussion instruments  to be composed as a single chromatic
set which is then converted to the real midi (or other) representation. 
For example 9 instruments: 5 toms, a snare and 3 temple-blocks, can be notated as 5.0 - 5.09 and then translated
to their real midi equivalents when saving the file.  Range-hi is determined by number of replacements.

(convert-pitches-to-template '(5.01 5.01 5.02 5.03 5.04 5.05) '((5.01 (6.06 5.03) bongos) (5.03 (70) cymbal)))
----->(6.06 6.06 5.03 7.1 5.04 5.05)|#

(defun convert-pitches-to-template (pitches template)
  (setq *perform-error* 'convert-pitches-to-template)
  (let ((template (correct-template template)))
    (cond ((null pitches) nil)
          ((null template) pitches) ;;this won't happen if we're coming from PERFORM
          ((or (not (listp pitches))(not (listp template))) pitches)
          ((floatp (first pitches)) ;;make sure list returned is same format as given.  For use outside perform.
           (make-pch (convert-to-template pitches template)))
          (t (make-midi (convert-to-template (make-pch pitches) template))))))


;;PITCHES are PCH and TEMPLATE replacements are in MIDI, to avoid problems converting 50->51, 51->52, etc...
(defun convert-to-template (pitches template)
  (setq *perform-error* 'convert-to-template)
  (loop for n from 0 below (length template) do
        (let* ((cur-template (nth n template))
               (range-lo (first cur-template))
               (replacements (second cur-template)))
          (loop for o from 0 below (length replacements) do
                (let ((replacement (nth o replacements)))
                  (setq pitches (subst replacement range-lo pitches)
                      range-lo (next-pch range-lo))))))
    (mapcar #'make-pch pitches)) ;;for non-converted pitches

;;makes sure RANGE-LO is PCH, and LIST-OF-NEW-PITCHES is midi.  
(defun correct-template (template)
  (if (or (null template)(not (listp template))) template
    (loop for n from 0 below (length template) collect
          (let* ((cur-template (nth n template))
                 (replacements (make-midi (second cur-template)))
                 (range-lo (make-pch (first cur-template))))
            (list range-lo replacements)))))


         
            
  
               