;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;ramp-functions.lisp
;;;nov 2001
;;;textures thesis project

(in-package :cl-user)

;;;;;;;;;;;;;;;;;;;;
;;;RAMP-FUNCTIONS;;;
;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;
;;;MULTI-RAMP;;;
;;;;;;;;;;;;;;;;

#|NOTE:  You can enter "*" to break a ramp
(multi-ramp '(1 2 * 3 4 * 5 * 6 7 8 * 9))
----> ramps are 1->2, 3->4, 5, 6->7, 7->8, 8. 
Note that elements segregated by *'s are repeated (i.e., 6->7, 7-> = 6 6.33 6.66 7.0 7.0 7.33 7.66 8.0)|#

;;Total revamp of multi-ramp.  Ramp-list can now contain *, meaning seperate items (do not ramp).
;;You can also decide how the ramp deciding how long to spend on each segment works (expt, etc.)
;;I will eventually implement the ability to decide the type of ramp between each segment.

(defun multi-ramp (buffer-ref ramp-list &key (type 'linear)(segment-ramp 'linear)(quantize nil))
  (let ((output (gethash buffer-ref *buffer-hash*)))
    (when (null output)
      (setq output (setf (gethash buffer-ref *buffer-hash*) 
                     (multi-ramp-f ramp-list type segment-ramp quantize))))
    (nth cur-position output)))

(defun multi-ramp-f (ramp-list type segment-ramp quantize)
  (setq *function-error* 'multi-ramp)
  (if (null ramp-list) nil
    (let* ((ramp-list (correct-ramp-list (get-data ramp-list)))
           (type (get-data type))
           (segment-ramp (get-data segment-ramp))
           (ramp-list (if (listp ramp-list) ramp-list
                        (list ramp-list)))
           (segments (num-segments ramp-list))
           (num-attacks-list (divide-into-parts num-attacks segments segment-ramp))
           (output
            (loop for n from 0 below segments append
                  (let* ((item1 (nth-segment n ramp-list))
                         (item2 (nth-segment n ramp-list t))
                         (length (nth n num-attacks-list)))
                    (if (<= length 1) (let ((location .5)) 
                                        (list (ramp item1 item2 :type type))) 
                      (loop for x from 0 below length collect
                            (let ((location (/ x (1- length))))
                              (ramp item1 item2 :type type))))))))
      (if quantize (mapcar #'quantize-to-n 
                     (location-loop quantize (length output))
                     output)
        output))))
              
;;returns a list of PARTS length totalling NUM.  
(defun divide-into-parts (num parts &optional (ramp 'linear))
  (let* ((list (loop for x from 0 to parts collect
                     (let ((location (/ x parts)))
                       (ramp 0 1 :type ramp))))
         (differences (mapcar #'(lambda (x y) 
                                  (when x (- x y)))
                        (rest list) list))
         (ideal-lengths (mapcar #'(lambda (x) (* x num)) differences)))
    (same-sum-integers ideal-lengths)))

;;takes a list of non-integers, rounds and adjusts so that they have the same total as the original
;;minus any remainder (which won't happen if called from divide-into-parts).
(defun same-sum-integers (list)
  (let* ((real-total (apply #'+ list))
         (floor-list (mapcar #'floor list))
         (floor-total (apply #'+ floor-list))
         (total-short (round (- real-total floor-total)))
         (add-to-each (/ total-short (length list)))
         (add (numerator add-to-each))
         (add-counter add)
         (per (denominator add-to-each))
         (per-counter per))
    (loop for n from 0 below (length floor-list) collect
          (let* ((item (nth n floor-list))
                 (output (if (> add-counter 0)
                             (progn (decf add-counter)
                               (1+ item))
                           item)))
            (decf per-counter)
            (when (zerop per-counter) 
              (setq per-counter per
                  add-counter add))
            output))))

;;gives number of segments not including *'s
(defun num-segments (list &optional (count 0))
  (cond ((= 1 (length list)) (1+ count))
        ((null list) count)
        (t
         (let ((second-item (second list))
               (third-item (third list)))
           (cond ((null third-item) (1+ count))
                 ((eql '* third-item) (num-segments (nthcdr 3 list) (1+ count)))
                 ((numberp second-item) (num-segments (rest list) (1+ count)))
                 (t (num-segments (nthcdr 2 list) (1+ count))))))))

(defun nth-segment (n list &optional (second-item nil))
  (if (<= n 0) 
      (if second-item (second list) (first list))
    (let ((item2(second list))
          (item3 (third list)))
      (cond ((null item3)
             (nth-segment (1- n) (nthcdr 2 list) second-item))
            ((eql '* item3)
             (nth-segment (1- n) (nthcdr 3 list) second-item))
            ((numberp item2) (nth-segment (1- n) (rest list) second-item))
            (t (nth-segment (1- n) (nthcdr 2 list) second-item))))))
        
          
 
          
        

        

#|removes * from front and back of list, duplicate **'s and any other innapropriate items
(correct-ramp-list '(* 1 2 (random 3) * 2 * * (not-a-listed-function 23) * 5 x 4 * 3 * *))
------> (1 2 (RANDOM 3) * 2 * 5 4 * 3)|#

(defun correct-ramp-list (list)
  (let ((list (reverse 
               (loop for n from 0 below (length list) with new-list = (reverse list) do
                     (let ((item (nth n new-list)))
                       (when (not (eql item '*))
                         (return (nthcdr n new-list))))))))
    (correct-ramp-list-f list)))

(defun correct-ramp-list-f (list &optional (front t) (asterisk nil))
  (let* ((item (first list)))
    (cond ((null list) nil)
      ((numberp item)
       (cons item (correct-ramp-list-f (rest list) nil)))
      ((listp item)
       (if (member (first item) (available-functions))
           (cons item (correct-ramp-list-f (rest list) nil))
         (correct-ramp-list-f (rest list) front asterisk)))
      ((eql item '*)
       (cond (front (correct-ramp-list-f (rest list)))
         (asterisk (correct-ramp-list-f (rest list) t t))
         (t (cons item (correct-ramp-list-f (rest list) nil t)))))
      (t (correct-ramp-list-f (rest list) front asterisk)))))
      

;;;;;;;;;;
;;;RAMP;;;
;;;;;;;;;;

(defun ramp (item1 item2 &key (type 'linear)(quantize '()))
  ;(format t "location ~S~%" location)
  (setq *function-error* 'ramp)
  (let ((item1 (get-data item1))
        (item2 (get-data item2)))
    (cond ((null item1)
           (if (null item2) nil item2))
          ((null item2)
           item1)
          (t
           (let* 
               ((quantize (get-data quantize))
                (output (cond ((zerop location) item1)
                              ((= 1 location) item2)
                              ((or (null type)
                                   (eql 'linear type))
                               (linear-ramp item1 item2))
                              ((eql type 'pch-linear) 
                               (midi-to-pch (round (ramp (pch-to-midi item1)(pch-to-midi item2) 'linear))))
                              ((eql type 'step)(if (< location .5) item1 item2))
                              ((eql type 'expt)
                               (expt-ramp item1 item2))
                              ((listp type)
                               (let ((args (butlast type))
                                     (type (first (last type))))
                                 (cond ((eql type 'expt)
                                        (expt-ramp item1 item2 args))
                                       (t (error "Ramp type ~S has not been implemented yet." type)))))
                              (t 
                               (progn (make-error 
                                       (format nil "RAMP: Type ~S has not been implemented. Using 1.~%" type))
                                 1)))))
             (if quantize (quantize-to-n quantize output)
               output))))))


;;;;;;;;;;;;;;;;;
;;;LINEAR-RAMP;;;
;;;;;;;;;;;;;;;;;

(defvar *linear-ramp-documentation* 
    "An even ramp using location to create 2 points.")

(defvar *linear-ramp-prompt*
    "(linear-ramp item1 item2)")

(defun linear-ramp (item1 item2)
  (setq *function-error* 'linear-ramp)
  (+ (* item1 (- 1 location))
     (* item2 location)))
    
;;;;;;;;;;;;;;;;
;;;EXPON-RAMP;;;
;;;;;;;;;;;;;;;;

(defvar *expt-amt* 2)

(defun expt-ramp (item1 item2 &optional expt-amt)
  (setq *function-error* 'expt-ramp)
  (let* ((expt-amt (if expt-amt expt-amt *expt-amt*))
         (expt-amt (if (listp expt-amt) (first expt-amt) expt-amt))
         (location (expt location expt-amt)))
    (+ (* item1 (- 1 location))
       (* item2 location))))


