;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;sequencer.lisp
;;;dec 2001
;;;textures thesis project

;;due to formatting problems with main-functions.lisp, I've divided it up into multiple files.

;;TO DO:  Fix sequencer...it's list formatting utilities, well, suck ass.
;;;(segment '(sequencer 2 '(1 2 3) :type 'cycle :cycles 2) 20)
;;;-------->(1 1 2 2 2 2 3 3 3 3 1 1 1 1 2 2 2 2 3 3) ???? should be '(1 1 2 2 3 3 1 1 2 2 3 3 1 1 2 2 3 3 1 1)
;;;fix shrink-distribution, it has no range checking and can possibly make instruments go below their real range..
;;;likewise stretch-distribution needs to be fixed because it might go above an instruments range...

;;also, some modularity has been sacrificed...the step sequencers now only acquire values once...

(in-package :cl-user)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;UTILITY FUNCTIONS;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;
;;;TRUNCATE-RHYTHM;;;
;;;;;;;;;;;;;;;;;;;;;

(defun truncate-rhythm (list beats &aux (beats (get-data beats)))
  (setq *function-error* 'truncate-rhythm)
  (if (null beats) list
    (do* ((beats (round beats))
          (temp nil)
          (sum 0 (apply #'+ temp))
          (cur-position cur-position (+ sum cur-position))
          (list (get-data list)))
         ((> sum beats) (reduce-total temp beats))
      (when (>= 0 beats) (progn (make-error "TRUNCATE: Beats is less than 0")(return list)))
      (if (listp list)
          (setq temp (append list temp))
        (setq temp (cons list temp))))))

(defun reduce-total (list n)
  (let ((sum (apply #'+ list)))
    (if (= sum n) list
      (let ((amt-to-add (- n sum))
            (last-value (first (last list))))
        (cond ((>= amt-to-add 0)
               (append (butlast list)
                       (list (+ (first (last list)) amt-to-add))))
              ((> last-value (abs amt-to-add))
               (append (butlast list)
                       (list (+ last-value amt-to-add))))
              ((= last-value (abs amt-to-add))
               (butlast list))
              (t 
               (append (reduce-total (butlast list) n))))))))

;;;;;;;;;;
;;;RIFF;;;
;;;;;;;;;;

(defparameter *riff-documentation*
  "Riff is like a special NTH function, because it evals it's parameters first.  In addition, it has the 
ability to effect an approximate 'mute.' If N is a list (i.e. '(15) for a 15 beat mute),
then RIFF returns a value dependent on CUR-PHRASE, which can effect a mute (or other preset data).")

(Defparameter *mute-rhythm* nil) ;;if non-nil, this will override the value input to mute.
(defparameter *mute-pitch* 7.0)
(defparameter *mute-duration* 0)
(defparameter *mute-dynamics* 0)
  
(defun riff (in list); &optional extract-length)
  (setq *function-error* 'riff)  
  (let* ((n (safe-eval in))
         (n (if (numberp n) 
                (round n) 
              n))
         (list (get-data list))
         (output    
          (cond ((listp n)
                 (let ((out (mute n)))
                   (if (eql out 'mark)
                       (return-from riff (riff in list))
                     out)))
                ((> n (length list)) (error "RIFF: ~s is not a valid index.~%" n))
                ((null list) nil)
                (t (nth (1- n) list)))) )
    (if (eql cur-phrase 'rhythm)
        (let* ((output (get-data output))  )
          (setq cur-beat (if (listp output)
                             (+ cur-beat (apply #'+ output))
                           (+ cur-beat output)))
          output)
      output)))


;;mute returns data which will effect a mute, using cur-phrase to determine what to output.
;;It also has the ability to act as a marker, something I will be expanding in the coming days.
;;Basically, it works like this; If cur-phrase is not 'rhythm, output a set value (i.e. *mute-pitch*, etc.).
;;If cur-phrase is RHYTHM, then if the incoming value is a number, use that as the value of the mute.
;;If the incoming value is a keyword, ':mute1, then store the value of cur-beat in a hash table as :MUTE1 and return 0.
;;If the incoming value is a symbol, 'mute1, then retrieve the value for symbol stored in the hash table and subtract
;;the value of cur-beat.  If this number is negative, signal an error.

;;Since this hash table is available to further calls, this can greatly simplify aligning different phrases.
;;Just make sure you set the marker with a :keyword before you try to access it.

(defvar *mute-hash* (make-hash-table :test 'eql))
(defun mute (&optional key)
  (let ((rhythm (first (first-list key))))
    (cond ((keywordp rhythm)
           (when (eql cur-phrase 'rhythm)
             ;(format t "~&~s = ~S~%" rhythm 
             (setf (gethash rhythm *mute-hash*) cur-beat))
           'mark) ;;This signifies that we set the mark, so get the next value.  Sequencer is buffered, so it ignores location.
          ((eql cur-phrase 'rhythm)
           (cond ((symbolp rhythm)
                  (let* ((output (or (- (gethash (add-colon rhythm) *mute-hash*) cur-beat)
                                     (error "No marker ~S in *mute-hash*" rhythm)))
                         (formula (subst output (list rhythm) key :test 'equal))
                         (out (get-data formula))
                         (out (if (numberp out) out (first out))))
                    (if (< out 1) 
                        (progn
                          (make-error "Marker accessed too late, ignoring.")
                          1)
                      ; (progn
                      ;   (format t "~&Access ~S from ~S = ~S~%" rhythm cur-beat output)
                      out)))
                 (*mute-rhythm* *mute-rhythm*)
                 (t (round rhythm))))
          ((eql cur-phrase 'pitch)
           *mute-pitch*)
          ((eql cur-phrase 'durations)
           *mute-duration*)
          ((eql cur-phrase 'dynamics)
           *mute-dynamics*)
          (t 0))))

(defun first-list (l)
  (when (listp l)
    (or (loop for n from 0 below (length l) do
              (when (listp (nth n l)) 
                (return (nth n l))))
        l)))

;;retrieves the value for key :VAR in *mute-hash*
;;Use this to retrieve values outside of modular-riff, i.e. clock-offset values.  Will not store a value.
(defun get-marker (var)
  (or (gethash (add-colon var) *mute-hash*) (format t "No marker ~S~%" var)))

;;;;;;;;;;;;;;;
;;;SEQUENCER;;;
;;;;;;;;;;;;;;;

(defun sequencer (buffer-ref list &key (type 'step) (cycles 1) (motion 'forward) (start-pos '0)
                             (reset-location 'default-t) print)
  (let* ((cycles (get-data cycles))
         (start-pos (get-data start-pos))
         (reset-location (get-data reset-location))
         (output  (sequencer-f buffer-ref list type cycles motion start-pos reset-location)))
  (when print (format t "~&Cur-phrase: ~A, output: ~A~%" cur-phrase output))
  output))

(defgeneric sequencer-f (buffer-ref list type cycles motion start-pos reset-location))

(defmethod sequencer-f (buffer-ref list (type (eql 'cycle)) cycles motion start-pos reset-location) 
  (setq *function-error* 'sequencer-cycle)
  (let* ((list (get-data list))
         (list (cond ((eql motion 'forward) list)
                 ((eql motion 'backward) (reverse list))
                 ((eql motion 'bounce) (append list (reverse list)))
                     (t (progn 
                          (make-error 
                           (format nil "Sorry, motion type '~S has not been implemented yet.  Setting to 'forward.~%" motion))
                     list))))
         (list (rotate-list list start-pos))
         (cycles (cond ((zerop cycles)
                        (when (zerop location) (make-error "Can't have 0 cycles, setting to 1."))
                        1)
                   ((numberp cycles) cycles)
                   (t (make-error
                       (format nil "Sorry, a cycles value of '~S has not been implemented yet.~%" cycles)))))
         (list (repeat-n-times (1- cycles) list))
         (num-items (length list))
         (item (let ((n (round
                         (* (1- num-items)  ;;gotta find a better way...edges de-emphasized, i.e.
                           location)))) ;;(1 2 2 3 3 4)...not right...should be even..
                 (nth n list)))
         (reset-location (if (eq reset-location 'default-t) nil reset-location))
         (location (if reset-location 
                       (make-location num-items)
                     location)))
    (get-data item)))

(defmethod sequencer-f (buffer-ref list (type (eql 'step)) cycles motion start-pos reset-location
                         &aux (reset-location (if (eq reset-location 'default-t) nil reset-location)))
  (setq *function-error* 'sequencer-step)
  (buffer buffer-ref `'(sequencer-step ',list ,cycles ',motion ,start-pos ,reset-location) 
    :storage-length num-attacks :output 'single))

(defun sequencer-step (list cycles motion start-pos reset-location)
  (let* ((list (get-data list))
         (list (rotate-list list start-pos))
         (cycles (if (zerop cycles) 
                     (progn (make-error "SEQUENCER: Can't have 0 cycles.  Setting to 1.")
                       1)
                   cycles))
         (list (mapcan #'(lambda (x) (repeat-n-times (1- cycles) (list x))) list))
         (list (cond ((eql motion 'forward) list)
                     ((eql motion 'backward) (reverse list))
                     ((eql motion 'bounce) (append list (reverse list)))
                     (t 
                      (make-error 
                       (format nil "Sorry, motion type '~S has not been implemented yet.  Setting to 'forward.~%" motion))
                      list))))
    list))

(defmethod sequencer-f (buffer-ref list (type (eql 'step-hold)) cycles motion start-pos reset-location
                         &aux (reset-location (if (eq reset-location 'default-t) nil reset-location)))
  (setq *function-error* 'sequencer-step-hold)
  (buffer buffer-ref `'(sequencer-step-hold ',list ,cycles ',motion ,start-pos ,reset-location)
    :storage-length num-attacks :output 'single))

(defun sequencer-step-hold (list cycles motion start-pos reset-location)
  (let* ((list (get-data list))
         (list (rotate-list list start-pos))
         (list (cond ((eql motion 'forward) list)
                     ((eql motion 'backward) (reverse list))
                     ((eql motion 'bounce) (append list (reverse list)))
                     (t (progn
                          (make-error 
                           (format nil "Sorry, motion type '~S has not been implemented yet.  Setting to 'forward.~%" motion))
                          list))))
         (list (check-and-modify-length list num-attacks))
         (cycles (if (zerop cycles) (progn (when (zerop location)
                                             (make-error "SEQUENCER: Can't have 0 cycles.  Setting to 1."))
                                      1)
                   cycles))
         (list (apply #'append (mapcar #'(lambda (x) (repeat-n-times (1- cycles) (list x))) list))))
        list))


;;n changed from floor to round
(defun sequencer-segment (list cycles reset-location)
  (let* ((n (round (* (1- (length list))
                     location)))
         (item (nth n list))
         (location (if reset-location 
                       (make-location (* 2 cycles))
                     location)))
    (get-data item)))

(defmethod sequencer-f (buffer-ref list (type (eql 'pattern)) cycles motion start-pos reset-location)
  (format t "Sorry, sequencer type 'pattern has not yet been implemented.~%"))

(defun make-location (num-segments)
  (setq *function-error* 'make-location)
  (rem (* num-segments location) 1))



;;;;;;;;;;;;
;;;BUFFER;;;
;;;;;;;;;;;;

;;this needs to be cleaned up a bit...but buffer has been updated...

(defun buffer (buffer-ref function &key (output 'single)(storage-length '1.0)(output-length 1)(repeat 0)reset-location)                     
  (setq *function-error* 'buffer)
  (let* ((storage-length (get-data storage-length))
         (repeat (get-data repeat))
         (repeat (if (fixnump repeat) repeat
                   (* num-attacks repeat)))
         (reset-location (get-data reset-location))
         (output (buffer-f buffer-ref function output output-length storage-length repeat reset-location)))
    output))
  
(defgeneric buffer-f (buffer-ref function output output-length storage-length repeat reset-location))

(defmethod buffer-f (buffer-ref function (output (eql 'single)) output-length storage-length repeat reset-location)
  (setf *function-error* 'buffer-f-single)
  (let ((storage-length (if (fixnump storage-length) storage-length
                          (round (* storage-length num-attacks))))
        (item (first (output-and-adjust-buffer buffer-ref 1))))
    (if item item
      (let ((data (repeat-n-times repeat (if reset-location (location-loop function storage-length t)
                                             (location-loop function storage-length t cur-position)))))
        (setf (gethash buffer-ref *buffer-hash*) (rest data))
        (first data)))))
      

(defmethod buffer-f (buffer-ref function (ouptut (eql 'list)) output-length storage-length repeat reset-location)
  (setf *function-error* 'buffer-f-list)
  (if reset-location 
      (let (;(storage-length (if (fixnump storage-length) storage-length   ;;;why is this here??
             ;                  (round (* storage-length num-attacks))))
             (data (repeat-n-times repeat
                                   (location-loop function output-length t))))
        data)
    (let* ((storage-length (if (fixnump storage-length) storage-length
                             (round (* storage-length num-attacks))))
           (item (output-and-adjust-buffer buffer-ref output-length))
           (item-length (length item))  ;;might not have enough in buffer
           (start-position (if (> cur-position num-attacks) num-attacks cur-position))
           (end-position (if (> (+ cur-position storage-length) num-attacks) num-attacks (+ cur-position storage-length)))
           (data
            (if (= item-length output-length)
                item
              (append item 
                      (repeat-n-times repeat 
                                      (if reset-location 
                                          (location-loop function output-length t)
                                        (location-loop function num-attacks t start-position end-position)))))))
      (when (< item-length output-length) (setf (gethash buffer-ref *buffer-hash*) (nthcdr output-length data)))
      (when *debug* (format t "~&data ~S hash ~S output-length ~S~%" data (gethash buffer-ref *buffer-hash*) output-length))
      (first-n output-length data))))       

(defun output-and-adjust-buffer (buffer-ref length)
  "Remove's up to LENGTH items from current buffer-ref."
  (setq *function-error* 'output-and-adjust-buffer)
  (let ((item (gethash buffer-ref *buffer-hash*)))
    (setf (gethash buffer-ref *buffer-hash*) (nthcdr length item))
    (when item (first-n length item))))



;;;;;;;;;;;;;;;;
;;;DISTRIBUTE;;;
;;;;;;;;;;;;;;;;

(defparameter *distribute-documentation*
  "Distribute is a very powerful, relatively simple and fast function that takes a number of realized
phrases and distribute them to the given instrumentation.  In other words, it can orchestrate a number
of phrases based on range and attack preferences.  It can also be used to combine a set of phrases into
a single new phrase, if a set of INPUT-PHRASES is larger than it's corresponding set of OUTPUT-INSTRUMENTS.")

#|  A description of options
INPUT-PHRASES-LIST: a list of lists, each list contains phrases that should be distributed together
NUM-ATTACKS-LIST: a single list of num-attacks relating to each individual input-phrase when CREATING
CLOCK-OFFSET-LIST: a single list of clock-offsets for each LIST input-phrase groups when CREATING
OUTPUT-SURNAMES: a single list off surnames for each LIST of input-phrase groups
OUTPUT-INSTRUMENTS-LIST: a list of lists, each list containing a list of instruments for each
list in INPUT-PHRASES to be distributed to
INSTRUMENT-DISTRIBUTION: 'even (default), 'random or 'random2.  EVEN distributes the ranges evenly, random is random, 
random2 is like random, but each instrument gets 2 notes before the next instrument is randomly found.
NOTE-PREFERENCE: method for deciding how to place notes in an individual instrument.  
- 'ATTACK says that attacks are more important than durations, so a new note can be placed on top
of the sustain for an old note.
- 'DURATION says that a new note will not be placed on top of the sustain for an old note, however
the sustain of the new note might be cut off by an attack already present.
TEMPO-LIST: is a single list of tempo's relating to each distribution group 
CREATING: decides whether to newly realize INPUT-PHRASES with NUM-ATTACKS-LIST and CLOCK-OFFSET-LIST,
or whether to skip this portion and use already created phrases (for changing instrument distributions, etc.)
This option defaults to the global value of CREATING, which is T if DISTRIBUTED-SECTION is called from within
a call to CREATE, i.e. (create 'distribute-1)  

To Do: change output-names to output-surnames, to be tacked onto instruments.
Make output-instruments a list of lists, a list for each input-phrase.
|#

(defun distributed-section (&key input-phrases-list num-attacks-list clock-offset-list 
                                 output-surnames output-instruments-list (creating creating)
                                 (instrument-distribution-list 'even)(note-preference-list 'attack)
                                 (tempo-list *tempo*))
  (setq *function-error* 'distributed-section)
  (when (or creating  ;;if creating, set these up and create each phrase in list
            (if (member nil (mapcar #'find-associated-phrase (apply #'append input-phrases-list)))
                (progn (format t "~&All input-phrases must be created before they can be distributed.~%")
                  T)
              nil))
    (let* ((input-phrases (apply #'append input-phrases-list)) 
           (num-attacks-list (check-and-modify-length num-attacks-list input-phrases))
           (clock-offset-list (check-and-modify-length clock-offset-list input-phrases-list))
           (clock-offset-listing (loop for n from 0 below (length input-phrases-list) append
                                       (check-and-modify-length (list (nth n clock-offset-list))
                                                                (nth n input-phrases-list))))
           (tempo-list (check-and-modify-length tempo-list input-phrases-list))
           (tempo-list (mapcar #'check-and-modify-length tempo-list input-phrases-list))
           (tempo-listing (apply #'append tempo-list)))
      (mapcar #'(lambda (input attacks clock tempo)
                  (create input (get-data attacks) (get-data clock) (get-data tempo)))
        input-phrases num-attacks-list clock-offset-listing tempo-listing)))
  (format t "~&Distributing: ")
  (let* ((input-phrases-list (check-and-modify-length input-phrases-list output-surnames))
         (input-phrases (mapcar #'(lambda (phrases)
                                    (loop for n from 0 below (length phrases) collect
                                          (find-associated-phrase (nth n phrases))))
                          input-phrases-list))
         (output-instruments-list (check-and-modify-length output-instruments-list output-surnames))
         (output-names (mapcar #'(lambda (instrument-list surname)
                                   (loop for n from 0 below (length instrument-list) collect
                                         (read-from-string (concatenate 'string 
                                                             (princ-to-string (nth n instrument-list))
                                                             (princ-to-string surname)))))
                         output-instruments-list output-surnames))
         (note-preference-list (check-and-modify-length note-preference-list output-surnames))
         (instrument-distribution-list (check-and-modify-length instrument-distribution-list output-surnames)))
    (loop for n from 0 below (length input-phrases) do
          (let* ((phrase-group (nth n input-phrases))
                 (clock-offset-list (mapcar #':clock-offset phrase-group))           
                 (clock-offset (first clock-offset-list))
                 (max-offset (apply #'max clock-offset-list))
                 (max-length (+ max-offset (last-beat phrase-group nil))) ;;array size
                 (tempo (:tempo (first phrase-group)))
                 (output-phrases 
                  (mapcar #'(lambda (name instrument) 
                              (make-phrase name instrument max-length))
                    (nth n output-names) (nth n output-instruments-list)))
                 (note-preference (nth n note-preference-list))
                 (instrument-distribution (nth n instrument-distribution-list))
                 (distribute-phrases (make-distribute-phrases output-phrases)))
            (format t "~A " (nth n output-surnames))
            (loop for o from 0 below (length phrase-group) do
                  (let* ((input-phrase (nth o phrase-group))
                         (offset (nth o clock-offset-list))
                         (offset (when (> offset 0) offset))
                         (rhythm (if offset 
                                     (cons offset (:rhythm input-phrase))
                                   (:rhythm input-phrase)))
                         (rhythm-length (num-attacks rhythm))
                         (duration (check-and-modify-length
                                    (if offset
                                        (cons offset (:duration input-phrase))
                                      (:duration input-phrase)) rhythm-length))
                         (rhythm-durations (combine-rhythm-duration rhythm duration))
                         (pitch (check-and-modify-length 
                                 (if offset
                                     (cons 7.0 (:pitch input-phrase))
                                   (:pitch input-phrase))
                                 rhythm-length))
                         (dynamics (check-and-modify-length 
                                    (if offset 
                                        (cons 0 (:dynamics input-phrase)) 
                                      (:dynamics input-phrase))
                                    rhythm-length)))
                    (setq *random2* nil)
                    (distribute instrument-distribution note-preference distribute-phrases rhythm-durations 
                                pitch dynamics)))
            (reduce-phrase-vectors output-phrases)))
    (format t "~%~S DISTRIBUTE'd to ~S~%" name output-names)))

;;uses nth-rhythm to get everything from rhythm-durations until the next attack and passes this on
(defun distribute (instrument-distribution note-preference distribute-phrases rhythm-durations pitches dynamics)
  (setq *function-error* 'distrubute)
  (loop with clock = 0 and n = 0 and rhythm-durations-length = (length rhythm-durations)
        do
        (let ((rhythm-duration (nth-rhythm clock rhythm-durations))
              (pitch (nth n pitches))
              (dynamic (nth n dynamics)))
          (distribute-single instrument-distribution note-preference pitch clock distribute-phrases 
                             rhythm-duration dynamic)
          (setq clock (+ clock (length rhythm-duration))
              n (1+ n))
          (when (>= clock rhythm-durations-length) (return)))))  
       
;;figures out which phrase is in the proper range and does not have an attack on the current beat
(defun distribute-single (instrument-distribution note-preference pitch clock distribute-phrases rhythm-duration 
                                                  dynamic)
  (setq *function-error* 'distribute-single)
  (let* ((possible-phrases 
          (loop for n from 0 below (length distribute-phrases) do
                (let ((item (nth n distribute-phrases)))
                  (when (in-range (cons (first item) 
                                        (list (second item))) 
                                  pitch)
                    (return (end item))))))
         (possible-phrases 
          (cond ((eql instrument-distribution 'random2)
                 (if *random2* 
                     (prog1 *random2* (nullify *random2*))
                   (setq *random2* (random-reorder possible-phrases))))
                ((eql instrument-distribution 'random)
                 (random-reorder possible-phrases))
                ((eql instrument-distribution 'even)
                 possible-phrases)
                (t (progn (make-error 
                           (format nil "DISTRIBUTE-SINGLE: Unknown instrument-distribution ~S." instrument-distribution))
                     possible-phrases)))))
    (if possible-phrases
        (cond ((eql note-preference 'attack)
               (loop for n from 0 below (length possible-phrases) doing
                     (let* ((phrase (nth n possible-phrases))
                            (phrase-rhythm 
                             (aref (:rhythm phrase) clock)))
                       (when (or (not (numberp phrase-rhythm))
                                 (zerop phrase-rhythm))
                         (return (distribute-items phrase clock pitch rhythm-duration dynamic))))))
              ((eql note-preference 'duration)
               (loop for n from 0 below (length possible-phrases) doing
                     (let* ((phrase (nth n possible-phrases))
                            (phrase-rhythm 
                             (aref (:rhythm phrase) clock)))
                       (when (and (numberp phrase-rhythm) (zerop phrase-rhythm))
                         (return (distribute-items phrase clock pitch rhythm-duration dynamic))))))
              (t (error "DISTRIBUTE: Sorry, unknown note preference ~S." note-preference)))
      0)))

;;finally, distributes each element to their proper places (if any)
(defun distribute-items (phrase clock pitch rhythm-duration dynamic)
  (setq *function-error* 'distribute-items)
  (setf (aref (:pitch phrase) clock) pitch
    (aref (:dynamics phrase) clock) dynamic)
  (let ((rhythm (:rhythm phrase))
        (length-rhythm-duration (length rhythm-duration))
        (n 0))
    (loop do         ;;with n=0 DOES NOT WORK!  If distributing more than 1 phrase, n is not reassigned!
          (let ((compare-rhythm (aref rhythm (+ clock n))))
            (setf (aref rhythm (+ clock n)) (nth n rhythm-duration))
            (when (or (= n (1- length-rhythm-duration))
                      (non-zero-nump compare-rhythm))
              (return (progn (setf (:rhythm phrase) rhythm) n)))
            (incf n)))))

;;when distributing, each phrase element is turned into a vector, this function
;;makes them into lists again by removing nil.
(defun reduce-phrase-vectors (phrases)
  (setq *utility-error* 'reduce-phrase-vectors)
  (loop for n from 0 below (length phrases) doing
        (let* ((phrase (nth n phrases))
               (rhythm-durations (reduce-vector (:rhythm phrase)))
               (clock-offset (+ (:clock-offset phrase)
                                (initial-zeros rhythm-durations))))
          (setf (:clock-offset phrase) clock-offset
            (:rhythm phrase) (extract-rhythm rhythm-durations)
            (:duration phrase) (extract-durations rhythm-durations)
            (:pitch phrase) (reduce-vector (:pitch phrase))
            (:dynamics phrase) (reduce-vector (:dynamics phrase))))))

;;returns a list starting with current position ending before the next attack (non-zero number)
(defun nth-rhythm (n rhythm-durations)
  (cons (nth n rhythm-durations)
        (do* ((clock (1+ n) (1+ clock))
              (item (nth clock rhythm-durations)(nth clock rhythm-durations))
              (temp nil))
             ((or (null item)
                  (and (numberp item)
                       (not (zerop item))))
              (return (reverse temp)))
          (push item temp))))
               

;;makes a list of ranges with a list of ordered instruments who can play that range.
;;works by making calling SINGLE-DISTRIBUTION, removing the instrument-specific ranges
;;used in that distribution, then calling again until all ranges are used up.

(defun make-distribute-phrases (phrases)
  (setq *function-error* 'make-distribute-phrases)
  (let* ((phrases (stable-sort (copy-seq phrases) #'< :key #'(lambda (x)(:range-hi (eval (:instrument x))))))
         (instruments (mapcar #'(lambda (x) (eval (:instrument x))) phrases))
         (ranges (mapcar #'(lambda (instrument)
                             (list (pch-to-midi (:range-lo instrument))
                                   (pch-to-midi (:range-hi instrument))))
                   instruments))
         (primary-distribution (single-distribution phrases ranges))
         (phrase-ranges (remove-used-ranges primary-distribution (mapcar #'list phrases ranges))))
    (do (distribution)
        ((null phrase-ranges)
         (return (consolidate-distribution primary-distribution distribution)))
      (let* ((phrases (mapcar #'first phrase-ranges))
             (ranges (mapcar #'second phrase-ranges))
             (new-distribution (single-distribution phrases ranges)))
        (setq phrase-ranges (remove-used-ranges new-distribution phrase-ranges)
            distribution (append distribution new-distribution))))))

;;same as make-distribute-phrases, but returns instrument names instead of phrase objects, for testing.
;;generates a primary distribution, decides which ranges weren't used in primary, then sends them back until
;;everything has been used.
(defun distribute-instruments (instruments)
  (setq *function-error* 'distribute-instruments*)
  (let* ((instruments (stable-sort (copy-seq instruments) #'< :key #'(lambda (x) (:range-hi (eval x)))))
         (instrument-objs (mapcar #'eval instruments))
         (ranges (mapcar #'(lambda (instrument)
                             (list (pch-to-midi (:range-lo instrument))
                                   (pch-to-midi (:range-hi instrument))))
                   instrument-objs))
         (primary-distribution (single-distribution instruments ranges))
         (instrument-ranges (remove-used-ranges primary-distribution (mapcar #'list instruments ranges))))
    (do (distribution)
        ((null instrument-ranges)
         (return (consolidate-distribution primary-distribution distribution)))
      (let* ((instruments (mapcar #'first instrument-ranges))
             (ranges (mapcar #'second instrument-ranges))
             (new-distribution (single-distribution instruments ranges)))
        (setq instrument-ranges (remove-used-ranges new-distribution instrument-ranges)
            distribution (append distribution new-distribution))))))
    
         
    
;;;starts with the primary distribution, compares to each item in distribution by calling ADJUST-DISTRIBUTION.
;;;It finished up by converting everything back to pch...
(defun consolidate-distribution (consolidated-distribution distribution)
  (setq *function-error* 'consolidate-distribution)
  (mapcar #'(lambda (group)
              (list (midi-to-pch (first group))(midi-to-pch (second group))(third group)))
    (condense-distribution
     (stable-sort 
      (loop for n from 0 to (length distribution) do
            (if (= n (length distribution)) (return consolidated-distribution)
              (setq consolidated-distribution 
                    (adjust-distribution consolidated-distribution 
                                         (nth n distribution)))))
      #'< :key #'second))))

;;assumes the list is sorted and correct, then searches for adjacent, EQUAL phrase-lists
;;and combines the ranges of the two into one.
;;(condense-distribution '((1 4 (marimba vibraphone))(5 20 (marimba vibraphone))(21 30 (marimba vibraphone))
;;                         (31 40 (marimba))))
;;-------> ((5 30 (marimba vibraphone))(31 40 (marimba)))


(defun condense-distribution (distribution)
  (if (<= (length distribution) 1) distribution  ;;if there is one thing or none, no more work to do.       
    (let* ((d1 (first distribution))
           (d2 (second distribution))
           (range-lo1 (first d1))
           (range-hi1 (second d1))
           (phrases1 (third d1))
           (range-lo2 (first d2))
           (range-hi2 (second d2))
           (phrases2 (third d2)))
      (if (equal phrases1 phrases2)  ;;if equal, condense to 1 and include in recursive call
          (condense-distribution (cons (list range-lo1 range-hi2 phrases1)
                                       (nthcdr 2 distribution)))
        (cons d1 (condense-distribution (rest distribution))))))) ;;otherwise recurse on rest
                
          


;;recursively checks distribution against each of instrument-ranges, removing any overlap from 
;;instrument-ranges and returns a list of lists of the unsed ranges.
(defun remove-used-ranges (distribution instrument-ranges)
  (setq *function-error* 'remove-used-ranges)
  (if (null distribution) 
      (sort instrument-ranges #'< :key #'caadr)
    (let* ((cur-distribution (first distribution))
           (range-lo (first cur-distribution))
           (range-hi (second cur-distribution))
           (instruments (third cur-distribution)))
      (loop for n from 0 below (length instrument-ranges) do
            (let ((instrument-range (nth n instrument-ranges)))
              (when (member (first instrument-range) instruments :test #'eql) 
                (setf (nth n instrument-ranges) nil
                  instrument-ranges (append (remove-range range-lo range-hi instrument-range)
                                            instrument-ranges)))))
      (remove-used-ranges (rest distribution) (remove nil instrument-ranges)))))


;;if there is overlap with input range and instrument-range, the overlap is removed.
;;returns a list of lists, since sometimes the overlap is in the middle of a range.
(defun remove-range (range-lo range-hi instrument-range)
  (setq *function-error* 'remove-range)
  (let ((instrument-range-lo (first (second instrument-range)))
        (instrument-range-hi (second (second instrument-range)))
        (instrument (first instrument-range)))
    (cond ((or (> instrument-range-lo range-hi)  
               (< instrument-range-hi range-lo))
           (list instrument-range))
          ((<= range-lo instrument-range-lo instrument-range-hi range-hi) ;;if instrument is contained.
           nil)
          ((> instrument-range-hi range-hi)
           (cons (list instrument (list (1+ range-hi) instrument-range-hi))
                 (remove-range range-lo range-hi (list instrument (list instrument-range-lo range-hi)))))
          (t
           (cons (list instrument (list instrument-range-lo (1- range-lo))) 
                 (remove-range range-lo range-hi (list instrument (list range-lo instrument-range-hi))))))))


    
#|Returns a list of instruments ranges followed by a single instrument in it's primary range.
This is easy if there is no overlap, but here is the algorithm for the rest:
Calculate average-range-amt, attempt to give each instr if possible.
start from bottom, give first instrument average-range-amt of range, continue to top.  If an instrument
is unable to carry it's full given range, adjust-single-distribution  is called, average-range-amt is recalculated, 
and we continue.  Voila!|#


;;instruments and ranges must correspond and be sorted from lowest to highest
(defun single-distribution (instruments ranges)
  (setq *function-error* 'single-distribution)
  (let ((gaps (add-to-end (find-gaps ranges) (1- (length instruments))))
        (initial-gap 0))
    (loop for o from 0 below (length gaps) append
          (let* ((cur-gap (nth o gaps))
                 (current-range-lo (first (nth initial-gap ranges)))  ;;range-lo is lowest note of group
                 (highest-note (second (nth cur-gap ranges)))
                 (instruments-length (- cur-gap initial-gap -1))
                 (distribution nil)
                 (top-of-gap (when (< o (1- (length gaps)))
                               (1- (first (nth (1+ cur-gap) ranges)))))
                 (gap-distribution (when (< o (1- (length gaps)))
                                     (if (= highest-note top-of-gap)
                                         (list highest-note highest-note '(nil))
                                       (list (1+ highest-note) top-of-gap '(nil))))))
            (loop for n from initial-gap to cur-gap do
                  (let* ((instrument (nth n instruments))
                         (range (nth n ranges))
                         (instrument-range-lo (first range))
                         (instrument-range-hi (second range))
                         (instruments-left (- cur-gap n -1))
                         (total-range-left (- highest-note (max instrument-range-lo current-range-lo)))
                         (avg-range (round (/ total-range-left instruments-left)))
                         (current-range-hi (min instrument-range-hi (+ current-range-lo avg-range))))
                    (cond ((and (in-range range current-range-lo)
                                (in-range range current-range-hi))
                           (setq distribution (append distribution
                                                      (list (list current-range-lo current-range-hi (list instrument)))))
                           (setq current-range-lo (1+ current-range-hi)))
                          ((and (= instrument-range-lo current-range-lo)
                                (< instrument-range-hi current-range-hi))
                           (push (list current-range-lo instrument-range-hi (list instrument)) distribution)
                           (setq current-range-lo (1+ instrument-range-hi)))
                          (t
                           (setq distribution (adjust-single-distribution 
                                               distribution avg-range instrument instrument-range-lo 
                                               instrument-range-hi current-range-lo current-range-hi))
                           (setq current-range-lo (1+ (second (end distribution))))))))
            (when gap-distribution (push gap-distribution distribution))
            (setq initial-gap (1+ cur-gap))
            (if (zerop o) (reverse distribution)
              distribution)))))


;;returns a list of where the gaps start. (begins at 0)
;;(find-gaps '((0 10) (8 19) (20 30) (40 50) (45 60) (80 90)))
;;-------------->(2 4)
(defun find-gaps (ranges &optional (counter 0))
  (setq *function-error* 'find-gaps)
  (if (null (rest ranges)) nil
    (let ((range1-top (second (first ranges)))
          (range2-bottom (first (second ranges))))
      (if (> range2-bottom (1+ range1-top)) ;;'((0 9) (10 20)) is not a gap
          (cons counter (find-gaps (rest ranges) (1+ counter)))
        (find-gaps (rest ranges) (1+ counter))))))
      

(defun adjust-single-distribution (distribution avg-range instrument instrument-range-lo instrument-range-hi
                                                 current-range-lo current-range-hi)                                                 
  (setq *function-error* 'adjust-single-distribution)
  (if (> instrument-range-lo current-range-lo)
      (stretch-distribution distribution avg-range instrument instrument-range-lo instrument-range-hi
                            current-range-hi)                            
    (shrink-distribution distribution instrument instrument-range-lo instrument-range-hi)))

(defun shrink-distribution (distribution new-instrument new-instrument-range-lo new-instrument-range-hi)
  (setq *function-error* 'shrink-distribution)
  (if (>= (second (end distribution)) new-instrument-range-hi)
      distribution
    (let* ((distribution (sort distribution #'< :key #'first))
           (range-lo-group (first (first distribution)))
           (distribution (append distribution (list (list new-instrument-range-lo new-instrument-range-hi 
                                                          (list new-instrument)))))
           (distribution-length (length distribution)))
      (loop for n from 0 below distribution-length do
            (let* ((distribute-item (nth n distribution))
                   (total-range (- new-instrument-range-hi range-lo-group))
                   (avg-range (round (/ total-range (- distribution-length n))))
                   (new-range-hi (min new-instrument-range-hi (+ avg-range range-lo-group))))
              (setf (first distribute-item) range-lo-group
                (second distribute-item) new-range-hi
                (nth n distribution) distribute-item
                range-lo-group (1+ new-range-hi))))
      distribution)))
  
                

;;we assume their are no gaps...gaps are dealt with in single-distribution
(defun stretch-distribution (distribution avg-range new-instrument new-instrument-range-lo new-instrument-range-hi
                                          current-range-hi)
  (setq *function-error* 'stretch-distribution)
  (let* ((distribution (sort distribution #'< :key #'first))
         (range-lo-group (first (first distribution)))
         (range-hi (+ avg-range new-instrument-range-lo))
         (new-range-hi (min new-instrument-range-hi (round (/ (+ range-hi current-range-hi) 2))))
         (distribution-addition (list (list new-instrument-range-lo new-range-hi
                                            (list new-instrument))))
         (distribution-length (length distribution)))
    (loop for n from 0 below distribution-length do
          (let* ((distribute-item (nth n distribution))
                 (total-range (- new-instrument-range-lo range-lo-group 1))
                 (avg-range (round (/ total-range (- distribution-length n))))
                 (new-range-hi (min new-instrument-range-hi (+ avg-range range-lo-group))))
            (setf (first distribute-item) range-lo-group
              (second distribute-item) new-range-hi
              (nth n distribution) distribute-item
              range-lo-group (1+ new-range-hi))))
    (append distribution distribution-addition)))
                 
         
(defun adjust-distribution (distribution-group new-distribution)
  (setq *function-error* 'adjust-distribution)
  (loop for n from 0 below (length distribution-group) append
        (let ((old-distribution (nth n distribution-group)))
          (setf (nth n distribution-group)
            (compare-distribution old-distribution new-distribution)))))


#|Compare-distribution compares two sets of ranges, with 13 possible situations.  See graphics
below for a representation of possible states.  The first bar represents range-new, and the second 
represents range-old.  Note that in 3) and 4) the range-hi of one is the same as the range-lo of the 
other, 5) and 6) have matching range-lo's. We start our distribution with the full range of the
ensemble, so we use the range-old as a boundary, i.e., if range-new extends farther than range-old, 
range-new is truncated.  

1) |    2)   |   3) |    4)   |  5) |   6)   |  7) | |   8) | |  9) | | 
                      |     |       |        |     | |        |     | 
     |     |          |     |       | |    | |     | |        |     | 


 10) |   11)   | 12)   |  13) |                                                              
     | |     | |     | |      | |                                                                 
       |       |     |        |       

Phew!|#

(defun compare-distribution (old-distribution new-distribution)
  (setq *function-error* 'compare-distribution)
  (if (member (first (end new-distribution)) (end old-distribution))
      (list old-distribution)
    (let* ((range-lo-old (first old-distribution))
           (range-hi-old (second old-distribution))
           (abs-range-old (- range-hi-old range-lo-old))
           (phrases-old (first (last old-distribution)))
           (range-lo-new (first new-distribution))
           (range-hi-new (second new-distribution))
           (phrases-new (first (last new-distribution))))
      (cond ((or (> range-lo-new range-hi-old)
                 (> range-lo-old range-hi-new)
                 (null (first phrases-new)))
             (list old-distribution))
            ((= range-lo-old range-hi-old)
             (list (list range-lo-old range-hi-old (append phrases-old phrases-new))))                                               
            ((= range-lo-new range-hi-old) 
             (list (list range-lo-old (1- range-hi-old) phrases-old)
                   (list range-hi-old range-hi-old (append phrases-old phrases-new))))
            ((= range-lo-old range-hi-new)
             (list (list (1+ range-lo-old) range-hi-old phrases-old)
                   (list range-lo-old range-lo-old (append phrases-old phrases-new))))
            ((> range-hi-new range-hi-old)
             (compare-distribution old-distribution (list range-lo-new range-hi-old phrases-new)))                                         
            ((< range-hi-new range-hi-old)
             (append (compare-distribution (list range-lo-old range-hi-new phrases-old)
                                               (list range-lo-new range-hi-new phrases-new))
                     (list (list (1+ range-hi-new) range-hi-old phrases-old))))
            ((< range-lo-new range-lo-old)
             (compare-distribution old-distribution (list range-lo-old range-hi-new phrases-new)))                                          
            ((> range-lo-new range-lo-old)
             (cons (list range-lo-old (1- range-lo-new) phrases-old)
                   (compare-distribution (list range-lo-new range-hi-old phrases-old)
                                               (list range-lo-new range-hi-new phrases-new))))                                      
            ((null phrases-old) (list (list range-lo-old range-hi-old phrases-new)))
            (t (list (list range-lo-old range-hi-old (append phrases-old phrases-new))))))))

;;;(defun adjust-single-distribution (algorithm old-distribution new-distribution &optional (new-real 'both))
;;;  (setq *function-error* 'adjust-single-distribution-hi)
;;;  (when old-distribution
;;;    (let* ((range-lo-old (first old-distribution))
;;;           (range-hi-old (second old-distribution))
;;;           (abs-range-old (- range-hi-old range-lo-old))
;;;           (phrases-old (first (last old-distribution)))
;;;           (range-lo-new (first new-distribution))
;;;           (range-hi-new (second new-distribution))
;;;           (phrases-new (first (last new-distribution))))
;;;      (cond ((or (> range-lo-new range-hi-old)
;;;                 (> range-lo-old range-hi-new))
;;;             (list old-distribution))
;;;            ((= range-lo-old range-hi-old)
;;;             (list (list range-lo-old range-hi-old (append phrases-old phrases-new))))                                               
;;;            ((= range-lo-new range-hi-old) 
;;;             (list (list range-lo-old (1- range-hi-old) phrases-old)
;;;                   (list range-hi-old range-hi-old (append phrases-new phrases-old))))
;;;            ((= range-lo-old range-hi-new)
;;;             (list (list (1+ range-lo-old) range-hi-old phrases-old)
;;;                   (list range-lo-old range-lo-old (append phrases-old phrases-new))))
;;;            ((> range-hi-new range-hi-old)
;;;             (adjust-single-distribution algorithm old-distribution
;;;                                         (list range-lo-new range-hi-old phrases-new) 
;;;                                         'bottom))
;;;            ((< range-hi-new range-hi-old)
;;;             (append (adjust-single-distribution algorithm 
;;;                                                 (list range-lo-old range-hi-old phrases-new)
;;;                                                 (list range-lo-new range-hi-new phrases-new)
;;;                                                 'both)                     
;;;                     (list (list (1+ range-hi-new) range-hi-old phrases-old))))
;;;            ((< range-lo-new range-lo-old)
;;;             (adjust-single-distribution algorithm old-distribution
;;;                                         (list range-lo-old range-hi-new phrases-new) 
;;;                                         (if (eql new-real 'bottom) 'none 'top)))
;;;            ((> range-lo-new range-lo-old)
;;;             (cons (list range-lo-old (1- range-lo-new) phrases-old)
;;;                     (adjust-single-distribution algorithm 
;;;                                                 (list range-lo-new range-hi-old phrases-old)
;;;                                                 (list range-lo-new range-hi-new phrases-new)
;;;                                                 new-real)))
;;;            ((null phrases-old) (list (list range-lo-old range-hi-old phrases-new)))
;;;            ((eql new-real 'both)
;;;             (list (list range-lo-old range-hi-old (append phrases-new phrases-old))))
;;;            ((< abs-range-old 5)
;;;             (list (list range-lo-old range-hi-old (append phrases-old phrases-new))))
;;;            ((> (length phrases-old) 1)
;;;             (list (list range-lo-old range-hi-old (append phrases-old phrases-new))))
;;;            (t (split-single-distribution algorithm range-lo-old range-hi-old 
;;;                                          (list phrases-old phrases-new)))))))
                                              

;;;(defun split-single-distribution (algorithm range-lo range-hi phrases-list-list)
;;;  (setq *function-error* 'split-distribution)
;;;  (let ((phrases-lo (first phrases-list-list))
;;;        (phrases-hi (second phrases-list-list)))
;;;    (cond ((eql algorithm 'even)
;;;           (let* ((absolute-range (- range-hi range-lo))
;;;                  (split-point (+ range-lo (floor (/ absolute-range 2)))))
;;;             (list (list range-lo split-point (append phrases-lo phrases-hi))
;;;                   (list (1+ split-point) range-hi (append phrases-hi phrases-lo)))))
;;;          ((eql algorithm 'range-hi)
;;;           (list (list range-lo range-hi (append phrases-hi phrases-lo))))
;;;          ((eql algorithm 'range-lo)
;;;           (list (list range-lo range-hi (append phrases-lo phrases-hi))))
;;;          (t (format t "~&SPLIT-DISTRIBUTION: Sorry, unknown distribution ALGORITHM ~S~%" algorithm)))))
            
        
                            


(defun make-phrase (name instrument length)
  (setf (gethash name *phrases*)
    (make-instance 'phrase :instrument instrument
      :rhythm (make-array length :initial-element 0)
      :pitch (make-array length)
      :dynamics (make-array length)
      :duration (make-array length)
      :clock-offset clock-offset
      :tempo tempo)))
                                 
(defun in-range (range pitch)
               (and (>= pitch (first range))
                    (<= pitch (second range))))

