;;Copyright (C) 2002 Soren Goodman
;;Algalog Music Score Generator
;;See readme.txt for more info

;;;templates.lisp

#| This whole file is quoted.  


;;;;;;;;;;;;;;;
;;;TEMPLATES;;;
;;;;;;;;;;;;;;;
    
Use these templates as an outline to what parameters a function call is expecting.  
Just copy, paste to a new file, and fill in the keywords.  Most keywords can be left blank
to make them inactive in the macros, and some can be left blank in the section tools.


Each of these templates should be SETQ'd to a variable.  The macro-templates should NOT be quoted, 
but the rest should be quoted (the macros output quoted phrases). 

Type (create PHRASE-VAR num-attacks) to create a phrase, the functions PITCH, RHYTHM DURATIONS 
and DYNAMICS will access the list of the given parameter, and VIEW will inspect the associated
phrase output.

Type (create SECTION/COMPISITION) to create a section, distributed-section or composition, since
the num-attacks are given.  This will automatically create the phrases within that section, and 
each individual can viewed as for an individual phrase. 

Type (create-n-save "filename" SECTION or (PHRASE num-attacks)) to save as a MIDI file called filename.midi.



;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;PHRASE-MACRO TEMPLATES;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

These are the basic building block of a piece.  Each outputs a phrase. 
:key '() indicates that the keyword is expecting data
:key NIL indicates that the keyword can be nil
:key (blank)  indicates that the keyword can be left out, but should have data if it is present

INSERT DOCUMENTATION HERE (not yet finished) 


  "SIMPLE-PHRASE is for basic entry of data.  Melody, rhythm and durations all take a list
and can modify this list with nth-n, transpose for pitch, and multiply for rhythm and durations.
Rhythm can also be truncated, and dynamics is unstructured."

(setq var (simple-phrase 
           :melody '() :n-p nil :num-items-p nil :transpose nil
           :rotate-p nil :repeat-p nil
           :rhythm '() :n-r nil :num-items-r nil :multiply-r nil :truncate nil
           :durations '() :n-d nil :num-items-d nil :multiply-d nil
           :dynamics 
           :extra nil 
           :instrument nil ;;if nil, this becomes the current value of INSTRUMENT...can also be a list of # for csound 
           :monophonic t :delay nil))
    
    
  "CHANCEW-SPLIT uses the chanceW function for both rhythm and duration and a 
transposed split-list for pitch."

(setq var (chancew-split 
           :melody '() :n-p nil :num-items-p nil 
           :rotate-p nil :repeat-p nil
           :length nil :include nil :transpose nil                      
           :max-r  :min-r  :window-curve-r  :chance-r 
           :multiply nil :truncate nil      
           :max-d  :min-d  :window-curve-d  :chance-d        
           :dynamics
           :monophonic t :delay nil))
    

  "VARIATION-SPLIT uses the variation function for rhythm, chancew function duration and 
split-list for pitch."

(setq var (variation-split 
           :melody '() :n-p nil :num-items-p nil  
           :length nil :include nil :transpose nil :rotate-p nil :repeat-p nil
           :rhythm '() :multiplier nil :n-r nil :num-items-r nil
           :max-r  :min-r  :chance-r  :items-r nil :rotate-r nil :repeat-r nil
           :max-d  :min-d  :chance-d  :window-curve-d     
           :dynamics 
           :monophonic t :delay nil :truncate nil))

  
  
  "SPLIT-VARIATION uses the split-list function for rhythm, chancew function duration and 
variation for pitch."

(setq var (split-variation 
           :melody '() :n-p nil :num-items-p nil :transpose nil                          
           :max-p  :min-p  :chance-p  :items-p nil :rotate-p nil :repeat-p nil
           :rhythm '() :multiplier nil :n-r nil :num-items-r nil
           :length nil :include nil
           :max-d  :min-d  :window-curve-d  :chance-d           
           :dynamics  
           :monophonic T :delay nil :truncate nil))

  

  "REORDER-SEQUENCER uses a randomly reordered buffered list for rhythm, chancew function duration and, 
a shrunk list for pitch (i.e. user input)."

(setq var (reorder-sequencer 
           :melody '() :n-p nil :num-items-p nil :transpose nil
           :rotate-p nil :repeat-p nil
           :rhythm '() :multiplier nil :n-r nil :num-items-r nil :rotate-r nil
           :output nil :storage-length nil
           :max-d  :min-d  :window-curve-d  :chance-d           
           :dynamics 
           :monophonic T :delay nil :truncate nil))
    


"Modular-riff takes any number of modular-phrases and turns them into a new modular-phrase that 
can sequence between the different phrases by way of the riff function.  Simply call modular-riff 
with a SEQUENCE where 1 is the first phrase and (1) is a 1 beat rest, sequence-args, a list of args for
the call to m-sequencer, a list of ATTACKS-PER-RIFF (for a 5 note 'riff', you don't have to have that 
sequence space referenced 5 times...you also can't do less...) a global delay (the delays within the 
individual phrases are ignored), :pre-eval and :post-eval (for other stuff) and finally a list of 
MODULAR-PHRASEs, which are the actual phrases to be sequenced."
 
(setq var (modular-riff 
           :sequence '() :sequence-args nil :attacks-per-riff '() 
           :delay nil :monophonic t 
           :reset-location-list '()
           :pre-eval nil :post-eval nil
           :symbolname nil ;;when present, setq's a variable "var-attacks" by calling (number-of-attacks)
           :phrases (list '())))

                          





;;;;;;;;;;;;;;;;;;;;;;;
;;;SECTION-TEMPLATES;;;
;;;;;;;;;;;;;;;;;;;;;;;

INSERT DOCUMENTATION HERE (not yet finished)

"Modular-phrase is the basic building block, however they should rarely be used directly.  
The PHRASE-MACRO-TEMPLATES given above all evaluate out to a call to MODULAR-PHRASE, with
many useful tools built in.  It is, however, useful to know the structure when examining 
the output of a phrase-macro."

(setq var '(modular-phrase 
            :num-attacks 
            :rhythm '() 
            :pitch '() 
            :duration '()
            :dynamics '() 
            :instrument  :tempo nil :monophonic t 
            :delay nil :pre-eval nil :post-eval nil))


"A section consists of a number of phrases, given in phrase list, 
with corresponding NUM-ATTACKS, CLOCK-OFFSET, INSTRUMENT and TEMPO."

(setq var '(section 
            :phrase-list '() 
            :num-attacks-list '()
            :clock-offset-list '()         
            :instrument-list '()
            :tempo-list '(240)))


"DISTRIBUTED-SECTION is like SECTION, but with an extra feature.
Each section consists of a number of phrases and instruments, instead of just 1, 
and the notes from these phrases are distributed between the given instruments."

(setq var '(distributed-section 
            :creating t
            :input-phrases-list '(())
            :num-attacks-list '()
            :clock-offset-list '()
            :output-surnames '()
            :output-instruments-list '(())
            :instrument-distribution-list nil
            :note-preference-list '(attack or duration)
            :tempo-list '(240)))


"A composition consists of a number of sections with clocks for offsetting each
phrase in a given section."

(setq var '(composition :section-clock-list '((section1 clock1)(section2 clock2)(...))))




  
  
  
  
  
|#