;;Copyright (C) 2001 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;utilities.lisp
;;;sept 2001
;;;textures thesis project

;;TO DO: optimize FIND-AVAILABLE-PITCHES, it is very slow.
;;fix location-loop, it doesn't work quite right for starting and ending in the middle...

(in-package :cl-user)

(defvar *suppress-display-output* (make-string-output-stream))
;;this temporarily suppresses display

;;WARNING: doesn't always return display to normal, for weird reasons...(setq *standard-output* *error-output*) in acl
(let ((*stdo* *standard-output*))
  (defmacro suppress-display (function)
    (setq *standard-output* (setq *suppress-display-output* (make-string-output-stream)))
    (eval function)
    (setq *standard-output* *stdo*)
    nil))

(defun fix-display ()
  (setq *standard-output* *error-output*))

;;not mine - found in a big collection of stuff...
(defun listify-string (string)
  "Turns a string into a list of symbols."
  (let ((eof (gensym))
        (result nil)
        (start 0)
        item)
    (loop
      (multiple-value-setq (item start)
        (read-from-string string nil eof :start start))
      (when (eq item eof)
        (return result))
      (setq result (nconc result (list item))))))


;;;takes a list for m-ramp, and does a location loop to create a pre-rendered ramp
;;(make-ramp '(5 10) 6)
;;--->(5 6 7 8 9 10)
(defun make-ramp (list length &optional quantize1) 
  (let* ((function (m-ramp list))
         (length (get-data length))
         (out (location-loop function length)))
    (if quantize1 (quantize-to-n 1 out)
      out)))

;;returns a series of LENGTH integers starting on START
(defun series (length &optional (start 0))
  (loop for n from start to (+ start length) collect n))
       

;;a defparameter that can set more than 1 variable (like a multiple setq but without the fucking "undeclared variable" bullshit).
(defmacro dsetq (&rest args)
  `(progn
     (do ((n 0 (+ 2 n)))
         ((>= n (length ',args)))
       (let ((item (nth n ',args))
             (value (nth (1+ n)',args)))
         (eval `(defparameter ,item ,value))))
     (eval (end (butlast ',args)))))
       

;;(make* 'hi)
;;-->*HI*
(defun make* (var)
  (concat-symbol "*"
                 (concat-symbol (princ-to-string var) '*)))

(defun buffer-ref ()
  (prog1 *buffer-ref* (incf *buffer-ref*)))

(defun rhythm (phrase)
  (let ((phrase (fap phrase)))
    (if phrase
        (:rhythm phrase)
      (format t "~S is not a phrase, cannot show RHYTHM.~%" phrase))))

(defun pitch (phrase)
  (let ((phrase (fap phrase)))
    (if phrase
        (:pitch phrase)
      (format t "~S is not a phrase, cannot show PITCH.~%" phrase))))

(defun durations (phrase)
  (let ((phrase (fap phrase)))
    (if phrase
        (:duration phrase)
    (format t "~S is not a phrase, cannot show DURATION.~%" phrase))))

(defun dynamics (phrase)
  (let ((phrase (fap phrase)))
    (if phrase
        (:dynamics phrase)
      (format t "~S is not a phrase, cannot show DYNAMICS.~%" phrase))))

(defun print-instrument (instrument)
  (if (listp instrument)
      (mapcar #'print-instrument instrument)
    (let* ((object (eval instrument))
           (range-lo (:range-lo object))
           (range-hi (:range-hi object)))
      (list instrument range-lo range-hi))))


;;Returns the current date in string format.  Doesn't like \, but will do / for seperator.
;;american is d/m/y, euro is m/d/y
(defun date-string (&optional (seperator #\.)(type 'american))
  (let* ((time (multiple-value-list (get-decoded-time)))
         (date (nth-n 3 3 time))
         (date-string
          (cond ((eql type 'american)
                 (format nil "~S~c~S~c~S" (second date) seperator (first date) seperator (third date)))
                ((eql type 'euro)
                 (format nil "~S~c~S~c~S" (first date) seperator (second date) seperator (third date))))))
    date-string))
    ;(read-from-string date-string)))
                                     
         

(defmacro nullify (var)
  `(setq ,var nil))

(defun available-functions (&optional disinclude-lisp-functions)
  (if disinclude-lisp-functions (apply #'append (mapcar #'eval *non-lisp-functions-list*))
    (apply #'append (mapcar #'eval *functions-list*)))) 
;;;
;;;(defun available-function (item)
;;;  (if (atom item) (and (member item (available-functions)) t) ;;We want to return T, not a list..
;;;    (let ((function (first item)))
;;;      (if (eql function 'quote) (available-function (second item))
;;;        (available-function function)))))

;;this should be replaced in the future...
(defun available-function (item)
  (fboundp item))



(defun make-error (error-string)
  (let ((error-string  
         (format nil "~A - ~A; ~A" cur-phrase name error-string)))
    (push error-string *runtime-errors*)
    nil))

(defun init () ;;I am lazy, this is shorter
  (initialize 'all))

(defun initialize (&optional (item 'all-but-phrases) &rest more)
  (initialize-f item more) 
  (when (fboundp 'gc)(gc))) ;;gc, garbage collect, is not common lisp...

(defgeneric initialize-f (item &rest more))

(defmethod initialize-f (item &rest more)
  (format t "Sorry, can't initialize ~S or ~S.~%" item more))

(defmethod initialize-f ((item null) &rest more)
 t)

(defmethod initialize-f ((item (eql 'buffers)) &rest more)
  (mapcar #'(lambda (x) (set x '())) *buffers*)
  (setq cur-beat 0)
  (clrhash *buffer-hash*)
  (when more (let ((more (clean-up-list more))) 
               (initialize (first more) (rest more)))))                               

(defmethod initialize-f ((item (eql 'phrases)) &rest more)
  (clrhash *phrases*)
  (when more (let ((more (clean-up-list more))) 
               (initialize (first more) (rest more)))))

(defmethod initialize-f ((item (eql 'dot)) &rest more)
  (setq print-next-dot 1)
  (when more (let ((more (clean-up-list more))) 
               (initialize (first more) (rest more)))))


(defmethod initialize-f ((item (eql 'errors)) &rest more)
  (mapcar #'(lambda (x) (set x '())) *error-vars*)
  (when more (let ((more (clean-up-list more))) 
               (initialize (first more) (rest more)))))


;;there is no real reason to call this function anymore, since buffer now uses a hash table which is initialized
;;when running CREATE...all that matters is that each time *buffer-ref* is accessed, it is a unique number.
(defmethod initialize-f ((item (eql 'macros)) &rest more)
  (when (> *buffer-ref* 99000) (setq *buffer-ref* 0))
  (when more (let ((more (clean-up-list more))) 
               (initialize (first more) (rest more)))))
  

(defmethod initialize-f ((item (eql 'all)) &rest more)
  (initialize 'errors 'dot 'phrases 'buffers 'macros))


(defmethod initialize-f ((item (eql 'all-but-phrases)) &rest more)
  (initialize 'errors 'dot 'buffers 'macros))

(defun durations-to-rhythm (rhythm)
"(durations-to-rhythm '(1 0 2 0 1 1 0 2))
---->(1 0 1 0 0 1 1 0 1 0)"
  (setq *utility-error* 'durations-to-rhythm)
  (loop for n from 0 to (1- (length rhythm)) append
        (let ((duration (nth n rhythm))) 
          (if (zerop duration) '(0)
            (check-and-modify-length '(1 0) duration)))))

;;checks if list is really a list...otherwise lists it.
(defun check-list (list)
  (if (listp list) list
    (list list)))

(defun biggest-number (list)
  (when list
    (let ((number-list 
           (loop for x from 0 to (1- (length list)) collect
                 (get-data (nth x list)))))
      (apply #'max number-list))))

(defmacro flip (symbol)
  `(setq ,symbol (if ,symbol nil t)))

(defun initial-zeros (list &optional (count 0))
  (cond ((null list) count)
        ((zerop (first list))
         (initial-zeros (rest list) (1+ count)))
        (t count)))
     
;;turns an array into a list and removes all nil's
(defun reduce-vector (vector)
  (setq *utility-error* 'reduce-vector)
  (remove nil
          (loop for n from 0 below (length vector) with temp = nil collect
                (svref vector n))))

(defun find-associated-phrase (symbol)
  (gethash symbol *phrases*))

;;this uses fap with the inspector...shorthand for looking at the output of a rendering using the inspector
(defun view (symbol)
  (inspect (fap symbol)))

;;just a nice shorthand...I use this function too much...
(defun fap (s)
  (find-associated-phrase s))

(defun concat-symbol (string symbol)
  (read-from-string (concatenate 'string string (princ-to-string symbol))))
               


;;this is for finding keyword parameters in a quoted function call, i.e.
;;(find-keyword ':test '(foo :baz 1 :bar 2 :test 'bingo :bap 'nope))
;;--------------->BINGO
(defun find-keyword (keyword section)
  (cond ((or (not (listp section)) 
             (null section))
         (format t "Sorry, couldn't find keyword ~S.~%" keyword))
        ((eql (first section) keyword)
         (eval (second section)))
        (t (find-keyword keyword (rest section)))))
        
(defun test-single (function &optional (num-attacks 100))
  (loop for x from 0 to (1- num-attacks) collect
        (let ((location (/ x num-attacks)))
          (eval function))))

(defun test-segment (function &optional (num-attacks 100) (start-pos 0))
  (initialize 'buffers 'errors 'dot)
  (segment function num-attacks start-pos))

;;does a descriptive print of an item and returns that item (instead of returning nil)
;;also prints other variables before eval'ing ITEM.  Useful for debugging.
(defmacro fprint (string &optional item &rest other-items)
  `(progn (format t "~&")
     (let ((out (mapcar #'(lambda (x) `(format t "~S " ,x)) ',other-items)))
       (mapc #'eval out)
       (if ,item
           (format t "~&~a ~S" ,string ,item) 
         (format t "~&~a" ,string))
       (terpri) 
       ,item)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;taken from  10/92-1994 Marty Hall. hall@aplcenmp.apl.jhu.edu;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro Def-Class (Class-Name Super-Class-List &rest Slot-Entries)
  "Expands into a defclass form. Simplest format:
  
  (Define-Class Foo (Bar) 
     (Slot-1 Val-1) 
     Slot-2) 
  -->
  (defclass Foo (Bar)
    ((Slot-1 :initform Val-1 :accessor :Slot-1 :initarg :Slot-1)
     (Slot-2                 :accessor :Slot-2 :initarg :Slot-2)))

  You can also add any of the normal slot-spec keywords after the slot value
  to specify the :allocation, :type, etc., plus there is an additional keyword
  called :DOC-STRING that is used to specify the doc string for the generic
  function having the same name as the slot. Finally, you can replace `Foo'
  with `(Foo (class-option)* )'. The syntax is *exactly* the same as
  Define-Class (but Named-Object is not mixed in as a parent)."

  (if
    (atom Class-Name)
    `(defclass ,Class-Name ,Super-Class-List
	 ,(mapcar #'Expand-Slot-Name-Value-Pair Slot-Entries))
    `(defclass ,(first Class-Name) ,Super-Class-List 
	 ,(mapcar #'Expand-Slot-Name-Value-Pair Slot-Entries)
       ,@(rest Class-Name)) )
)

(defun Expand-Slot-Name-Value-Pair (Slot-Entry)
  (let (Slot-Name Slot-Value Extra-Keywords Doc-String)
    (cond
      ((listp Slot-Entry)
       (setq Slot-Name (first Slot-Entry)
	     Slot-Value (second Slot-Entry)
	     Extra-Keywords (rest (rest Slot-Entry))
	     Doc-String (getf Extra-Keywords :Doc-String))
       (when Doc-String
	 (remf Extra-Keywords :Doc-String)
	 (setf (documentation Slot-Name 'function) Doc-String))
       (append
	 (list Slot-Name
	       :accessor (Add-Colon Slot-Name)
	       :initform Slot-Value
	       :initarg (Add-Colon Slot-Name))
	 Extra-Keywords) )
      (t
       (setq Slot-Name Slot-Entry)	; Redundant for consistency w/ above
       (list Slot-Name
	     :accessor (Add-Colon Slot-Name)
	     :initarg (Add-Colon Slot-Name)) ) )
    ))

(defmethod Add-Colon ((Sym symbol))
  (if (keywordp Sym)
      Sym
      (intern (symbol-name Sym) :keyword)) )

(defmethod Add-Colon ((Str string))
  (intern (string-upcase Str) :keyword) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;
;;;GET-DATA;;;
;;;;;;;;;;;;;;

;;this function looks at *functions* and if the first item of a list is a function, it is eval'd.

(defun get-data (item)
  (setq *utility-error* 'get-data)
  (cond ((listp item)
         (if (fboundp (first item))
             (get-data (eval item))
           item))
        ((numberp item) item)
        ((null item) nil)
        ((member item (append *function-keys* *ramp-types*)) item)
        ((symbolp item)
         (if (boundp item)
             (eval item)
           item))

        ((objectp item) item)
        (t (format t "Sorry, I don't know how to get-data from ~S.~%" item))))

;;Does as much evaling as possible.  
;;(setq egg 50) 
;;(makunbound 'surprise)
;;(safe-eval '(+ egg (+ 3 surprise (* 3 egg))))
;;------>(+ 50 (+ 3 Surprise 150))

(defun safe-eval (l)
  (cond ((atom l) l)
        ((fboundp (first l))
         (eval (cons (first l)
                     (safe-eval1 (rest l)))))
        (t (safe-eval1 l))))

(defun safe-eval1 (l)
  (setq *utility-error* 'safe-eval)
  (if (atom l)
      l
    (let ((new (mapcar #'safe-eval1 l)))
      (if (and (fboundp (first new))
               (eval (cons 'and  ;;can't (apply #'and
                           (mapcar #'(lambda (x) (and (atom x)
                                                      (or (numberp x)
                                                          (boundp x)))) ;;nothing unbound
                             (rest new))))) ;;first should be function
          (eval new)
        (mapcar #'safe-eval1 new)))))

        
;;;(defun ok-evalp (l)
;;;  (if (atom l)
;;;      (if (or (numberp l)
;;;              (objectp l)
;;;              (boundp l))
;;;          t
;;;        nil)
;;;    (and (fboundp (first l))
;;;         (eval (cons 'and
;;;                     (mapcar #'ok-evalp (rest l)))))))


;;;;;;;;;;;;;;;
;;;GET-MACRO;;;
;;;;;;;;;;;;;;;

;;get macro searches through layers of a list, and if it encounters a member of *macro-functions* it evaluates that
;;segment, thus expanding every "macro" function in a list. 


(defun get-macro (list)
  (when list
    (let ((list (unquote list)))
      (if (not (listp list))
          (quote-check list)
        (let ((function (first list)))
          (cond ((member function *macro-functions*)
                 (eval list))
                ((listp function)
                 (mapcar #'get-macro list))
                (t list)))))))
  

;;;(defun get-macro (list)
;;;  (when list
;;;    (if (not (listp list))
;;;        list
;;;      (let ((function (first list)))
;;;        (if (member function *macro-functions*)
;;;            list)
;;;        list))))
           

;;;;;;;;;;;;;;;
;;;GET-RANGE;;;
;;;;;;;;;;;;;;;

;;get-range checks to make sure range is not > (:range-hi instrument) or < (:range-lo instrument)

(defun get-range (item)
  (let* ((range (get-data item))
         (instrument (if instrument instrument
                       default-instrument))
         (range-hi (:range-hi instrument))
         (range-lo (:range-lo instrument)))
    (cond ((>= range range-hi) range-hi)
          ((<= range range-lo) range-lo)
          (t range))))
         

;;;;;;;;;;;;;;;
;;;LAST-BEAT;;;
;;;;;;;;;;;;;;;

(defun last-beat (phrase &optional (include-clock t))
  (setq *utility-error* 'last-beat)
  (cond ((null phrase) 0)
        ((numberp phrase) phrase)
        ((listp phrase)
         (let ((item (first phrase)))
           (cond ((member item *control-functions*)
                  (make-error "LAST-BEAT: Input phrase should be quoted."))
                 ((not (typep item 'phrase))
                  (mapcar #'(lambda (x) (last-beat x include-clock))
                    phrase))
                 ((eql item 'section)
                  (let* ((phrase-list (find-keyword ':phrase-list phrase))
                         (last-beat-list (mapcar #'(lambda (x) (last-beat x include-clock)) phrase-list)))
                    (biggest-number last-beat-list)))
                 ((typep item 'phrase)
                  (biggest-number (mapcar #'(lambda (x) (last-beat x include-clock)) phrase)))
                 ((member item (available-functions))
                  (last-beat (get-data phrase))))))
        ((typep phrase 'phrase)
         (let ((clock (if include-clock (round (get-data (:clock-offset phrase)))
                        0))
               (rhythm (:rhythm phrase)))
           (+ clock (apply #'+ rhythm))))
        ((symbolp phrase)
         (let* ((phrase-output (find-associated-phrase phrase))
                (phrase-output (if (null phrase-output) (eval phrase)
                                 phrase-output)))
           (if (or (listp phrase-output)
                   (objectp phrase-output))
               (last-beat phrase-output include-clock)
             (format t "Sorry, ~S is not a valid phrase or section; can't find last-beat." phrase))))
        (t (prog1 0 (format t "Sorry, ~S is not a valid phrase or section; can't find last-beat." 
                      phrase)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;RHYTHM/LIST UTILITIES;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;this set of functions returns the members of LIST that are (>, >=, <, <=) COMPARE.

;;(swap-positions 0 2 '(1 2 3 4) --> '(3 2 1 4)
(defun swap-positions (pos1 pos2 list)
  (let ((list (copy-list list))) ;;don't want destructive...
        (psetf (nth pos1 list) (nth pos2 list)
          (nth pos2 list) (nth pos1 list))
    list))

(defun list> (compare list)
  (let ((compare (get-data compare))
        (list (get-data list)))
    (remove-if (lambda (x) (<= x compare)) list)))

(defun list>= (compare list)
  (let ((compare (get-data compare))
        (list (get-data list)))
    (remove-if (lambda (x) (< x compare)) list)))

(defun list< (compare list)
  (let ((compare (get-data compare))
        (list (get-data list)))
    (remove-if (lambda (x) (>= x compare)) list)))

(defun list<= (compare list)
  (let ((compare (get-data compare))
        (list (get-data list)))
    (remove-if (lambda (x) (> x compare)) list)))

(defparameter *nth-n-documentation*
  "Returns the middle chunk of a list.  N starts at 0.
A value of nil for num-items causes nth-n to act as nthcdr.
(nth-n 2 4 '(1 2 3 4 5 6 7 8))
------>(3 4 5 6)")

(defparameter *nth-n-prompt*
  "(nth-n n num-items list)")

(defun nth-n (n num-items list)
  (let* ((n (get-data n))
         (n (if (numberp n) 
                (round n)
              nil))
         (list (get-data list))
         (num-items (get-data num-items))
         (num-items (if (numberp num-items)
                        (round num-items)
                      nil)))
  (cond ((null n)
         (if (null num-items) list
           (first-n num-items list)))
        ((null num-items)
         (nthcdr n list))
        (t (let ((list (nthcdr n list)))
             (if (> num-items (length list))
                    list
               (first-n num-items list)))))))

;;;;;;;;;;;;;;;;;;;;;;
  #|Takes lists of rhythm and duration and merges them by
1) converting rests into 0's 2) converting duration to S's and 3) converting staccato's into .5's
(combine-rhythm-duration '(4 4 4) '(4 2 0))
--->(1 S S S 1 S 0 0 0.5 0 0 0)|#

(defun combine-rhythm-duration (rhythm duration)
  (setq *utility-error* 'combine-rhythm-duration)
  (let ((rhythm (get-data rhythm))
        (rhythm (if (and (listp rhythm)
                         (member 0 rhythm))
                    (rhythm-to-durations rhythm)
                  rhythm))
        (duration (get-data duration))
        (duration (if (null duration) 
                      (check-and-modify-length '(1) rhythm)
                    (when (listp rhythm)
                      (fix-durations rhythm duration)))))
    (unless (and (listp duration) (listp rhythm))
      (error "Combine-rhythm-duration must have 2 lists for input!"))
    (append
     (loop for n from 0 below (1- (length rhythm)) append
           (make-single-rhythm-duration (nth n rhythm) (nth n duration)))
     (make-single-rhythm-duration (end rhythm) (end duration) t)))) 
;;last beat is switched to allow for full duration

;;assume rhythm has no 0's (run through rhythm-to-duration)
(defun make-single-rhythm-duration (rhythm duration &optional last-beat)
  (setq *utility-error* 'make-single-rhythm-duration)
  (cond ((= rhythm duration) (make-sustain rhythm))
        ((> duration rhythm) (if last-beat (list duration)
                               (make-sustain rhythm)))
        (t
         (if (zerop duration) 
             (cons .5 
                   (loop for x from 1 below rhythm collect 0))
           (append (make-sustain duration)
                   (loop for x from 1 to (- rhythm duration) collect 0))))))

(defun make-sustain (duration)
  (cons 1 (loop for x from 1 below duration collect 'S)))

;;converts 1's (or bigger) and 0's to length of beat
;;(rhythm-to-durations '(1 0 0 4 0)) --> (3 5)
(defun rhythm-to-durations (list)
  (let ((count 0)
        (temp nil))
     (do* ((temp nil)(n (1- (length list)) (decf n)))
          ((zerop (1+ n)) temp)
       (let* ((item (nth n list))
              (item (if (= .5 item) 1 item)))
         (if (zerop item)
               (incf count)
               (if (> count 0) 
                   (prog1 (push (+ count item) temp)(setq count 0))
                 (push item temp)))))))


;;use this instead of segment for smallruns of location from 0 to 1...segment is weird..
;;use the append option if you are unsure whether the incoming data is an atom or list
(defun location-loop (function num-attacks &optional (append t)(start-position 0)
                               (end-position num-attacks))
  (setq *utility-error* 'location-loop)
  (if (>= start-position end-position)
      (let ((location 1))
        (get-data function))
    (let ((output-length (abs (- end-position start-position))))
      (if append
          (let (temp)            
            (do* ((cur-position start-position (+ start-position (length temp))))
                 ((>= cur-position end-position)
                  (return (check-and-modify-length temp output-length)))
              (let* ((location (/ cur-position (1- num-attacks)))
                     (output (get-data function)))
                (if (and output (listp output)) (setq temp (append temp output))
                  (setq temp (add-to-end temp output))))))
        (check-and-modify-length 
         (loop for cur-position from start-position below end-position collect 
               (let ((location (/ cur-position (1- num-attacks))))
                 (get-data function)))
         output-length)))))


;;;(defun location-loop (function num-attacks &optional (append t)(start-position 0)
;;;                               (end-position num-attacks))
;;;  (setq *utility-error* 'location-loop)
;;;  (if (>= start-position end-position)
;;;      (let ((location 1))
;;;        (get-data function))
;;;    (let ((output-length (abs (- end-position start-position))))
;;;      (if append
;;;          (let (temp)            
;;;            (do ((cur-position start-position (+ start-position (length temp))))
;;;                ((>= cur-position end-position)
;;;                 (return (check-and-modify-length temp output-length)))
;;;              (let* ((location (/ cur-position (1- num-attacks)))
;;;                     (output (get-data function)))
;;;                (if (and output (listp output)) (setq temp (append temp output))
;;;                  (setq temp (add-to-end temp output))))))
;;;        (check-and-modify-length 
;;;         (loop for cur-position from start-position below end-position collect 
;;;               (let ((location (/ cur-position (1- num-attacks))))
;;;                 (get-data function)))
;;;         output-length)))))



  
(defun end (list)
  (first (last list)))

;;Take a rhythm-duration list and extract just duration.  Note that if duration > rhythm, duration gets chopped.
(defun extract-durations (list &optional (remove-zeros t))
  (setq *utility-error* 'extract-durations)
  (unless (listp list)
    (error "Extract-durations: ~s is not a list!" list))
  (let ((temp nil)
        (list (reverse (if remove-zeros
                           (remove-zeros-from-end list) 
                         list)))
        (count 0))
    (loop for n from 0 below (length list) do
          (let ((item (nth n list)))
            (cond ((eql item 'S)
                   (incf count))
              ((zerop item) nil)
              ((= item .5)
               (push 0 temp))
              (t 
               (prog1 (push (+ count item) temp)
                 (setq count 0))))))
    temp))

(defun extract-rhythm (list)
  (let ((temp nil)
        (list (reverse list))
        (count 1))
    (loop for n from 0 below (length list) do
          (let ((item (nth n list)))
            (if (or (eql item 'S) (zerop item))
                (incf count)
              (progn (push count temp) (setq count 1)))))
    temp))

;;removes all zeros at the end of the list, i.e. '(1 2 3 4 0 1 2 0 0 0) --> (1 2 3 4 0 1 2)
(Defun remove-zeros-from-end (list)
  (let ((list (reverse list)))
    (do* ((n 0 (incf n))(item (first list)(nth n list)))
         ((not (eq item 0))
          (if (>= n (length list)) nil
            (append (reverse (nthcdr (1+ n) list))
                    (list item)))))))

(defun total-length (list)
  "Gives the total length of a list, when possible, by evaluating sublists and using the total elements."
  (setq *utility-error* 'total-length)
  (cond ((null list) 0)
    ((atom list) 1)
    (t
     (loop for x from 0 below (length list) sum
           (total-length (get-data (nth x list)))))))

(defun debug-switch ()
  (flip *debug*))

(defun repeat-n-times (n list)
  "Note that this function accepts floats for n, so that a list can be partially repeated."
  (setq *utility-error* 'repeat-n-times)
  (let ((list (check-list list)))
    (if (typep n 'fixnum)
        (loop for x from 0 to n append
              list)
      (let ((extra-chunk (first-n (round (* (rem n 1) (length list))) list)))
        (append (repeat-n-times (floor n) list) extra-chunk)))))

;;same as repeat-n-times, but repeated items are not placed at the end of the list and N must be an integer.
;;(duplicate-n-times 1 '(1 2 3))-->(1 1 2 2 3 3)
(defun duplicate-n-times (n list)
  (setq *utility-error* 'duplicate-n-times)
  (let ((list (get-data list))
        (n (1+ (round (get-data n)))))
    (loop for o from 0 below (length list) append
          (repeat (nth o list) n))))
  
(defun rotate-list (list amount)
  (setq *utility-error* 'rotate-list)
  (let* ((list (get-data list))
         (amount (get-data amount))
         (amount (if (fixnump amount) amount
                   (round (* (length list) amount)))))
    (append (last-n (- (length list) amount) list)
            (first-n amount list))))

(defun total-duration (beats &optional (tempo *tempo*))
  (* beats (ms-per-beat tempo)))

(defun duration-to-beats (duration &optional (time 'ms)(tempo *tempo*))
  (let ((duration (cond ((eql time 'seconds)
                         (* 1000 duration))
                        ((eql time 'minutes)
                         (* 60000 duration))
                        (t duration))))
    (round (/ duration (ms-per-beat tempo)))))
                    

(defun beats-to-measures (beats &optional (beats-per-measure 16))
  (1+ (if beats
          (floor (/ beats beats-per-measure))
        0)))

#|
use member, faster...
(defun in-listp (item list)
  "(in-listp 3 '(1 2 3))
>>>3"
  (cond ((null list) ())
        ((eql item (first list)) item)
        (t (in-listp item (rest list)))))
|#

(defun remove-first-occurence (item list)
  (remove item list :test 'eql :count 1))

(defun remove-items (items list)
  (if (null items)
      list
    (remove-items (rest items) (remove (first items) list :test 'eql :count 1))))


;;removes the nth item from list
(defun remove-nth (n list)
  (remove n list :count 1
          :test (lambda (x y) (equal x (position y list)))))

;;inserts ITEM after position N in LIST
;;(insert-nth nil 2 '(1 2 3 4 5 6))
;;--------->(1 2 3 Nil 4 5 6)
(defun insert-nth (item n list)
  (append (first-n (1+ n) list) (list item) (nthcdr (1+ n) list)))

;;duplicates whatever is at the nth position of list.
;;(duplicate-nth 2 '(1 2 3 4))
;;--------->(1 2 3 3 4)
(defun duplicate-nth (n list)
  (append (first-n n list) (list (nth n list)(nth n list)) (nthcdr (1+ n) list)))

;;reorders using list as template
(defun order-to-list (list unordered &optional (position 0))
  "(order-to-list '(0 1 0 2 0 3 0 4) '(chicken 5 1 3 0 2 4 1 0 shit))
>>>>>(0 1 0 2 3 4 chicken 5 1 shit)"
  (setq *error* 'order-to-list)
  (let ((item (nth position unordered)))
    (cond ((null list)
           unordered)
          ((null item)  
           (order-to-list (rest list) unordered))
          ((eql item (first list))
           (cons item 
                 (order-to-list (rest list) (remove item unordered :count 1 :test 'eql))))
          (t (order-to-list list unordered (1+ position))))))


;;reorders using a LIST of integers to define placement.  If LIST is not a series '(0 1 2 ... ) then 
;;FILLER is used to fill the empty spaces.

;;note, list must not be below

;;(order-by-list '(4 1 2) '(a b c) 'd)
;;-------------->(D B C D A)

(defun order-by-list (list unordered &optional (filler 0))
  (let* ((max (1+ (apply #'max list)))
         (min (apply #'min list))
         (list (remove-duplicates list)) ;;each space must be unique,
         (unordered (if (< (length list) (length unordered))
                        (check-and-modify-length unordered list)
                      unordered))
         (output (repeat filler max)))
    (when (minusp min) (error "ORDER-BY-LIST:  Smallest item of LIST, ~A, is less than 0." min))
    (when (< max (length unordered)) (make-error "ORDER-BY-LIST: max of LIST is smaller than length of UNORDERED"))
    (loop for n from 0 below (length list) do
          (setf (nth (nth n list) output) (nth n unordered)))
    output))
  
  
  
(defun add-to-end (list item)
  (append list (list item)))

(defun repeat-end (list)
  (append list (last list)))

(defun pick-one (list)
  (let ((list (get-data list)))
    (nth (random (length list)) list)))

(defun randomize (list)
  (loop for x from 0 below (length list) collect
        (let ((item (pick-one list)))
          (setq list (remove-first-occurence item list))
          item)))

(defun pick-n (n list)
  (unless (or (zerop n)
              (null list))
    (let* ((item (pick-one list))
           (new-list (remove-first-occurence item list)))
      (cons item (pick-n (1- n) new-list)))))

(defun num-attacks (beat-list &optional (current 0))
  (cond ((null beat-list) current)
    ((eql (first beat-list) 'S)(num-attacks (rest beat-list) current))
    ((> (first beat-list) 0) (num-attacks (rest beat-list) (1+ current)))
    (t (num-attacks (rest beat-list) current))))

#|
Calculates the number of 0's after the first element.  It is assumed that the first element is an attack.
|#

;;beats-to-next-attack adds the length of each attack, i.e. '(3 0 0 2) --> 5
(defun beats-to-next-attack (beat-list)
  (loop for n from 1 to (1- (length beat-list)) do 
        (if (zerop (nth n beat-list))
            (when (null (nth (1+ n) beat-list))
              (return 0))
          (return (+ (first beat-list) n)))))

;;like beats-to-next-attack but doesn't include length of first element
(defun elements-to-next-attack (beat-list)
  (loop for n from 1 to (1- (length beat-list)) do 
        (if (zerop (nth n beat-list))
            (when (null (nth (1+ n) beat-list))
              (return 0))
          (return n))))

(defun nth-attack (n beat-list)
  (if (<= n 0)
      beat-list
    (let ((next-attack (elements-to-next-attack beat-list)))
      (nth-attack (1- n) (nth-rest next-attack beat-list)))))

(defun ms-per-beat (tempo)
  (* 1000
     (/ 60 tempo)))

(defun nth-rest (n list)
  (let ((n (round (get-data n)))
        (list (get-data list)))
    (nthcdr n list)))

#| 
(clean-up-list '((1 ((2)) ((((3))) (4 5)))))
---> (1 2 (3 (4 5)))
(clean-up-list '(1) --> (1)
               ;;I spent WAY TOO FUCKING MUCH TIME ON THIS.  I didn't need to.  I did.  I'll never use it..
|#

(defun clean-up-list (list)
  (let ((data (simplify-parens list)))
    (if (listp data)
        data
      (list data))))

(defun fix-durations (rhythm duration)
  (setq *utility-error* 'fix-durations)
  (let* ((rhythm (rhythm-to-durations rhythm))
         (duration (mapcar #'(lambda (x) 
                               (if (null x) 
                                   (progn (make-error "Duration had NIL in list; Using 1")
                                     1) 
                                 x)) duration))
         (duration (if (null duration) '(1) duration))
         (duration (check-and-modify-length duration rhythm))
         (temp nil))
    (loop for n from 0 to (1- (length rhythm)) do
          (let ((r-item (nth n rhythm))
                (d-item (nth n duration)))
            (push (min r-item d-item) temp)))
    (reverse temp)))
        

(defun simplify-parens (list &aux (list (remove-parens list)))
  (cond ((null list) ())
        ((atom list) list)
        ((= (length list) 1) 
         (remove-parens (first list)))
        (t (loop for n from 0 to (1- (length list)) collect
                 (simplify-parens (nth n list))))))
        
(defun remove-parens (list)
  (cond ((null list) ())
        ((atom list) list)
        ((= (length list) 1)
         (remove-parens (first list)))
        (t list)))

#|Checks the output of ITEM by testing *average-resolution* different location settings.
Item must produce a number|#
(defun get-average (item)
  (setq *utility-error* 'get-average)
  (if (numberp item) item
    (let* ((temp '())
           (num-attacks (if (not (zerop num-attacks)) num-attacks 100))
           (resolution (if (> *average-resolution* num-attacks) num-attacks
                         *average-resolution*))
           (constant (/ num-attacks resolution)))
      (loop as cur-position = (* constant (length temp)) 
          while (< cur-position num-attacks) do
            (let* ((location (/ cur-position num-attacks))
                   (data (get-data item))
                   (adjust-data (if (listp data)
                                    (apply #'+ data)
                                  data)))
              (push (/ adjust-data resolution) temp)))
      (apply #'+ temp))))
          

(defun quantize-to-n (step-size item &optional (round-up t))
  "outputs item so that it is evenly divisible by step-size"
  ;(format t "~&location ~S~%" location)
  (let ((step-size (get-data step-size))
        (item (get-data item))
        (round-up (get-data round-up)))
  (if (listp item) 
      (mapcar #'(lambda (x)
                  (quantize-to-n step-size x round-up))
        item)
    (let ((remainder (rem item step-size)))
      (if round-up
          (if (>= remainder (* .5 step-size))
              (+ item (- step-size remainder))
            (- item remainder))
        (- item remainder))))))


;;repeats each ITEM AMOUNT times. (repeat-items '(1 2 3) 2) --> (1 1 2 2 3 3) 
(defun repeat-items (items amount)
  (let ((items (get-data items))
        (amount (get-data amount)))
    (mapcan #'(lambda (x) (repeat-n-times (1- amount) x)) items)))

(defun repeat (item amount)
  (when (numberp amount)
    (make-list amount :initial-element item)))
  
(defun check-and-modify-length (list length)
  "checks the length of list and either truncates or duplicates the end to achieve LENGTH"
  (let* ((list (if (listp list) list (list list)))
         (length (if (listp length) (length length) length))
         (length (truncate (+ .5 length)))) ;;this is for decimal values, staccato = .5 so it rounds up
    (cond ((zerop length) ())
      ((< length 0)
       (progn (format t "Check-and-modify-length error: length ~S is less than 0.~%" length) list))
      ((null list)
       (check-and-modify-length '(nil) length))
      ((= (length list) 
         length) 
       list)
      ((> (length list)
         length)
       (check-and-modify-length (butlast list) length))
      (t (check-and-modify-length (repeat-end list) length)))))

(defun check-and-modify-rhythm (list length)
  "makes sure there are enough num-attacks in list and adjusts if necessary" 
  (cond ((zerop length)())
        ((< length 0)
         (format t "Check-and-modify-rhythm error: length ~S is less than 0.~%" length))
        ((zerop (first (last list)))
         (check-and-modify-rhythm (butlast list) length))
        ((null list)
         (check-and-modify-rhythm '(1) length))
        ((= (num-attacks list) 
            length) 
         list)
        ((> (num-attacks list)
            length)
         (check-and-modify-rhythm (butlast list) length))
        (t (check-and-modify-rhythm (add-to-end list 1) length))))
         
;;;;;;;;;;;;;;;;;;;;;;
;;;EXTRA MATH STUFF;;;
;;;;;;;;;;;;;;;;;;;;;;

;;first click starts the timer, second click prints the distance between clicks plus any input seconds.
(let (start)
  (defun stopwatch (&optional (plus-seconds 0))
    (let* ((time (multiple-value-list (get-decoded-time)))
           (seconds (+ (if start plus-seconds 0) (first time)(* 60 (second time))(* 60 60 (third time)))))
      (if start 
          (progn 
            (print-convert-to-minutes (- seconds start))
            (setq start nil))
        (format t "Starting stopwatch.~%" (setq start seconds))))))

(defun print-convert-to-minutes (seconds)
  (let* ((minutes (/ seconds 60))
         (hours (floor (mod (/ minutes 60) 99)))
         (minutes (floor (mod minutes 60)))
         (seconds (floor (mod seconds 60))))
    (format t "~s:~S:~S;  ~S hour~p, ~S minute~p and ~S second~p.~%"  hours minutes seconds 
      hours hours minutes minutes seconds seconds))
  t)

(defun scale-to-n (list n)
  (let ((max (max smallnum (apply #'max list))))
    (mapcar (lambda (x) (* n (/ x max))) list)))


;;checks for duplicated items and adds1.  Checks for any number of repetitions.
;;(add1-to-duplicates '(1 1 1 3 5 7))
;;;------------------->(1 2 3 4 5 7)
(defun add1-to-duplicates (numlist)
  (loop for n from 0 below (length numlist) with items collect
        (let ((num (nth n numlist)))
          (do ((n 0 (1+ n)))
              ((>= n (length items)) num)
            (when (member num items) (incf num)))
          (push num items)
          num)))

(defun binary-listp (list)
"(binary-listp '(1 0 0 1 0 1 0))
--->T"
  (cond ((null list) t)
        ((> (first list) 1)
         ())
        (t (binary-listp (rest list)))))

;;deals with lists (get-data)
(defun times (item1 item2)
  (setq *utility-error* 'times)
  (let* ((item1 (get-data item1))
         (item2 (get-data item2)))
    (cond ((or (symbolp item1)(symbolp item2))
           (make-error (format nil "Times: ~S or ~S is not a number of list of numbers." item1 item2)))
          ((and (numberp item1) (numberp item2))
           (* item1 item2))
          ((numberp item1)
           (mapcar #'(lambda (x)(* item1 x)) item2))
          ((numberp item2)
           (mapcar #'(lambda (y)(* item2 y)) item1))
          ((and (listp item1)(listp item2))
           (mapcar #'* item1 item2)))))

(defun non-zero-nump (item)
  (and (numberp item)
       (not (zerop item))))

;;this fixes many floating point errors, such as (+ .9 .01)--->(fix-float .909999997)----> .91
(defun fix-float (number)
  (log (exp number)))

(defun round-to-n (n number)
  (let ((diviser (expt 10 n)))
    (float (/ (round (* number diviser))
              diviser)))) 

(defun positive-decimal (float)
  (cond ((< float 0) 
         (positive-decimal (abs float)))
        ((> float 1)
         (positive-decimal (1- float)))
        (t float)))

(defgeneric biggest-num (num-or-list-of-numbers))

(defmethod biggest-num ((num number))
  num)

(defmethod biggest-num ((list list))
  (let* ((rest (rest list))
         (compare1 (first list))
         (compare2 (first rest)))
    (cond ((null compare2) compare1)
          ((> compare2 compare1)
           (biggest-num rest))
          (t 
           (biggest-num (cons compare1 (rest rest)))))))

(defgeneric smallest-num (num-or-list-of-numbers))

(defmethod smallest-num ((num number))
  num)

(defmethod smallest-num ((list list))
  (let* ((rest (rest list))
         (compare1 (first list))
         (compare2 (first rest)))
    (cond ((null compare2) compare1)
          ((< compare2 compare1)
           (smallest-num rest))
          (t 
           (smallest-num (cons compare1 (rest rest)))))))
       
;;;;;;;;;;;;;;;;;;;;;;
;;;OBJECT FUNCTIONS;;;
;;;;;;;;;;;;;;;;;;;;;;

(defun slot-access (slot object)
  (eval (cons slot (list object))))

(defun objectp (item)
  (if (typep item (find-class 'standard-object))
      item))

(defun objects-only (list)
  (let ((item (first list)))
    (cond ((null list)())
          ((objectp item)
           (cons item
                 (objects-only (rest list))))
          (t (objects-only (rest list))))))

(defun data-only (list)
  (let ((item (first list)))
    (cond ((null list)())
          ((objectp item)
           (data-only (rest list)))
          (item (cons item
                      (data-only (rest list))))
          (t (data-only (rest list))))))

;;;;;;;;;;;;;;;;;;;;;
;;;PITCH FUNCTIONS;;;
;;;;;;;;;;;;;;;;;;;;;

;;converts PC pitch-classes (0 - 11+) into a pch at OCTAVE 
;;(pc-to-pch '(0 7 12) 5)-------->(5.0 5.07 6.0)
;;(pc-to-pch '(0 7 12) 5.01)----->(5.01 5.08 6.01)
(defun pc-to-pch (pc &optional (octave 5))
  (if (listp pc)
      (mapcar #'(lambda (x) (pc-to-pch x octave)) pc)
    (fix-pch (+ octave (* pc .01)))))
  

;;input can be a list
(defun pch-to-cps (pch)
  (if (listp pch)
      (mapcar #'pch-to-cps pch)
    (let ((12throot2 (expt 2 1/12))
          (distance (- (make-midi pch) midi-c4)))
      (* cps-c4 (expt 12throot2 distance)))))


;;adjusts pch such as 5.12 -->6.0
(defun fix-pch (pch)
  (let ((pitch (round (* 100 (- pch (truncate pch))))))
    (if (> pitch 11)
        (fix-pch (round-to-2 (+ pch 1 -.12)))
      (round-pch pch))))

(defun make-pch (item)
  (if (atom item)
      (if (numberp item)
          (round-pch (if (floatp item) item (midi-to-pch item)))
        (named-to-pch item))
    (mapcar #'make-pch item)))

(defun make-midi (item)
  (if (atom item)
      (if (floatp item) (pch-to-midi item) item)
    (mapcar #'make-midi item)))

(defun transpose-midi (pitch transpose)
  (+ pitch transpose))


(defgeneric convert-pitch (from to list)
  (:documentation 
   "Converts LIST FROM one pitch type TO another, i.e. 'pch TO 'midi."))

(defmethod convert-pitch ((from (eql 'pch))(to (eql 'midi)) list)
  (mapcar #'pch-to-midi list))

(defmethod convert-pitch ((from (eql 'midi))(to (eql 'pch)) list)
  (mapcar #'midi-to-pch list))

;;converts a "named" pitch, i.e. C4, Db2, etc., to pch notation
(defun named-to-pch (symbol)
  (if (numberp symbol) (make-pch symbol)
    (let* ((named (coerce (string symbol) 'list))
           (item1 (read-from-string (string (first named))))
           (item2 (second named))
           (pitch (cond ((eql item1 'c) .0)
                        ((eql item1 'd) .02)
                        ((eql item1 'e) .04)
                        ((eql item1 'f) .05)
                        ((eql item1 'g) .07)
                        ((eql item1 'a) .09)
                        ((eql item1 'b) .11)
                        (t (error "NAMED-TO-PITCH: ~A is not a valid pitch." item1))))
           (pitch (cond ((eql item2 #\#) (+ pitch .01))
                        ((eql item2 #\B) (- pitch .01))
                        (t pitch)))                        
           (octave (if (not (or (eql item2 #\#)(eql item2 #\B)))
                       (+ (- pch-c4 4) (read-from-string (string item2)))
                     (+ (- pch-c4 4) (read-from-string (string (third named)))))))
      (round-pch (+ pitch octave)))))
      
         

(defun pch-to-midi (pch)
  (let* ((octave-pch (round (- (round pch) pch-c4)))
         (pc (round (* 100 (rem (round-pch pch) 1))))
         (octave-midi (+ midi-c4 (* 12 octave-pch)))
         (midi (+ octave-midi pc)))
    (if (> midi 0) midi 0)))

(defun midi-to-pch (midi &aux (midi (round midi)))
  (let* ((octave-midi (- (midi-octave midi) (midi-octave midi-c4)))
         (pc (round-to-n 2 (* .01 (midi-to-pc midi))))
         (octave-pch (+ octave-midi (round pch-c4)))
         (pch (+ octave-pch pc)))
    (round-to-n 2 pch)))

(defun pch-to-pc (pitch)
  (round (* 100 (rem pitch 1))))

(defun midi-to-pc (midi)
  (cond ((> midi 11)
         (midi-to-pc (- midi 12)))
        ((< midi 0)
         (midi-to-pc (+ midi 12)))
        (t
         midi)))

(defun midi-octave (midi &optional (cur '0))
  (cond ((> midi 11) 
         (midi-octave (- midi 12) (1+ cur)))
        ((< midi 0)
         (midi-octave (+ midi 12) (1- cur)))
        (t
         cur)))

(defun midi-in-pc-setp (midi pc-set)
  (cond ((> midi 11)
         (midi-in-pc-setp (- midi 12) pc-set))
        ((null pc-set)())
        ((= midi (first pc-set)) t)
        (t (midi-in-pc-setp midi (rest pc-set)))))


;;;need to optimize big time...this section is SLOOOOOOOOW.
#|
(defun find-available-pitches (pc-set range-lo range-hi)
  (let ((octave-lo (floor range-lo))
        (octave-hi (floor range-hi))
        (pc-lo (round-to-n 2 (rem range-lo 1)))
        (pc-hi (round-to-n 2 (rem range-hi 1))))
    (loop for
    (loop for x from octave-lo to octave-hi collect
          (+
|#
(defun find-available-pitches (pc-set range-lo range-hi &optional (current range-lo))
  (let ((next (next-pch current))
        (range-hi (round-pch range-hi))
        (range-lo (round-pch range-lo)))
  (cond ((> current range-hi) ())
        ((member (pch-to-pc current) pc-set) 
         (cons current (find-available-pitches pc-set range-lo range-hi next)))
        (t (find-available-pitches pc-set range-lo range-hi next)))))

(defun round-pch (pch)
  (let ((pch (round-to-n 2 pch))
        (pc (round (* 100 (rem pch 1)))))
    (if (> pc 11) (float (ceiling pch))
        pch)))

(defun round-to-2 (num)
  (float (/ (round (* 100 num)) 100)))

(defun next-pch (pch)
  (if (>= (round-to-2 (+ pch .01))
          (round-to-2 (+ 0.12 (round pch))))
      (round-to-2 (+ 1.0 (round pch)))
    (round-to-2 (+ 0.01 pch))))

(defun prev-pch (pch)
  (if (< (round-to-2 (- pch .01))
         (round-to-2 (floor pch)))
      (round-to-2 (- (round pch) 0.89))
    (round-to-2 (- pch .01))))

(defun print-hash (&optional (table *phrases*))
  (maphash #'(lambda (k s) (print (list k s))) table))

;;;;;;;;;;;;;
;;;FIRST-N;;;
;;;;;;;;;;;;;

(defvar *first-n-documentation* 
    "Returns the first N items of a list.  If N is nil, returns the entire list.")

(defvar *first-n-prompt*
    "(first-n n list)")

(defun first-n (n list)
  (setq *utility-error* 'first-n)
    (let ((n (round (get-data n)))
          (list (get-data list)))
      (loop for o from 0 below n collect
            (nth o list))))

;;;;;;;;;;;;
;;;LAST-N;;;
;;;;;;;;;;;;

(defvar *last-n-prompt*
    "(last-n n list)")

(defvar *last-n-documentation* '())

(defun last-n (n list)
  (let ((list (get-data list)))
    (reverse (first-n n (reverse list))))) 


;;;;;;;;;;;;;;;
;;;TRANSPOSE;;;
;;;;;;;;;;;;;;;

(defvar *transpose-documentation* '())

(defvar *transpose-prompt*
    "(transpose list steps &key (type 'pch))")

(defun transpose (list steps &key (type 'pch))
  (setq *function-error* 'transpose)
  (let* ((list (check-list (get-data list)))
         (steps (check-and-modify-length (get-data steps) list)))
    (mapcar #'(lambda (x y) (transpose-f x (round y) type)) list steps)))

(defgeneric transpose-f (list steps type))

(defmethod transpose-f (list steps type)
  (declare (ignore steps list))
  (make-error (format nil "Sorry, bad data for TRANSPOSE-F type ~S." type)))

(defmethod transpose-f ((list list) (steps fixnum) (type (eql 'pch)))
  (setq *function-error* 'transpose-f)
  (let ((midi-list (convert-pitch 'pch 'midi list)))
    (convert-pitch 'midi 'pch (transpose-f midi-list steps 'midi))))

(defmethod transpose-f ((list number) (steps fixnum) (type (eql 'pch)))
  (let ((pitch (pch-to-midi list)))
    (midi-to-pch (transpose-midi pitch steps))))

(defmethod transpose-f ((list list) (steps fixnum) (type (eql 'midi)))
  (mapcar #'(lambda (x) (transpose-midi x steps)) list))

(defmethod transpose-f (list (steps float) type)
  (transpose-f list (round steps) type))

(defmethod transpose-f (list (steps ratio) type)
  (transpose-f list (round steps) type))

(defmethod transpose-f (list (steps list) type)
  (mapcar #'(lambda (x y) (transpose-f x (round y) type)) steps list))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#| I don't know if these are needed, so I'll wait until they are to save bloat.

;;;;;;;;;;;;;;;;
;;;GET-number;;;
;;;;;;;;;;;;;;;;

(defgeneric get-number (number)
  (:documentation 
   "Finds the first numberp in a list"))

(defmethod get-number ((number number))
  number)

(defmethod get-number ((object symbol))
  (when object (get-number (eval object))))

(defmethod get-number ((object t))
  )

(defmethod get-number ((list list))
  (let ((number (first list)))
    (if (numberp number) number
      (get-number (rest list)))))
|#


  